' =============================================================
' This is a special version with the DropDownListBox and
' TextBox control procedures removed to eliminate out of
' memory problems when compiling from the command line
' under Windows 2000 (no problems under Windows 98 SE).
' =============================================================
' DIALOG.BAS, v1.2
'
' Copyright (C) 2002 Michael Webster
'
' This program is free software; you can redistribute it and/or
' modify it under the terms of the GNU General Public License as
' published by the Free Software Foundation; either version 2 of
' the License, or (at your option) any later version.
'
' This program is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public
' License along with this program; if not, write to the Free
' Software Foundation, Inc., 59 Temple Place, Suite 330,
' Boston, MA  02111-1307  USA

DEFINT A-Z

' This file contains a QuickBASIC module that provides the core components
' for a text mode, dialog based user interface with mouse support.
'
' To use these procedures within the QuickBASIC or QuickBASIC Extended
' programming environment, you must load the default Quick library
' (QB.QLB or QBX.QLB) when the programming environment is started.
' Refer to QUICKLIB.TXT for more information.
'
' Note that this module, in combination with the necessary support modules,
' has substantial variable storage requirements. One way to reduce these
' memory requirements is to minimize the number of windows, the number of
' controls per window, and the value of cMaxControls (declared below).
'
' To help reduce the size of the source code, the bulk of the header
' comments were moved to DIALOG.TXT.

' $INCLUDE: 'INTX.BI'
' $INCLUDE: 'DIALOG.BI'
' $INCLUDE: 'GENERAL.BI'
' $INCLUDE: 'MOUSE.BI'
' $INCLUDE: 'ERROR.BI'

CONST cFalse = 0, cTrue = NOT cFalse

' Declare the procedures that are private to this module.
DECLARE SUB InitModule ()
DECLARE FUNCTION ActiveControlType% ()
DECLARE SUB SetActiveControl (callerIndex%, direction%)
DECLARE SUB PrintControlCap (row%, col%, fore%, back%, accessFore%, accessCol%, caption$)

' Declare the constants, data types, and variables that are private to
' this module.

' Declare a constant for the maximum number of controls per window.
CONST cMaxControls = 20

' Declare constants for the SetActiveControl direction parameter.
CONST cAcThis = 0
CONST cAcNext = 1
CONST cAcPrev = 2
CONST cAcUp = 3
CONST cAcDown = 4
CONST cAcLeft = 5
CONST cAcRight = 6

' Define a data type and declare a shared array to store window information.
TYPE WindowDataType
    row             AS INTEGER
    col             AS INTEGER
    rows            AS INTEGER
    cols            AS INTEGER
    fore            AS INTEGER
    back            AS INTEGER
    fTitleBar       AS INTEGER
    activeControl   AS INTEGER
    controlCount    AS INTEGER
END TYPE
DIM SHARED sWindowData(0 TO 7) AS WindowDataType

' Declare constants for the control types. A list box control will enable
' an internal item search feature whenever the fSearch parameter is true.
' The command button, option button, check box, and label controls allow
' the user to select the control by pressing the access key without the
' Alt key whenever the active control does not trap character keys
' internally (i.e., whenever the active control is not a text box or a
' list box with the internal item search feature enabled). To make this
' work correctly, we need a separate type constant for a list box (or a
' drop-down list box) with the internal item search feature enabled.
' Also, we need a separate type constant for a disabled command button
' so the SetActiveControl procedure can correctly bypass such buttons.
CONST cCtCheckBox = 1
CONST cCtCommandButton = 2
CONST cCtDisabledCommandButton = 3
CONST cCtDropDownListBox = 4
CONST cCtDropDownListBoxWithSearch = 5
CONST cCtLabel = 6
CONST cCtListBox = 7
CONST cCtListBoxWithSearch = 8
CONST cCtOptionButton = 9
CONST cCtTextBox = 10

' Define a data type and declare a shared array to store control
' information. The first array dimension identifies the window and
' the second identifies the control (by index). The row and column
' elements are used for keyboard navigation. The fScrollBar element
' is for list box controls. The group and selected state elements
' are for option button controls.
TYPE ControlDataType
    controlType     AS INTEGER
    row             AS INTEGER
    col             AS INTEGER
    lastRow         AS INTEGER
    lastCol         AS INTEGER
    fScrollBar      AS INTEGER
    group           AS INTEGER
    fSelected       AS INTEGER
END TYPE
DIM SHARED sControlData(0 TO 7, 1 TO cMaxControls) AS ControlDataType

' Declare a shared variable to store the number of the most recently
' opened window. Each window is on a separate screen page, so the value
' in this variable also identifies the active and visible screen page.
DIM SHARED sCurrentWindow

' Declare a shared flag to store the insert/typeover state for all controls.
DIM SHARED sfTypeOver

' Declare a shared flag to control keyboard input. The flag is set by a
' control procedure whenever the current keyboard input has been handled
' and should be discarded. The flag is cleared during the next call to
' the GetUserInput procedure.
DIM SHARED sfDiscardKey

' Declare a shared flag to control access to mouse input. This flag is set
' by the first control to detect that the left mouse button is pressed and
' the mouse cursor is on the control, and cleared by the same control when
' the left mouse button is released. Other controls ignore the mouse while
' the flag is set.
DIM SHARED sfMouseLock
                                                                                                                                                                   
' Declare a shared flag to store the module initialization state.
DIM SHARED sfInitModuleOK

' Declare a shared flag to enable toggling of the display between the
' current window and screen page 0 with the F4 key.
DIM SHARED sfEnableToggle

' *** START TEST CODE ***
' *** END TEST CODE ***

END

FUNCTION ActiveControlIndex%

    ' Returns the tab index for the active control in
    ' the current window.

    ActiveControlIndex = sWindowData(sCurrentWindow).activeControl

END FUNCTION

FUNCTION ActiveControlType

    ' Returns the control type for the active control in the
    ' current window.
    '
    ' Constants for the control types are defined in the module
    ' level code of this module.
   
    ' The following statement will trigger a Subscript out of range
    ' error if the current window contains only label controls and
    ' the user presses an access key for one of them.
    ActiveControlType = sControlData(sCurrentWindow, sWindowData(sCurrentWindow).activeControl).controlType

END FUNCTION

SUB CloseWindow

    ' Resets the control count for the current window and removes the
    ' window from the screen.
    '
    ' The window is removed from the screen by switching the active and
    ' visible screen page to the previous screen page, thus restoring
    ' the image preserved most recently by the OpenWindow procedure.
    '
    ' Note that this procedure is also used to restore the image preserved
    ' by the OpenScreen procedure.

    ' Simulate an Illegal function call error if no windows are open.
    IF sCurrentWindow < 1 THEN
        LogError "DIALOG.BAS", "CloseWindow", "sCurrentWindow < 1"
        ERROR cErrorIllegalFunctionCall
    END IF

    sWindowData(sCurrentWindow).controlCount = 0
    sCurrentWindow = sCurrentWindow - 1
  
    HideMouseCursor

    SCREEN , , sCurrentWindow, sCurrentWindow
    
    IF sCurrentWindow THEN ShowMouseCursor

    ' This statement solves a problem with the drop down list box
    ' procedure causing the module to ignore keyboard input under
    ' specific conditions.
    sfDiscardKey = cTrue

END SUB

SUB CpCheckBox (chk AS CheckBoxType, caption$)

    ' Displays and operates a check box control.

    ' If this is the first call, then init and return.
    IF chk.internal.tabindex = 0 THEN

        IF sCurrentWindow < 1 THEN
            LogError "DIALOG.BAS", "CpCheckBox", "sCurrentWindow < 1"
            ERROR cErrorIllegalFunctionCall
        END IF
        IF caption$ = "" THEN
            LogError "DIALOG.BAS", "CpCheckBox", "caption$ = <null>"
            ERROR cErrorIllegalFunctionCall
        END IF

        ' Get the position of the access key character and remove
        ' the delimiter.
        accessPosition = INSTR(caption$, "&")
        IF accessPosition THEN
            cap$ = LEFT$(caption$, accessPosition - 1) + MID$(caption$, accessPosition + 1)
        ELSE
            cap$ = caption$
        END IF

        chk.internal.lastCol = chk.col + LEN(cap$) + 3

        IF chk.row < 1 THEN
            LogError "DIALOG.BAS", "CpCheckBox", "chk.row < 1"
            ERROR cErrorIllegalFunctionCall
        END IF
        IF chk.row > 25 THEN
            LogError "DIALOG.BAS", "CpCheckBox", "chk.row > 25"
            ERROR cErrorIllegalFunctionCall
        END IF
        IF chk.col < 1 THEN
            LogError "DIALOG.BAS", "CpCheckBox", "chk.col < 1"
            ERROR cErrorIllegalFunctionCall
        END IF
        IF chk.internal.lastCol > 80 THEN
            LogError "DIALOG.BAS", "CpCheckBox", "chk.internal.lastCol > 80"
            ERROR cErrorIllegalFunctionCall
        END IF

        ' Save the access key character and the Alt+ access key code.
        IF accessPosition THEN
            chk.internal.accessKey = UCASE$(MID$(cap$, accessPosition, 1))
            chk.internal.accessCode = AltPlusKeyCode(MID$(cap$, accessPosition, 1))
        END IF

        ' Bump the control count for the current window and copy it to the
        ' tab index element for this check box.
        sWindowData(sCurrentWindow).controlCount = sWindowData(sCurrentWindow).controlCount + 1
        chk.internal.tabindex = sWindowData(sCurrentWindow).controlCount

        ' Copy the data for this check box to the shared array.
        sControlData(sCurrentWindow, chk.internal.tabindex).controlType = cCtCheckBox
        sControlData(sCurrentWindow, chk.internal.tabindex).row = chk.row
        sControlData(sCurrentWindow, chk.internal.tabindex).col = chk.col
        sControlData(sCurrentWindow, chk.internal.tabindex).lastRow = chk.row
        sControlData(sCurrentWindow, chk.internal.tabindex).lastCol = chk.internal.lastCol

        ' Add the check box to the caption.
        IF chk.fChecked THEN
            cap$ = "[X] " + cap$
        ELSE
            cap$ = "[ ] " + cap$
        END IF

        ' Print the caption.
        fore = sWindowData(sCurrentWindow).fore
        back = sWindowData(sCurrentWindow).back
        IF accessPosition THEN
            accessCol = chk.col + accessPosition + 3
        END IF
        PrintControlCap chk.row, chk.col, fore, back, cAccessFore, accessCol, cap$

    ELSE
        GetUserInput key$, keyCode, mouseRow, mouseCol, fMouseLeft, 0, chk.internal.tabindex

        ' Handle any mouse input. The chk.internal.fMouseLeft flag stores
        ' the left mouse button state between calls to this instance.
       
        IF fMouseLeft THEN

            ' If the mouse cursor is on this check box...
            IF mouseRow = chk.row AND mouseCol >= chk.col AND mouseCol <= chk.internal.lastCol THEN

                ' If no control has responded, then make this check box
                ' the active control and set the necessary flags.
                IF NOT chk.internal.fMouseLeft AND NOT sfMouseLock THEN
                    SetActiveControl chk.internal.tabindex, cAcThis
                    chk.internal.fMouseLeft = cTrue
                    sfMouseLock = cTrue
                END IF
            END IF

        ELSEIF chk.internal.fMouseLeft THEN
            ' The left mouse button has been released so clear the flags.
            chk.internal.fMouseLeft = cFalse
            sfMouseLock = cFalse

            ' If the mouse cursor is still on this check box, then
            ' toggle the checked state.
            IF mouseRow = chk.row AND mouseCol >= chk.col AND mouseCol <= chk.internal.lastCol THEN
                chk.fChecked = NOT chk.fChecked
            END IF

        END IF

        ' Handle any keyboard input. Set the shared
        ' discard key flag for any key handled.

        ' If this check box has an access key...
        IF chk.internal.accessCode THEN

            ' If the user pressed the access key without the Alt key...
            IF UCASE$(key$) = chk.internal.accessKey THEN

                ' If the active control is not a text box or a
                ' list box with the internal item search feature
                ' enabled, then make this check box the active
                ' control and toggle the checked state.
                act = ActiveControlType
                IF act <> cCtTextBox AND act <> cCtListBoxWithSearch AND act <> cCtDropDownListBoxWithSearch THEN
                    SetActiveControl chk.internal.tabindex, cAcThis
                    chk.fChecked = NOT chk.fChecked
                    sfDiscardKey = cTrue
                END IF

            ELSEIF keyCode = chk.internal.accessCode THEN
                ' If the user pressed the access key in combination with
                ' the Alt key, then make this check box the active control
                ' and toggle the checked state.
                SetActiveControl chk.internal.tabindex, cAcThis
                chk.fChecked = NOT chk.fChecked
                sfDiscardKey = cTrue
            END IF
        END IF

        ' Handle any other keys only if this check box is the active control.
        IF sWindowData(sCurrentWindow).activeControl = chk.internal.tabindex THEN
          
            SELECT CASE keyCode
                CASE cKeyTab
                    SetActiveControl chk.internal.tabindex, cAcNext
                    sfDiscardKey = cTrue
        
                CASE cKeyShiftTab
                    SetActiveControl chk.internal.tabindex, cAcPrev
                    sfDiscardKey = cTrue

                CASE cKeyUp
                    SetActiveControl chk.internal.tabindex, cAcUp
                    sfDiscardKey = cTrue
        
                CASE cKeyDown
                    SetActiveControl chk.internal.tabindex, cAcDown
                    sfDiscardKey = cTrue

                CASE cKeyLeft
                    SetActiveControl chk.internal.tabindex, cAcLeft
                    sfDiscardKey = cTrue

                CASE cKeyRight
                    SetActiveControl chk.internal.tabindex, cAcRight
                    sfDiscardKey = cTrue
             
                CASE cKeyInsert
                    sfTypeOver = NOT sfTypeOver
                    sfDiscardKey = cTrue

                CASE cKeySpace
                    chk.fChecked = NOT chk.fChecked
                    sfDiscardKey = cTrue

            END SELECT
        END IF
    END IF

    ' If the checked state has changed since the last call, then update
    ' the visual state and copy the new state to an instance variable.
    IF chk.fChecked <> chk.internal.fChecked THEN
        SetColor sWindowData(sCurrentWindow).fore, sWindowData(sCurrentWindow).back
        HideTextCursor
        LOCATE chk.row, chk.col + 1
        HideMouseCursorIf chk.row, chk.col + 1, 1, 1
        IF chk.fChecked THEN
            PRINT "X";
            chk.internal.fChecked = cTrue
        ELSE
            PRINT " ";
            chk.internal.fChecked = cFalse
        END IF
        ShowMouseCursor
    END IF

    ' If this check box is the active control, then position the
    ' text cursor in the box and make it visible.
    IF sWindowData(sCurrentWindow).activeControl = chk.internal.tabindex THEN
        LOCATE chk.row, chk.col + 1
        ShowTextCursor
    END IF

END SUB

SUB CpCommandButton (cmd AS CommandButtonType, caption$)
    
    ' Displays and operates a command button control.
    '
    ' Returns with the fClicked element of <cmd> set to true if the user:
    '
    '   Clicked the button with the mouse.
    '
    '   Pressed the access key in combination with the Alt key.
    '
    '   Pressed the access key alone (if the active control is not a text
    '   box or a list box with the internal item search feature enabled).
    '
    '   Pressed Enter (if the button is the active control or the default
    '   flag is set and the active control is not a command button).
    '
    '   Pressed the spacebar (if the button is the active control).
    '
    '   Pressed Escape (if the cancel flag is set).
  
    ' Get the position of the access key character, remove the
    ' delimiter, and save the access key character and the Alt+
    ' access key code. This must be done on each call because the
    ' caption argument can be changed between calls and the
    ' caption must be printed each time the actuated state changes.
    cap$ = caption$
    accessPosition = INSTR(caption$, "&")
    IF accessPosition THEN
        cap$ = LEFT$(cap$, accessPosition - 1) + MID$(cap$, accessPosition + 1)
        cmd.internal.accessKey = UCASE$(MID$(cap$, accessPosition, 1))
        cmd.internal.accessCode = AltPlusKeyCode(cmd.internal.accessKey)
    END IF
    IF cap$ = "" THEN
        LogError "DIALOG.BAS", "CpCommandButton", "cap$ = <null>"
        ERROR cErrorIllegalFunctionCall
    END IF

    ' If this is the first call, then init and return.
    IF cmd.internal.tabindex = 0 THEN

        IF sCurrentWindow < 1 THEN
            LogError "DIALOG.BAS", "CpCommandButton", "sCurrentWindow < 1"
            ERROR cErrorIllegalFunctionCall
        END IF

        ' Allow for the button shadow when calculating the last column.
        cmd.internal.lastCol = cmd.col + LEN(cap$)
        lastRow = cmd.row + 1

        IF cmd.row < 1 THEN
            LogError "DIALOG.BAS", "CpCommandButton", "cmd.row < 1"
            ERROR cErrorIllegalFunctionCall
        END IF
        IF lastRow > 25 THEN
            LogError "DIALOG.BAS", "CpCommandButton", "lastRow > 25"
            ERROR cErrorIllegalFunctionCall
        END IF
        IF cmd.col < 1 THEN
            LogError "DIALOG.BAS", "CpCommandButton", "cmd.col < 1"
            ERROR cErrorIllegalFunctionCall
        END IF
        IF cmd.internal.lastCol > 80 THEN
            LogError "DIALOG.BAS", "CpCommandButton", "cmd.internal.lastCol > 80"
            ERROR cErrorIllegalFunctionCall
        END IF

        ' Bump the control count for the current window and copy it to the
        ' tab index element for this command button.
        sWindowData(sCurrentWindow).controlCount = sWindowData(sCurrentWindow).controlCount + 1
        cmd.internal.tabindex = sWindowData(sCurrentWindow).controlCount

        ' Copy the data for this command button to the shared array.
        IF cmd.fDisabled THEN
            sControlData(sCurrentWindow, cmd.internal.tabindex).controlType = cCtDisabledCommandButton
            ' Set the internal disabled flag so
            ' we can detect a run-time change.
            cmd.internal.fDisabled = cmd.fDisabled
        ELSE
            sControlData(sCurrentWindow, cmd.internal.tabindex).controlType = cCtCommandButton
        END IF
        sControlData(sCurrentWindow, cmd.internal.tabindex).row = cmd.row
        sControlData(sCurrentWindow, cmd.internal.tabindex).col = cmd.col
        sControlData(sCurrentWindow, cmd.internal.tabindex).lastRow = cmd.row
        sControlData(sCurrentWindow, cmd.internal.tabindex).lastCol = cmd.internal.lastCol - 1

        ' If there is no active control for the current window and this
        ' command button is not disabled, then make this command button
        ' the active control and the current default.
        IF sWindowData(sCurrentWindow).activeControl = 0 AND NOT cmd.fDisabled THEN
            SetActiveControl cmd.internal.tabindex, cAcThis
            cmd.internal.fCurrentDefault = cTrue
        END IF
        
        ' Print the command button in the normal position.
        GOSUB PrintNormal

    ELSE
        ' Clear the return value.
        cmd.fClicked = cFalse

        ' Handle any changes in the enabled state.
        IF cmd.fDisabled THEN
            IF NOT cmd.internal.fDisabled THEN
                cmd.internal.fDisabled = cmd.fDisabled
                cmd.internal.fActuated = cFalse
                sControlData(sCurrentWindow, cmd.internal.tabindex).controlType = cCtDisabledCommandButton
                cmd.internal.fCurrentDefault = cFalse
                GOSUB PrintNormal
                IF sWindowData(sCurrentWindow).activeControl = cmd.internal.tabindex THEN
                    SetActiveControl cmd.internal.tabindex, cAcNext
                END IF
                sfDiscardKey = cTrue
            END IF
            EXIT SUB
        ELSE
            IF cmd.internal.fDisabled THEN
                cmd.internal.fDisabled = cmd.fDisabled
                sControlData(sCurrentWindow, cmd.internal.tabindex).controlType = cCtCommandButton
                cmd.fUpdate = cTrue
            END IF
        END IF

        ' If the fUpdate parameter is set, then
        ' reprint the command button.
        IF cmd.fUpdate THEN
            IF cmd.internal.fActuated THEN
                GOSUB PrintActuated
            ELSE
                GOSUB PrintNormal
            END IF
        END IF
       
        ' Clear the fUpdate parameter to avoid a repeat.
        cmd.fUpdate = cFalse

        GetUserInput key$, keyCode, mouseRow, mouseCol, fMouseLeft, 0, cmd.internal.tabindex

        ' Handle any mouse input.
        IF fMouseLeft THEN

            ' If the mouse cursor is on this command button...
            IF mouseRow = cmd.row AND mouseCol >= cmd.col AND mouseCol < cmd.internal.lastCol THEN

                ' If no control has responded, then make this command button
                ' the active control and the current default, print the
                ' command button in the actuated position, and set the
                ' necessary flags.
                IF NOT cmd.internal.fActuated AND NOT sfMouseLock THEN
                    SetActiveControl cmd.internal.tabindex, cAcThis
                    cmd.internal.fCurrentDefault = cTrue
                    GOSUB PrintActuated
                    cmd.internal.fActuated = cTrue
                    sfMouseLock = cTrue
                END IF

            ' If the mouse cursor is not on this command button and this
            ' command button is actuated...
            ELSEIF cmd.internal.fActuated THEN
              
                ' Print the command button in the normal position and
                ' clear the actuated flag.
                GOSUB PrintNormal
                cmd.internal.fActuated = cFalse
            END IF

        ELSE
            ' The left mouse button has been released.
            sfMouseLock = cFalse

            ' If this command button is actuated...
            IF cmd.internal.fActuated THEN
           
                ' Print the command button in the normal position
                ' and clear the actuated flag.
                GOSUB PrintNormal
                cmd.internal.fActuated = cFalse
          
                ' If the mouse cursor is still on this command button,
                ' then set the return value and return to the caller.
                IF mouseRow = cmd.row AND mouseCol >= cmd.col AND mouseCol <= cmd.internal.lastCol THEN
                    cmd.fClicked = cTrue
                    EXIT SUB
                END IF
            END IF
        END IF

        ' Handle any keyboard input. Set the shared discard key flag
        ' for any key handled.

        ' If this command button has an access key...
        IF cmd.internal.accessCode THEN
       
            ' Unlike the other controls, keyboard selection of a
            ' command button does NOT change the active control.

            ' If the user pressed the access key without the Alt key...
            IF UCASE$(key$) = cmd.internal.accessKey THEN

                ' If the active control is not a text box or a list
                ' box with the internal item search feature enabled,
                ' then set the return value and return to the caller.
                act = ActiveControlType
                IF act <> cCtTextBox AND act <> cCtListBoxWithSearch AND act <> cCtDropDownListBoxWithSearch THEN
                    cmd.fClicked = cTrue
                    sfDiscardKey = cTrue
                    EXIT SUB
                END IF

            ELSEIF keyCode = cmd.internal.accessCode THEN
                ' If the user pressed the access key in combination
                ' with the Alt key, then set the return value, and
                ' return to the caller.
                cmd.fClicked = cTrue
                sfDiscardKey = cTrue
                EXIT SUB
            END IF

        END IF

        ' Handle the enter, space, and escape keys.
        SELECT CASE keyCode
            CASE cKeyEnter
                ' If this command button is the active control, or the
                ' default flag is set and the active control is not a
                ' command button, then set the return value and return
                ' to the caller.
                IF sWindowData(sCurrentWindow).activeControl = cmd.internal.tabindex THEN
                    cmd.fClicked = cTrue
                    sfDiscardKey = cTrue
                    EXIT SUB
                ELSEIF sControlData(sCurrentWindow, sWindowData(sCurrentWindow).activeControl).controlType <> cCtCommandButton THEN
                    IF cmd.fDefault THEN
                        cmd.fClicked = cTrue
                        sfDiscardKey = cTrue
                        EXIT SUB
                    END IF
                END IF

            CASE cKeySpace
                ' If this command button is the active control, then
                ' set the return value and return to the caller.
                IF sWindowData(sCurrentWindow).activeControl = cmd.internal.tabindex THEN
                    cmd.fClicked = cTrue
                    sfDiscardKey = cTrue
                    EXIT SUB
                END IF

            CASE cKeyEscape
                ' If this command button is the Cancel button, then
                ' set the return value and return to the caller.
                IF cmd.fCancel THEN
                    cmd.fClicked = cTrue
                    sfDiscardKey = cTrue
                    EXIT SUB
                END IF

        END SELECT

        ' Handle any other keys only if this command button is the
        ' active control. Check the active control before clearing
        ' the current default flag and updating the visual state
        ' because SetActiveControl may fail if the direction argument
        ' has any value other than cAcThis.
        IF sWindowData(sCurrentWindow).activeControl = cmd.internal.tabindex THEN
            
            SELECT CASE keyCode
                CASE cKeyTab
                    SetActiveControl cmd.internal.tabindex, cAcNext
                    IF sWindowData(sCurrentWindow).activeControl <> cmd.internal.tabindex THEN
                        cmd.internal.fCurrentDefault = cFalse
                        GOSUB PrintNormal
                    END IF
                    sfDiscardKey = cTrue
          
                CASE cKeyShiftTab
                    SetActiveControl cmd.internal.tabindex, cAcPrev
                    IF sWindowData(sCurrentWindow).activeControl <> cmd.internal.tabindex THEN
                        cmd.internal.fCurrentDefault = cFalse
                        GOSUB PrintNormal
                    END IF
                    sfDiscardKey = cTrue

                CASE cKeyUp
                    SetActiveControl cmd.internal.tabindex, cAcUp
                    IF sWindowData(sCurrentWindow).activeControl <> cmd.internal.tabindex THEN
                        cmd.internal.fCurrentDefault = cFalse
                        GOSUB PrintNormal
                    END IF
                    sfDiscardKey = cTrue
          
                CASE cKeyDown
                    SetActiveControl cmd.internal.tabindex, cAcDown
                    IF sWindowData(sCurrentWindow).activeControl <> cmd.internal.tabindex THEN
                        cmd.internal.fCurrentDefault = cFalse
                        GOSUB PrintNormal
                    END IF
                    sfDiscardKey = cTrue

                CASE cKeyLeft
                    SetActiveControl cmd.internal.tabindex, cAcLeft
                    IF sWindowData(sCurrentWindow).activeControl <> cmd.internal.tabindex THEN
                        cmd.internal.fCurrentDefault = cFalse
                        GOSUB PrintNormal
                    END IF
                    sfDiscardKey = cTrue

                CASE cKeyRight
                    SetActiveControl cmd.internal.tabindex, cAcRight
                    IF sWindowData(sCurrentWindow).activeControl <> cmd.internal.tabindex THEN
                        cmd.internal.fCurrentDefault = cFalse
                        GOSUB PrintNormal
                    END IF
                    sfDiscardKey = cTrue
               
                CASE cKeyInsert
                    sfTypeOver = NOT sfTypeOver
                    sfDiscardKey = cTrue

            END SELECT
        END IF
    END IF

    ' The code above handles any changes in the active control that
    ' originate in this instance. The code below handles any changes
    ' in the active control that originate in other controls or other
    ' instances of this control.
    '
    ' The current default command button has button arrows. An active
    ' command button is always the current default. The default command
    ' button (if there is one) is the current default command button
    ' whenever the active control is not a command button.
    '
    ' If the button is not actuated and active or if the current default
    ' state has changed, then print the button in the normal position.
    IF NOT cmd.internal.fActuated THEN
        IF sWindowData(sCurrentWindow).activeControl = cmd.internal.tabindex THEN
            IF NOT cmd.internal.fCurrentDefault OR NOT cmd.internal.fActive THEN
                cmd.internal.fCurrentDefault = cTrue
                cmd.internal.fActive = cTrue
                GOSUB PrintNormal
            END IF

        ' The following statement will trigger a Subscript
        ' out of range error if the first command button in
        ' the tab order is disabled.
        ELSEIF cmd.fDefault AND sControlData(sCurrentWindow, sWindowData(sCurrentWindow).activeControl).controlType <> cCtCommandButton THEN
            IF NOT cmd.internal.fCurrentDefault THEN
                cmd.internal.fCurrentDefault = cTrue
                GOSUB PrintNormal
            END IF
            cmd.internal.fActive = cFalse
        ELSE
            IF cmd.internal.fCurrentDefault THEN
                cmd.internal.fCurrentDefault = cFalse
                GOSUB PrintNormal
            END IF
            cmd.internal.fActive = cFalse
        END IF
    END IF

    ' If this command button is the active control, then position the text
    ' cursor at the first non-space character of the caption and make it
    ' visible.
    IF sWindowData(sCurrentWindow).activeControl = cmd.internal.tabindex THEN
        FOR i = 2 TO LEN(cap$)
            IF MID$(cap$, i, 1) <> " " THEN EXIT FOR
        NEXT
        LOCATE cmd.row, cmd.col + i - 1
        ShowTextCursor
    END IF

EXIT SUB

PrintNormal:
    ' Print the command button in the normal (not actuated) position.
    GOSUB AddButtonArrows
    IF accessPosition THEN
        accessCol = cmd.col + accessPosition - 1
    END IF
    
    IF cmd.fDisabled THEN
        PrintControlCap cmd.row, cmd.col, cCmdDisabledFore, cCmdBack, cCmdDisabledFore, accessCol, cap$
    ELSE
        PrintControlCap cmd.row, cmd.col, cCmdFore, cCmdBack, cCmdAccessFore, accessCol, cap$
    END IF

    ' Add the button shadow. Note that the shadow will cover the
    ' remaining (end) character from the actuated button.
    HideMouseCursorIf cmd.row, cmd.col, 2, cmd.internal.lastCol - cmd.col + 1
    SetColor cBlack, sWindowData(sCurrentWindow).back
    LOCATE cmd.row, cmd.internal.lastCol
    PRINT CHR$(220);
    LOCATE cmd.row + 1, cmd.col + 1
    PRINT STRING$(LEN(cap$), CHR$(223));
    ShowMouseCursor

    RETURN

PrintActuated:
    ' Print the command button in the actuated position.

    ' This will never be called for a disabled button.
    
    ' Remove the first character of the normal button and the lower
    ' part of the button shadow.
    SetColor sWindowData(sCurrentWindow).fore, sWindowData(sCurrentWindow).back
    HideTextCursor
    HideMouseCursorIf cmd.row, cmd.col, 2, cmd.internal.lastCol - cmd.col + 1
    LOCATE cmd.row, cmd.col
    PRINT " ";
    LOCATE cmd.row + 1, cmd.col + 1
    PRINT SPACE$(LEN(cap$));
    ShowMouseCursor
    
    ' Note that the entire button is shifted 1 character to the right.
    GOSUB AddButtonArrows
    IF accessPosition THEN
        accessCol = cmd.col + accessPosition
    END IF
    PrintControlCap cmd.row, cmd.col + 1, cCmdFore, cCmdBack, cCmdAccessFore, accessCol, cap$
    RETURN

AddButtonArrows:
    ' If this command button is the current default, then add button arrows.
    IF cmd.internal.fCurrentDefault THEN
        MID$(cap$, 1, 1) = CHR$(16)
        MID$(cap$, LEN(cap$), 1) = CHR$(17)
    END IF
    RETURN

END SUB

SUB CpLabel (lbl AS LabelType, caption$)

    ' Displays and operates a label control.
    '
    ' A label cannot be the active control and does not respond to any
    ' input other than the access key. If the user presses the access
    ' key and the active control is not a text box or a list box with
    ' the internal item search feature enabled, or the user presses
    ' the access key in combination with the Alt key, then the active
    ' control is set to the next control in the tab order.
    
    ' If this is the first call, then init and return.
    IF lbl.internal.tabindex = 0 THEN

        IF sCurrentWindow < 1 THEN
            LogError "DIALOG.BAS", "CpLabel", "sCurrentWindow < 1"
            ERROR cErrorIllegalFunctionCall
        END IF
        IF caption$ = "" THEN
            LogError "DIALOG.BAS", "CpLabel", "caption$ = <null>"
            ERROR cErrorIllegalFunctionCall
        END IF

        ' Get the position of the access key character and remove
        ' the delimiter.
        accessPosition = INSTR(caption$, "&")
        IF accessPosition THEN
            cap$ = LEFT$(caption$, accessPosition - 1) + MID$(caption$, accessPosition + 1)
        ELSE
            cap$ = caption$
        END IF

        lastCol = lbl.col + LEN(cap$) - 1

        IF lbl.row < 1 THEN
            LogError "DIALOG.BAS", "CpLabel", "lbl.row < 1"
            ERROR cErrorIllegalFunctionCall
        END IF
        IF lbl.row > 25 THEN
            LogError "DIALOG.BAS", "CpLabel", "lbl.row > 25"
            ERROR cErrorIllegalFunctionCall
        END IF
        IF lbl.col < 1 THEN
            LogError "DIALOG.BAS", "CpLabel", "lbl.col < 1"
            ERROR cErrorIllegalFunctionCall
        END IF
        IF lastCol > 80 THEN
            LogError "DIALOG.BAS", "CpLabel", "lastCol > 80"
            ERROR cErrorIllegalFunctionCall
        END IF

        ' Save the access key character and the Alt+ access key code.
        IF accessPosition THEN
            lbl.internal.accessKey = UCASE$(MID$(cap$, accessPosition, 1))
            lbl.internal.accessCode = AltPlusKeyCode(MID$(cap$, accessPosition, 1))
        END IF

        ' Bump the control count for the current window and copy
        ' it to the tab index element for this label.
        sWindowData(sCurrentWindow).controlCount = sWindowData(sCurrentWindow).controlCount + 1
        lbl.internal.tabindex = sWindowData(sCurrentWindow).controlCount

        ' Copy the data for this label to the shared array.
        sControlData(sCurrentWindow, lbl.internal.tabindex).controlType = cCtLabel
        sControlData(sCurrentWindow, lbl.internal.tabindex).row = lbl.row
        sControlData(sCurrentWindow, lbl.internal.tabindex).col = lbl.col
        sControlData(sCurrentWindow, lbl.internal.tabindex).lastRow = lbl.row
        sControlData(sCurrentWindow, lbl.internal.tabindex).lastCol = lastCol
        
        ' Print the caption.
        fore = sWindowData(sCurrentWindow).fore
        back = sWindowData(sCurrentWindow).back
        IF accessPosition THEN
            accessCol = lbl.col + accessPosition - 1
        END IF
        PrintControlCap lbl.row, lbl.col, fore, back, cAccessFore, accessCol, cap$

    ELSE
        GetUserInput key$, keyCode, mouseRow, mouseCol, fMouseLeft, 0, lbl.internal.tabindex

        ' Handle any keyboard input. Set the shared discard key flag
        ' for any key handled.

        ' If this label has an access key...
        IF lbl.internal.accessCode THEN

            ' If the user pressed the access key without the Alt key...
            IF UCASE$(key$) = lbl.internal.accessKey THEN

                ' If the active control is not a text box or a list
                ' box with the internal item search feature enabled,
                ' then set the active control to the next control
                ' in the tab order.
                act = ActiveControlType
                IF act <> cCtTextBox AND act <> cCtListBoxWithSearch AND act <> cCtDropDownListBoxWithSearch THEN
                    SetActiveControl lbl.internal.tabindex, cAcNext
                    sfDiscardKey = cTrue
                END IF

            ELSEIF keyCode = lbl.internal.accessCode THEN
                ' If the user pressed the access key in combination with the
                ' Alt key, then set the active control to the next control in
                ' the tab order.
                SetActiveControl lbl.internal.tabindex, cAcNext
                sfDiscardKey = cTrue
            END IF
        END IF
    END IF
   
END SUB

SUB CpListBox (lst AS ListBoxType, list$())

    ' Displays and operates a single selection list box control.
    '
    ' A vertical scroll bar is automatically displayed whenever the
    ' number of list items exceeds the number of visible items.
    '
    ' Setting the lst.itemCount parameter to a value > 0 specifies
    ' an "owner draw" list. This is useful for cases where the full
    ' list would exceed the available string capacity. For these
    ' cases, the window procedure can use the parameter to specify
    ' the number of list items, periodically check the current
    ' position in the list (lst.internal.firstVisibleItem), and
    ' generate and display the actual display data over the control.
    ' Note that item highlighting and the internal item search
    ' feature are disabled.
    '
    ' Setting the fDoubleClick parameter to true enables a feature
    ' that allows the user to double-click a list item to select the
    ' item and activate the current window's default command button.
    ' The double-click interval is declared as a local constant.
    '
    ' Setting the fSearch parameter to true enables an internal item
    ' search feature. Whenever the list box is the active control,
    ' this feature allows the user to search for an item by pressing
    ' the key which corresponds to the first character of the item
    ' (case-insensitive).
    '
    ' Setting the fText parameter to true adapts this control for
    ' viewing text. Note that setting this parameter to true disables
    ' item highlighting.
    '
    ' The fSelect flag was added to allow CpDropDownListBox to
    ' determine if a list item has been selected with the mouse.
    '
    ' This procedure was difficult to code and debug because it is
    ' too long and does too much!

    ' Declare constants for parameters that might be changed later.
    CONST cRepeatDelay = .5
    CONST cDoubleClickInterval = .2

    IF lst.itemCount THEN
        itemCount = lst.itemCount
    ELSE
        IF LBOUND(list$) <> 1 THEN
            LogError "DIALOG.BAS", "CpListBox", "LBOUND(list$) <> 1"
            ERROR cErrorIllegalFunctionCall
        END IF
        itemCount = UBOUND(list$)
    END IF
  
    lastRow = lst.row + lst.rows - 1
    lastCol = lst.col + lst.cols - 1

    ' If the number of items is greater than the number of item rows, then
    ' set a flag so a vertical scroll bar will be displayed. If the number
    ' of rows is greater than 6, then set a flag so the vertical scroll bar
    ' will contain a slider. Although the scroll bar would have room for a
    ' slider for any number of rows greater than 4, the value 6 makes the
    ' slider logic simpler.
    IF itemCount > lst.rows - 2 THEN
        lst.internal.fScrollBar = cTrue
        IF lst.rows > 6 THEN
            fSlider = cTrue
        ELSE
            fSlider = cFalse
        END IF
    ELSE
        lst.internal.fScrollBar = cFalse
        fSlider = cFalse
    END IF
    IF lst.internal.tabindex THEN
        sControlData(sCurrentWindow, lst.internal.tabindex).fScrollBar = lst.internal.fScrollBar
    END IF
   
    ' If this is the first call, then init and return.
    IF lst.internal.tabindex = 0 THEN

        IF sCurrentWindow < 1 THEN
            LogError "DIALOG.BAS", "CpListBox", "sCurrentWindow < 1"
            ERROR cErrorIllegalFunctionCall
        END IF
        IF lst.rows < 4 THEN
            LogError "DIALOG.BAS", "CpListBox", "lst.rows < 4"
            ERROR cErrorIllegalFunctionCall
        END IF
        IF lst.cols < 6 THEN
            LogError "DIALOG.BAS", "CpListBox", "lst.cols < 6"
            ERROR cErrorIllegalFunctionCall
        END IF
        IF lst.row < 1 THEN
            LogError "DIALOG.BAS", "CpListBox", "lst.row < 1"
            ERROR cErrorIllegalFunctionCall
        END IF
        IF lastRow > 25 THEN
            LogError "DIALOG.BAS", "CpListBox", "lastRow > 25"
            ERROR cErrorIllegalFunctionCall
        END IF
        IF lst.col < 1 THEN
            LogError "DIALOG.BAS", "CpListBox", "lst.col < 1"
            ERROR cErrorIllegalFunctionCall
        END IF
        IF lastCol > 80 THEN
            LogError "DIALOG.BAS", "CpListBox", "lastCol > 80"
            ERROR cErrorIllegalFunctionCall
        END IF
     
        ' Bump the control count for the current window and copy
        ' it to the tab index element for this list box.
        sWindowData(sCurrentWindow).controlCount = sWindowData(sCurrentWindow).controlCount + 1
        lst.internal.tabindex = sWindowData(sCurrentWindow).controlCount

        ' Copy the data for this list box to the shared array.
        IF lst.fSearch AND lst.itemCount = 0 THEN
            sControlData(sCurrentWindow, lst.internal.tabindex).controlType = cCtListBoxWithSearch
        ELSE
            sControlData(sCurrentWindow, lst.internal.tabindex).controlType = cCtListBox
        END IF
        sControlData(sCurrentWindow, lst.internal.tabindex).row = lst.row
        sControlData(sCurrentWindow, lst.internal.tabindex).col = lst.col
        sControlData(sCurrentWindow, lst.internal.tabindex).lastRow = lastRow
        sControlData(sCurrentWindow, lst.internal.tabindex).lastCol = lastCol
        sControlData(sCurrentWindow, lst.internal.tabindex).fScrollBar = lst.internal.fScrollBar

        ' If there is no active control for the current window,
        ' then make this list box the active control.
        IF sWindowData(sCurrentWindow).activeControl = 0 THEN
            SetActiveControl lst.internal.tabindex, cAcThis
        END IF

        ' Save the initial item count to an instance variable so subsequent
        ' calls will be able to detect a change in the number of items.
        lst.internal.itemCount = itemCount
 
        ' Note that the following values are for the list item area.
        IF lst.internal.firstVisibleItem = 0 THEN
            lst.internal.firstVisibleItem = 1
        END IF
        lst.internal.firstRow = lst.row + 1
        lst.internal.lastRow = lastRow - 1
        lst.internal.firstCol = lst.col + 2
        lst.internal.lastCol = lastCol - 2
        lst.internal.rows = lst.rows - 2
        lst.internal.cols = lst.cols - 4
        IF lst.internal.fScrollBar THEN lst.internal.sliderRow = lst.internal.firstRow + 1
     
        ' Print everything.
        PrintFrame lst.row, lst.col, lst.rows, lst.cols, 1, cFalse, ""
        SetColor sWindowData(sCurrentWindow).fore, sWindowData(sCurrentWindow).back
        IF lst.internal.fScrollBar THEN GOSUB PrintScrollBarArrows
        GOSUB UpdateListAndScrollBar
        ShowMouseCursor
       
        lst.internal.fSelect = cFalse

    ELSE
        ' Because the fSearch parameter and the item count can change
        ' between calls, we need to set the control type on every call.
        IF lst.fSearch AND lst.itemCount = 0 THEN
            sControlData(sCurrentWindow, lst.internal.tabindex).controlType = cCtListBoxWithSearch
        ELSE
            sControlData(sCurrentWindow, lst.internal.tabindex).controlType = cCtListBox
        END IF

        ' If the selected item has changed since the last call...
        IF lst.selectedItem <> lst.internal.selectedItem THEN
       
            IF lst.selectedItem > itemCount THEN
                LogError "DIALOG.BAS", "CpListBox", "lst.selectedItem > itemCount"
                ERROR cErrorIllegalFunctionCall
            END IF
            IF lst.selectedItem < 1 THEN
                LogError "DIALOG.BAS", "CpListBox", "lst.selectedItem < 1"
                ERROR cErrorIllegalFunctionCall
            END IF

            ' If the selected item is above the first visible item,
            ' then set the first visible item to the selected item.
            IF lst.selectedItem < lst.internal.firstVisibleItem THEN
                lst.internal.firstVisibleItem = lst.selectedItem
            END IF

            ' If the selected item is below the last visible item,
            ' then set the first visible item such that the selected
            ' item will be on the last row.
            IF lst.selectedItem >= lst.internal.firstVisibleItem + lst.internal.rows THEN
                lst.internal.firstVisibleItem = lst.selectedItem - lst.internal.rows + 1

            END IF

            GOSUB UpdateListAndScrollBar
        END IF

        ' If the item count has changed since the last call
        ' or the fUpdate parameter is set...
        IF itemCount <> lst.internal.itemCount OR lst.fUpdate THEN
     
            ' If the last item is above the last row...
            IF itemCount < lst.internal.firstVisibleItem + lst.internal.rows - 1 THEN

                ' Adjust the first visible item as necessary to place the
                ' last item on or below the last row.
                lst.internal.firstVisibleItem = itemCount - lst.internal.rows + 1
         
                ' Adjust the first visible item as necessary to
                ' keep the value positive and non-zero.
                IF lst.internal.firstVisibleItem < 1 THEN lst.internal.firstVisibleItem = 1
        
                ' Adjust lst.selectedItem if its value is no longer valid.
                IF lst.selectedItem > itemCount THEN
                    lst.selectedItem = itemCount
                END IF
            END IF
           
            lst.internal.itemCount = itemCount
         
            SetColor sWindowData(sCurrentWindow).fore, sWindowData(sCurrentWindow).back
            HideTextCursor

            ' If this list box should have a scroll bar, then print the up
            ' and down arrows. Otherwise fill the entire scroll bar area
            ' with the normal frame characters (single line border assumed).
            IF lst.internal.fScrollBar THEN
                GOSUB PrintScrollBarArrows
            ELSE
                FOR row = lst.row + 1 TO lastRow - 1
                    HideMouseCursorIf row, lastCol, 1, 1
                    LOCATE row, lastCol
                    PRINT "";
                    ShowMouseCursor
                NEXT
            END IF
          
            GOSUB UpdateListAndScrollBar

            ' Clear the fUpdate parameter to avoid a repeat.
            lst.fUpdate = cFalse

        END IF

        GetUserInput key$, keyCode, mouseRow, mouseCol, fMouseLeft, 0, lst.internal.tabindex

        ' Handle any mouse input. The lst.internal.fMouseLeft flag stores
        ' the left mouse button state between calls to this instance.

        IF fMouseLeft THEN

            ' This block of code performs an auto-repeat
            ' of the up, down, page up, and page down
            ' scroll bar operations. It is placed here so
            ' it can execute while the mouse is locked.
            IF lst.internal.fScrollBar AND mouseCol = lst.col + lst.cols - 1 AND lst.internal.fMouseLeft THEN
                SELECT CASE mouseRow
                    CASE lst.internal.firstRow
                        ' The mouse cursor is on the up arrow.
                        IF NOT lst.internal.fMouseLeftOnSlider THEN
                            IF TIMER > lst.internal.timeOut THEN
                                GOSUB Up
                                lst.internal.timeOut = TIMER
                            END IF
                        END IF

                    CASE lst.internal.lastRow
                        ' The mouse cursor is on the down arrow.
                        IF NOT lst.internal.fMouseLeftOnSlider THEN
                            IF TIMER > lst.internal.timeOut THEN
                                GOSUB Down
                                lst.internal.timeOut = TIMER
                            END IF
                        END IF

                    CASE lst.internal.firstRow + 1 TO lst.internal.sliderRow - 1
                        ' The mouse cursor is on the slider area
                        ' of the scroll bar above the slider.
                        IF fSlider AND NOT lst.internal.fMouseLeftOnSlider THEN
                            IF TIMER > lst.internal.timeOut THEN
                                GOSUB PageUp
                                lst.internal.timeOut = TIMER
                            END IF
                        END IF

                    CASE lst.internal.sliderRow + 1 TO lst.internal.lastRow - 1
                        ' The mouse cursor is on the slider area
                        ' of the scroll bar below the slider.
                        IF fSlider AND NOT lst.internal.fMouseLeftOnSlider THEN
                            IF TIMER > lst.internal.timeOut THEN
                                GOSUB PageDown
                                lst.internal.timeOut = TIMER
                            END IF
                        END IF
                END SELECT
            END IF
           
            ' The block of code allows the user to drag the slider.
            ' It is placed here so it can execute while the mouse
            ' is locked. Note that the code ingores the mouse cursor
            ' column so the user can wander off of the scroll bar
            ' while dragging the slider.
            IF lst.internal.fMouseLeftOnSlider THEN
                SELECT CASE mouseRow
                    CASE IS <= lst.internal.firstRow
                        ' If the mouse cursor is above the slider area
                        ' of the scroll bar, then move to the start of
                        ' the list. This feature provides the mouse
                        ' equivalent of the Home key.
                        IF lst.internal.firstVisibleItem <> 1 OR lst.selectedItem <> 1 OR lst.internal.sliderRow <> lst.internal.firstRow + 1 THEN
                            lst.internal.firstVisibleItem = 1
                            lst.selectedItem = 1
                            lst.internal.sliderRow = lst.internal.firstRow + 1
                            GOSUB UpdateListAndScrollBar
                        END IF
                  
                    CASE lst.internal.firstRow + 1 TO lst.internal.lastRow - 1
                        ' If the mouse cursor row falls within the slider
                        ' area of the scroll bar and the mouse cursor is
                        ' not on the slider row, then the user is
                        ' attempting to drag the slider. Set the slider
                        ' row to the mouse cursor row and calculate new
                        ' values for the first visible item and the
                        ' selected item.
                        IF mouseRow <> lst.internal.sliderRow THEN
                            lst.internal.sliderRow = mouseRow
                            IF lst.internal.sliderRow = lst.internal.firstRow + 1 THEN
                                lst.internal.firstVisibleItem = 1
                            ELSE
                                ' The long integer constant used below
                                ' prevents an overflow by promoting the
                                ' intermediate result.
                                lst.internal.firstVisibleItem = ((lst.internal.sliderRow - lst.internal.firstRow - 1&) * (itemCount - lst.internal.rows + 1) / (lst.internal.rows - 3))

                                ' The statement above maintains the
                                ' correct relationship between the
                                ' position of the slider on the scroll
                                ' bar and the position of the first
                                ' visible item in the item list.
                                ' Because lst.internal.rows is an
                                ' integer, some combinations of values
                                ' for itemCount and lst.internal.rows
                                ' will cause the statement to assign a
                                ' value of 0. The statement below
                                ' corrects this problem.
                                IF lst.internal.firstVisibleItem < 1 THEN lst.internal.firstVisibleItem = 1
                            END IF
                            IF lst.selectedItem < lst.internal.firstVisibleItem THEN
                                lst.selectedItem = lst.internal.firstVisibleItem
                            ELSEIF lst.selectedItem > lst.internal.firstVisibleItem + lst.internal.rows - 1 THEN
                                lst.selectedItem = lst.internal.firstVisibleItem + lst.internal.rows - 1
                            END IF
                            GOSUB UpdateListAndScrollBar
                        END IF

                    CASE IS >= lst.internal.lastRow
                        ' If the mouse cursor is below the slider area
                        ' of the scroll bar, then move to the end of
                        ' the list. This feature provides the mouse
                        ' equivalent of the End key.
                        IF lst.internal.firstVisibleItem <> itemCount - lst.internal.rows + 1 OR lst.selectedItem <> itemCount OR lst.internal.sliderRow <> lst.internal.lastRow - 1 THEN
                            lst.internal.firstVisibleItem = itemCount - lst.internal.rows + 1
                            lst.selectedItem = itemCount
                            lst.internal.sliderRow = lst.internal.lastRow - 1
                            GOSUB UpdateListAndScrollBar
                        END IF
                END SELECT
            END IF

            ' The NOT condition prevents the scroll bar from
            ' responding to a mouse-over when the left mouse
            ' button is pressed but was not initially pressed
            ' over this list box.
            IF NOT sfMouseLock THEN

                ' If the mouse cursor is on the list field...
                IF mouseRow >= lst.internal.firstRow AND mouseRow <= lst.internal.lastRow AND mouseCol >= lst.internal.firstCol AND mouseCol <= lst.internal.lastCol THEN

                    ' If no control has responded, then make this list box
                    ' the active control, set the selected item to match
                    ' the mouse cursor location (if the mouse cursor is on
                    ' a list item), set the necessary flags, and update
                    ' the item list and the scroll bar.
                    IF NOT lst.internal.fMouseLeft AND NOT sfMouseLock THEN
                        SetActiveControl lst.internal.tabindex, cAcThis
                        cursorItem = lst.internal.firstVisibleItem + mouseRow - lst.internal.firstRow
                        IF cursorItem <= itemCount THEN lst.selectedItem = cursorItem
                        lst.internal.fMouseLeft = cTrue

                        ' The NOT condition fixes the problem with a
                        ' drop-down list box item being selected and
                        ' the list closed when the user wanders into
                        ' the list area while dragging the slider.
                        IF NOT lst.internal.fMouseLeftOnSlider THEN
                            lst.internal.fSelect = cTrue
                        END IF

                        sfMouseLock = cTrue
                        GOSUB UpdateListAndScrollBar

                        ' If double-clicking is enabled, and the previous
                        ' block set a new selected item, and the selected
                        ' item is the same as it was the last time the left
                        ' mouse button was released on a list item, and the
                        ' double-click interval has not expired, then place
                        ' the character code and scan code for the Enter
                        ' key in the keyboard buffer. This will activate
                        ' the current window's default command button.
                        IF lst.fDoubleClick AND cursorItem THEN
                            IF lst.selectedItem = lst.internal.prevSelectedItem THEN
                                IF TIMER < lst.internal.doubleClickTimeout THEN

                                    ' Ignore any error return.
                                    PushKey &H1C, 13, 0
                                END IF
                            END IF
                        END IF
                    END IF

                ' If this list box has a scroll bar and the mouse
                ' cursor is on the scroll bar...
                ELSEIF lst.internal.fScrollBar AND mouseCol = lst.col + lst.cols - 1 AND mouseRow >= lst.internal.firstRow AND mouseRow <= lst.internal.lastRow THEN

                    sfMouseLock = cTrue
                  
                    ' Tis block of code performs the initial up, down,
                    ' page up, and page down operations, sets an auto-
                    ' repeat timeout value, and enables dragging of the
                    ' slider. For the cases that auto-repeat, the
                    ' timeout value is set to the value in the constant
                    ' cRepeatDelay and the repeat interval is effectively
                    ' 55 ms (18.2 repeats per second). The value returned
                    ' by the TIMER function is based on the system timer
                    ' tick. Because the system timer tick has a period
                    ' of 55 ms, the value returned by the TIMER function
                    ' return value changes in increments of 55 ms.
                    SELECT CASE mouseRow
                        CASE lst.internal.firstRow
                            ' The mouse cursor is on the up arrow.
                            IF NOT lst.internal.fMouseLeft AND NOT lst.internal.fMouseLeftOnSlider THEN
                                GOSUB Up
                                lst.internal.fMouseLeft = cTrue
                                lst.internal.timeOut = TIMER + cRepeatDelay
                            END IF

                        CASE lst.internal.lastRow
                            ' The mouse cursor is on the down arrow.
                            IF NOT lst.internal.fMouseLeft AND NOT lst.internal.fMouseLeftOnSlider THEN
                                GOSUB Down
                                lst.internal.fMouseLeft = cTrue
                                lst.internal.timeOut = TIMER + cRepeatDelay
                            END IF

                        CASE lst.internal.firstRow + 1 TO lst.internal.sliderRow - 1
                            ' The mouse cursor is on the slider area
                            ' of the scroll bar above the slider.
                            IF fSlider AND NOT lst.internal.fMouseLeft AND NOT lst.internal.fMouseLeftOnSlider THEN
                                GOSUB PageUp
                                lst.internal.fMouseLeft = cTrue
                                lst.internal.timeOut = TIMER + cRepeatDelay
                            END IF

                        CASE lst.internal.sliderRow + 1 TO lst.internal.lastRow - 1
                            ' The mouse cursor is on the slider area
                            ' of the scroll bar below the slider.
                            IF fSlider AND NOT lst.internal.fMouseLeft AND NOT lst.internal.fMouseLeftOnSlider THEN
                                GOSUB PageDown
                                lst.internal.fMouseLeft = cTrue
                                lst.internal.timeOut = TIMER + cRepeatDelay
                            END IF

                        CASE lst.internal.sliderRow
                            ' The mouse cursor is on the slider.
                            IF fSlider AND NOT lst.internal.fMouseLeft AND NOT lst.internal.fMouseLeftOnSlider THEN
                                lst.internal.fMouseLeftOnSlider = cTrue

                                ' This statement solves a problem
                                ' with dropping the slider when the
                                ' user wanders into the list area
                                ' while dragging the slider.
                                lst.internal.fMouseLeft = cTrue
                            END IF
                      
                    END SELECT
                END IF
            END IF
        ELSE
            ' The left mouse button has been released. If double-clicking
            ' is enabled, and the left mouse button was pressed with the
            ' mouse cursor on the control on the previous call to this
            ' instance, then save the current selected item and calculate
            ' and save a double-click timeout value.
            IF lst.fDoubleClick AND lst.internal.fMouseLeft THEN
                lst.internal.prevSelectedItem = lst.selectedItem
                lst.internal.doubleClickTimeout = TIMER + cDoubleClickInterval
            END IF

            ' Clear the flags.
            lst.internal.fMouseLeft = cFalse
            lst.internal.fMouseLeftOnSlider = cFalse
            sfMouseLock = cFalse

        END IF

        ' If this list box is the active control for the current window,
        ' then handle any keys. Some cases include logic to prevent
        ' redundant operations, including redundant updates of the list
        ' item field. Set the shared discard key flag for any key that
        ' is handled.
        IF sWindowData(sCurrentWindow).activeControl = lst.internal.tabindex THEN
            SELECT CASE keyCode
        
                CASE cKeyTab
                    SetActiveControl lst.internal.tabindex, cAcNext
                    sfDiscardKey = cTrue
                    GOSUB UpdateListAndScrollBar
                  
                CASE cKeyShiftTab
                    SetActiveControl lst.internal.tabindex, cAcPrev
                    sfDiscardKey = cTrue
                    GOSUB UpdateListAndScrollBar
                     
                CASE cKeyHome
                    IF lst.internal.firstVisibleItem > 1 OR lst.selectedItem > 1 THEN
                        lst.internal.firstVisibleItem = 1
                        lst.selectedItem = 1
                        sfDiscardKey = cTrue
                        GOSUB UpdateListAndScrollBar
                    END IF

                CASE cKeyEnd
                    IF lst.internal.firstVisibleItem < itemCount - lst.internal.rows + 1 OR lst.selectedItem < itemCount THEN
                        IF itemCount > lst.internal.rows THEN
                            lst.internal.firstVisibleItem = itemCount - lst.internal.rows + 1
                        END IF
                        lst.selectedItem = itemCount
                        sfDiscardKey = cTrue
                        GOSUB UpdateListAndScrollBar
                    END IF
            
                CASE cKeyInsert
                    sfTypeOver = NOT sfTypeOver
                    sfDiscardKey = cTrue

                CASE cKeyUp
                    GOSUB Up
                    sfDiscardKey = cTrue

                CASE cKeyDown
                    GOSUB Down
                    sfDiscardKey = cTrue

                CASE cKeyLeft
                    SetActiveControl lst.internal.tabindex, cAcLeft
                    sfDiscardKey = cTrue

                CASE cKeyRight
                    SetActiveControl lst.internal.tabindex, cAcRight
                    sfDiscardKey = cTrue

                CASE cKeyPageUp
                    GOSUB PageUp
                    sfDiscardKey = cTrue

                CASE cKeyPageDown
                    GOSUB PageDown
                    sfDiscardKey = cTrue

                CASE 32 TO 255
                    ' The key is a character key (or an ASCII character
                    ' code entered as Alt+n). If the search feature is
                    ' enabled and the list is not an owner draw list,
                    ' then, starting at the item following the selected
                    ' item, search for the next item where the first
                    ' character matches the key character (case-
                    ' insensitive). If a match is found, then set the
                    ' selected item to the match and adjust
                    ' firstVisibleItem.
                    IF lst.fSearch AND lst.itemCount = 0 THEN
                        key$ = UCASE$(key$)
                        i = lst.selectedItem + 1
                        IF i > itemCount THEN i = 1
                        FOR j = 1 TO itemCount
                            IF key$ = UCASE$(LEFT$(list$(i), 1)) THEN
                                lst.selectedItem = i
                               
                                ' If the selected item is above the first
                                ' visible item, then set the first visible
                                ' item to the selected item.
                                IF lst.selectedItem < lst.internal.firstVisibleItem THEN
                                    lst.internal.firstVisibleItem = lst.selectedItem
                                END IF

                                ' If the selected item is below the last
                                ' visible item, then set the first visible
                                ' item so the selected item will be on the
                                ' last row.
                                IF lst.selectedItem >= lst.internal.firstVisibleItem + lst.internal.rows THEN
                                    lst.internal.firstVisibleItem = lst.selectedItem - lst.internal.rows + 1
                                END IF

                                GOSUB UpdateListAndScrollBar
                                sfDiscardKey = cTrue
                                EXIT FOR
                            END IF
                            i = i + 1
                            IF i > itemCount THEN i = 1
                        NEXT
                    END IF
            END SELECT
        END IF

        ' If this list box is the active control for the current window, then
        ' position the text cursor and make it visible. If no selected item
        ' has been specified, then position the text cursor at the first row.
        IF sWindowData(sCurrentWindow).activeControl = lst.internal.tabindex THEN
            IF lst.selectedItem = 0 THEN
                LOCATE lst.internal.firstRow, lst.internal.firstCol
            ELSE
                LOCATE lst.row + lst.selectedItem - lst.internal.firstVisibleItem + 1, lst.internal.firstCol
            END IF
            ShowTextCursor
        END IF
     
        ' Save the number of the selected item to an instance variable so
        ' subsequent calls will be able to detect a change in the selected
        ' item.
        lst.internal.selectedItem = lst.selectedItem

    END IF

EXIT SUB

PrintScrollBarArrows:
    ' If this list box should have scroll bar, then print the scroll bar
    ' arrows. Printing the arrows here, instead of in the subroutine that
    ' updates the list item field and the scroll bar, avoids multiple hides
    ' of the mouse cursor when the mouse is used to scroll the list.
    IF lst.internal.fScrollBar THEN
        HideMouseCursorIf lst.internal.firstRow, lastCol, 1, 1
        LOCATE lst.internal.firstRow, lastCol
        PRINT CHR$(24);
        ShowMouseCursor
        HideMouseCursorIf lst.internal.lastRow, lastCol, 1, 1
        LOCATE lst.internal.lastRow, lastCol
        PRINT CHR$(25);
        ShowMouseCursor
    END IF

    RETURN

Up:
    ' Move the selected item up 1 row, adjusting firstVisibleItem
    ' as necessary.
    IF lst.selectedItem = 0 THEN
        lst.selectedItem = 1
        GOSUB UpdateListAndScrollBar
    ELSEIF lst.selectedItem > lst.internal.firstVisibleItem THEN
        lst.selectedItem = lst.selectedItem - 1
        GOSUB UpdateListAndScrollBar
    ELSEIF lst.internal.firstVisibleItem > 1 THEN
        lst.internal.firstVisibleItem = lst.internal.firstVisibleItem - 1
        lst.selectedItem = lst.selectedItem - 1
        GOSUB UpdateListAndScrollBar
    END IF

    RETURN


Down:
    ' Move the selected item down 1 row, adjusting firstVisibleItem
    ' as necessary.
    IF lst.selectedItem < lst.internal.firstVisibleItem + lst.internal.rows - 1 AND lst.selectedItem < itemCount THEN
        lst.selectedItem = lst.selectedItem + 1
        GOSUB UpdateListAndScrollBar
    ELSEIF lst.internal.firstVisibleItem < itemCount - lst.internal.rows + 1 THEN
        lst.internal.firstVisibleItem = lst.internal.firstVisibleItem + 1
        lst.selectedItem = lst.selectedItem + 1
        GOSUB UpdateListAndScrollBar
    END IF

    RETURN

PageUp:
    ' Scroll the list down such that the first visible item will be on the
    ' last row and move the highlight to the first row.
    IF lst.internal.firstVisibleItem > 1 OR lst.selectedItem > 1 THEN
        lst.internal.firstVisibleItem = lst.internal.firstVisibleItem - lst.internal.rows + 1
        IF lst.internal.firstVisibleItem < 1 THEN lst.internal.firstVisibleItem = 1
        lst.selectedItem = lst.internal.firstVisibleItem
        GOSUB UpdateListAndScrollBar
    END IF
    RETURN

PageDown:
    ' Scroll the list up such that the last visible item will be on the
    ' first row and move the highlight to the last row.
    IF lst.internal.rows < itemCount THEN
        IF lst.internal.firstVisibleItem < itemCount - lst.internal.rows + 1 OR lst.selectedItem < itemCount THEN
            lst.internal.firstVisibleItem = lst.internal.firstVisibleItem + lst.internal.rows - 1
            IF lst.internal.firstVisibleItem > itemCount - lst.internal.rows + 1 THEN
                lst.internal.firstVisibleItem = itemCount - lst.internal.rows + 1
            END IF
            lst.selectedItem = lst.internal.firstVisibleItem + lst.internal.rows - 1
            GOSUB UpdateListAndScrollBar
        END IF
    ELSE
        lst.selectedItem = itemCount
        GOSUB UpdateListAndScrollBar
    END IF
    RETURN

UpdateListAndScrollBar:

    IF lst.itemCount = 0 THEN
        HideTextCursor
        FOR i = 1 TO lst.internal.rows
            item = i + lst.internal.firstVisibleItem - 1
            IF item = lst.selectedItem AND NOT lst.fText THEN
                SetColor cHighlightFore, cHighlightBack
            ELSE
                SetColor sWindowData(sCurrentWindow).fore, sWindowData(sCurrentWindow).back
            END IF

            ' Truncate each item and fill any empty rows at
            ' the end with spaces.
            IF item <= itemCount THEN
                temp$ = LEFT$(list$(item), lst.internal.cols)
            ELSE
                temp$ = SPACE$(lst.internal.cols)
            END IF
      
            HideMouseCursorIf lst.row + i, lst.internal.firstCol, 1, lst.internal.cols
            LOCATE lst.row + i, lst.internal.firstCol
            PRINT temp$;

            ' If the cursor is not at the last column, then
            ' fill the remainder of the row with spaces to
            ' overwrite any old characters.
            IF item <= itemCount THEN
                cursorColTemp = POS(0)
                IF cursorColTemp <= lst.internal.lastCol THEN
                    PRINT SPACE$(lst.internal.lastCol - cursorColTemp + 1);
                END IF
            END IF
            ShowMouseCursor
        NEXT
    END IF
  
    SetColor sWindowData(sCurrentWindow).fore, sWindowData(sCurrentWindow).back
 
    ' If this list box should have a scroll bar...
    IF lst.internal.fScrollBar THEN
 
        ' If the scroll bar contains a slider and the user is not attempting
        ' to drag the slider, then calculate the slider row.
        IF fSlider AND NOT lst.internal.fMouseLeftOnSlider THEN
            IF lst.internal.firstVisibleItem = 1 THEN
                lst.internal.sliderRow = lst.internal.firstRow + 1
            ELSE

                ' The following statement is a transposed version of the
                ' statement used to calculate the first visible item in
                ' the code that handles mouse inputs. Because the variables
                ' are integers instead of reals, this statement may assign
                ' a value that is different from the value that was used
                ' to calculate the first visible item. A simple fix for
                ' this problem is to skip these calculations whenever the
                ' the user is attempting to drag the slider.
                lst.internal.sliderRow = (lst.internal.rows - 3) * (lst.internal.firstVisibleItem / (itemCount - lst.internal.rows + 1)) + lst.internal.firstRow + 1

                ' The statement above maintains the correct relationship
                ' between the position of the slider on the scroll bar and
                ' the position of the first visible item in the item list.
                ' Depending on the number of list item rows and the number
                ' list items, the calculated slider row may cause problems
                ' with the scroll bar page up and page down functions.
                ' The statements below adjust the slider row as necessary
                ' to prevent these problems.
                IF lst.internal.sliderRow = lst.internal.firstRow + 1 THEN
                    IF lst.internal.firstVisibleItem > 1 THEN
                        lst.internal.sliderRow = lst.internal.sliderRow + 1
                    END IF
                END IF
                IF lst.internal.sliderRow = lst.internal.lastRow - 1 THEN
                    IF lst.internal.firstVisibleItem < itemCount - lst.internal.rows + 1 THEN
                        lst.internal.sliderRow = lst.internal.sliderRow - 1
                    END IF
                END IF
            END IF
        END IF
      
        HideMouseCursorIf lst.internal.firstRow + 1, lastCol, lst.internal.rows - 2, 1
        FOR row = lst.internal.firstRow + 1 TO lst.internal.lastRow - 1
            LOCATE row, lastCol
            IF row = lst.internal.sliderRow THEN
                IF fSlider THEN PRINT CHR$(219);  ELSE PRINT CHR$(176);
            ELSE
                PRINT CHR$(176);
            END IF
        NEXT
    END IF
    ShowMouseCursor
  
    RETURN

END SUB

SUB CpOptionButton (opt AS OptionButtonType, caption$)

    ' Displays and operates an option button control.

    ' If this is the first call, then init and return.
    IF opt.internal.tabindex = 0 THEN
        
        IF sCurrentWindow < 1 THEN
            LogError "DIALOG.BAS", "CpOptionButton", "sCurrentWindow < 1"
            ERROR cErrorIllegalFunctionCall
        END IF
        IF caption$ = "" THEN
            LogError "DIALOG.BAS", "CpOptionButton", "caption$ = <null>"
            ERROR cErrorIllegalFunctionCall
        END IF

        ' Get the position of the access key character and remove
        ' the delimiter from the caption.
        accessPosition = INSTR(caption$, "&")
        IF accessPosition THEN
            cap$ = LEFT$(caption$, accessPosition - 1) + MID$(caption$, accessPosition + 1)
        ELSE
            cap$ = caption$
        END IF

        opt.internal.lastCol = opt.col + LEN(cap$) + 3

        IF opt.row < 1 THEN
            LogError "DIALOG.BAS", "CpOptionButton", "opt.row < 1"
            ERROR cErrorIllegalFunctionCall
        END IF
        IF opt.row > 25 THEN
            LogError "DIALOG.BAS", "CpOptionButton", "opt.row > 25"
            ERROR cErrorIllegalFunctionCall
        END IF
        IF opt.col < 1 THEN
            LogError "DIALOG.BAS", "CpOptionButton", "opt.col < 1"
            ERROR cErrorIllegalFunctionCall
        END IF
        IF opt.internal.lastCol > 80 THEN
            LogError "DIALOG.BAS", "CpOptionButton", "opt.internal.lastCol > 80"
            ERROR cErrorIllegalFunctionCall
        END IF

        ' Save the access key character and the Alt+ access key code.
        IF accessPosition THEN
            opt.internal.accessKey = UCASE$(MID$(cap$, accessPosition, 1))
            opt.internal.accessCode = AltPlusKeyCode(MID$(cap$, accessPosition, 1))
        END IF

        ' Bump the control count for the current window and copy
        ' it to the tab index element for this option button.
        sWindowData(sCurrentWindow).controlCount = sWindowData(sCurrentWindow).controlCount + 1
        opt.internal.tabindex = sWindowData(sCurrentWindow).controlCount

        ' Copy the data for this option button to the shared array.
        sControlData(sCurrentWindow, opt.internal.tabindex).controlType = cCtOptionButton
        sControlData(sCurrentWindow, opt.internal.tabindex).row = opt.row
        sControlData(sCurrentWindow, opt.internal.tabindex).col = opt.col
        sControlData(sCurrentWindow, opt.internal.tabindex).lastRow = opt.row
        sControlData(sCurrentWindow, opt.internal.tabindex).lastCol = opt.internal.lastCol
       
        ' Copy the group number and the initial selected state for this
        ' option button to the shared array.
        sControlData(sCurrentWindow, opt.internal.tabindex).group = opt.group
        sControlData(sCurrentWindow, opt.internal.tabindex).fSelected = opt.fSelected
       
        ' Copy the initial selected state to an instance variable.
        opt.internal.fSelected = opt.fSelected

        ' Add the option box to the caption.
        IF opt.fSelected THEN
            cap$ = "(*) " + cap$
        ELSE
            cap$ = "( ) " + cap$
        END IF

        ' Print the caption.
        fore = sWindowData(sCurrentWindow).fore
        back = sWindowData(sCurrentWindow).back
        IF accessPosition THEN
            accessCol = opt.col + accessPosition + 3
        END IF
        PrintControlCap opt.row, opt.col, fore, back, cAccessFore, accessCol, cap$

    ELSE
        GetUserInput key$, keyCode, mouseRow, mouseCol, fMouseLeft, 0, opt.internal.tabindex

        ' Handle any mouse input. The opt.internal.fMouseLeft flag stores
        ' the left mouse button state between calls to this instance.

        IF fMouseLeft THEN

            ' If the mouse cursor is on this option button...
            IF mouseRow = opt.row AND mouseCol >= opt.col AND mouseCol <= opt.internal.lastCol THEN

                ' If no control has responded, then make this option button
                ' the active control and set the necessary flags.
                IF NOT opt.internal.fMouseLeft AND NOT sfMouseLock THEN
                    SetActiveControl opt.internal.tabindex, cAcThis
                    opt.internal.fMouseLeft = cTrue
                    sfMouseLock = cTrue
                END IF
            END IF

        ELSEIF opt.internal.fMouseLeft THEN

            ' The left mouse button has been released so clear the flags.
            opt.internal.fMouseLeft = cFalse
            sfMouseLock = cFalse

            ' If the mouse cursor is still on this option button,
            ' then set the selected flag.
            IF mouseRow = opt.row AND mouseCol >= opt.col AND mouseCol <= opt.internal.lastCol THEN
                opt.fSelected = cTrue
            END IF

        END IF

        ' Handle any keyboard input.  Set the shared discard key flag
        ' for any key handled.

        ' If this check box has an access key...
        IF opt.internal.accessCode THEN

            ' If the user pressed the access key without the Alt key...
            IF UCASE$(key$) = opt.internal.accessKey THEN

                ' If the active control is not a text box or a
                ' list box with the internal item search feature
                ' enabled, then make this option button the
                ' active control and set the selected flag.
                act = ActiveControlType
                IF act <> cCtTextBox AND act <> cCtListBoxWithSearch AND act <> cCtDropDownListBoxWithSearch THEN
                    SetActiveControl opt.internal.tabindex, cAcThis
                    opt.fSelected = cTrue
                    sfDiscardKey = cTrue
                END IF

            ELSEIF keyCode = opt.internal.accessCode THEN
                ' If the user pressed the access key in combination with
                ' the Alt key, then make this option button the active
                ' control and set the selected flag.
                SetActiveControl opt.internal.tabindex, cAcThis
                opt.fSelected = cTrue
                sfDiscardKey = cTrue
            END IF
        END IF

        ' Handle any other keys only if this option button is the active
        ' control.
        IF sWindowData(sCurrentWindow).activeControl = opt.internal.tabindex THEN
           
            SELECT CASE keyCode
                CASE cKeyTab
                    SetActiveControl opt.internal.tabindex, cAcNext
                    sfDiscardKey = cTrue
         
                CASE cKeyShiftTab
                    SetActiveControl opt.internal.tabindex, cAcPrev
                    sfDiscardKey = cTrue

                CASE cKeyUp
                    SetActiveControl opt.internal.tabindex, cAcUp
                    sfDiscardKey = cTrue
         
                CASE cKeyDown
                    SetActiveControl opt.internal.tabindex, cAcDown
                    sfDiscardKey = cTrue

                CASE cKeyLeft
                    SetActiveControl opt.internal.tabindex, cAcLeft
                    sfDiscardKey = cTrue

                CASE cKeyRight
                    SetActiveControl opt.internal.tabindex, cAcRight
                    sfDiscardKey = cTrue
              
                CASE cKeyInsert
                    sfTypeOver = NOT sfTypeOver
                    sfDiscardKey = cTrue

                CASE cKeySpace
                    IF NOT opt.fSelected THEN
                        opt.fSelected = cTrue
                    END IF
                    sfDiscardKey = cTrue

            END SELECT
        END IF
    END IF
  
    ' External code can set or clear opt.fSelected. Other instances of
    ' this procedure can clear the selected flag in the shared control
    ' data.
    '
    ' If opt.fSelected does not match opt.internal.fSelected (because
    ' opt.fSelected has been changed by external code or by the code
    ' above), then set the visual state and the other flags to match
    ' opt.fSelected. Otherwise, if the selected flag in the shared
    ' control data does not match opt.internal.fSelected (because the
    ' selected flag in the shared control data has been cleared by
    ' another instance), then set the visual state and the other flags
    ' to match the selected flag in the shared control data.
    IF opt.fSelected <> opt.internal.fSelected THEN
        IF opt.fSelected THEN
            GOSUB DeselectGroup
            GOSUB PrintSelected
            opt.internal.fSelected = cTrue
            sControlData(sCurrentWindow, opt.internal.tabindex).fSelected = cTrue
        ELSE
            GOSUB PrintNotSelected
            opt.internal.fSelected = cFalse
            sControlData(sCurrentWindow, opt.internal.tabindex).fSelected = cFalse
        END IF
    ELSEIF sControlData(sCurrentWindow, opt.internal.tabindex).fSelected <> opt.internal.fSelected THEN
        IF sControlData(sCurrentWindow, opt.internal.tabindex).fSelected THEN
            GOSUB DeselectGroup
            GOSUB PrintSelected
            opt.fSelected = cTrue
            opt.internal.fSelected = cTrue
        ELSE
            GOSUB PrintNotSelected
            opt.fSelected = cFalse
            opt.internal.fSelected = cFalse
        END IF
    END IF

    ' If this option button is the active control for the current
    ' window, then position the text cursor and make it visible.
    IF sWindowData(sCurrentWindow).activeControl = opt.internal.tabindex THEN
        LOCATE opt.row, opt.col + 1
        ShowTextCursor
    END IF

EXIT SUB

DeselectGroup:
    ' Deselect all option buttons in the same group as this option button.
    FOR j = 1 TO sWindowData(sCurrentWindow).controlCount
        IF sControlData(sCurrentWindow, j).controlType = cCtOptionButton AND sControlData(sCurrentWindow, j).group = opt.group THEN
            IF sControlData(sCurrentWindow, j).fSelected THEN
                sControlData(sCurrentWindow, j).fSelected = cFalse
            END IF
        END IF
    NEXT
    RETURN

PrintSelected:
    ' Set the visual state of this option button to selected.
    SetColor sWindowData(sCurrentWindow).fore, sWindowData(sCurrentWindow).back
    HideTextCursor
    LOCATE opt.row, opt.col + 1
    HideMouseCursorIf opt.row, opt.col + 1, 1, 1
    PRINT "*";
    ShowMouseCursor
    RETURN

PrintNotSelected:
    ' Set the visual state of this option button to not selected.
    SetColor sWindowData(sCurrentWindow).fore, sWindowData(sCurrentWindow).back
    HideTextCursor
    LOCATE opt.row, opt.col + 1
    HideMouseCursorIf opt.row, opt.col + 1, 1, 1
    PRINT " ";
    ShowMouseCursor
    RETURN

END SUB

SUB EnableHiBackColors

    ' Enables high-intensity background colors.
    '
    ' This procedure calls the VGA BIOS Toggle Blink/Intensity Bit
    ' function. That function sets or clears the blink/intensity bit
    ' in the Attribute Controller's Mode Control Register. When the
    ' bit is clear, the Attribute Controller uses the blink/intensity
    ' bit of the character attribute to control the background color
    ' intensity.
    '
    ' This procedure disables blinking foreground colors. Note that
    ' the QB programming environment will re-enable blinking whenever
    ' it regains control (after breakpoints, etc). This will cause
    ' all characters with a high-intensity background to revert to a
    ' blinking foreground.
    '
    ' Note that the COLOR statement cannot be used in the normal
    ' manner to select a high-intensity background color (even with
    ' high-intensity background colors enabled). You should use the
    ' SetColor procedure in place of the COLOR statement.

    DIM regX AS RegTypeX

    ' For the value in BX, 0 = enable intensity, 1 = enable blinking.
    regX.ax = &H1003
    regX.bx = 0
    InterruptX &H10, regX, regX

END SUB

SUB GetUserInput (key$, keyCode%, mouseRow%, mouseCol%, fMouseLeft%, fMouseRight%, callerIndex%) STATIC

    ' Returns information about keyboard and mouse input.
    '
    ' The control procedures call this procedure with the tab index
    ' of the current control instance in <callerIndex%>. The mouse
    ' status is read on each call. The keyboard is read whenever the
    ' shared discard key flag is set (indicating that the previous
    ' caller processed the current key or the current window changed)
    ' and whenever the value in <callerIndex%> repeats (indicating
    ' that all controls have ignored the current key).
    '
    ' If the shared flag sfEnableToggle is set, then this procedure
    ' traps the F4 key to toggle the display between the screen page
    ' of the current window and screen page 0. While screen page 0
    ' is being displayed, this procedure ignores all input events.
    ' Note that the Escape key can be used to toggle to the screen
    ' page of the current window, but not to screen page 0.
   
    IF sfEnableToggle THEN

        ' If currently displaying screen page 0, then
        ' read the keyboard.
        IF fPage0 THEN

            GetKey k$, kCode

            ' If the user pressed the F4 or Escape key, then
            ' toggle to the screen page of the current window.
            IF kCode = cKeyF4 OR kCode = cKeyEscape THEN
                HideMouseCursor
                SCREEN , , sCurrentWindow, sCurrentWindow
                ShowMouseCursor
                fPage0 = cFalse
            ELSE

                ' Otherwise, return to the caller
                ' ignoring all input events.
                EXIT SUB
            END IF
        END IF
    END IF
   
    GetMouseStatus mouseRow, mouseCol, fMouseLeft, 0
   
    ' We need to lock the mouse if the left mouse button is
    ' pressed while the mouse cursor is NOT on a control.
    ' This feature is used by the control procedures to keep
    ' the controls from responding to a mouse-over when the
    ' left mouse button is pressed but was not initially
    ' pressed over a control (in the area for which the
    ' control procedure locks the mouse).
    IF fMouseLeft THEN
        IF NOT fMouseChecked THEN
            FOR i = 1 TO sWindowData(sCurrentWindow).controlCount
                IF sControlData(sCurrentWindow, i).controlType = cCtListBox OR sControlData(sCurrentWindow, i).controlType = cCtListBoxWithSearch THEN
                    IF mouseRow > sControlData(sCurrentWindow, i).row AND mouseRow < sControlData(sCurrentWindow, i).lastRow THEN
                        IF sControlData(sCurrentWindow, i).fScrollBar THEN
                            IF mouseCol > sControlData(sCurrentWindow, i).col + 1 AND mouseCol < sControlData(sCurrentWindow, i).lastCol - 1 OR mouseCol = sControlData(sCurrentWindow, i).lastCol THEN
                                fOnControl = cTrue
                                EXIT FOR
                            END IF
                        ELSE
                            IF mouseCol > sControlData(sCurrentWindow, i).col + 1 AND mouseCol < sControlData(sCurrentWindow, i).lastCol - 1 THEN
                                fOnControl = cTrue
                                EXIT FOR
                            END IF
                        END IF
                    END IF
                ELSE
                    IF mouseRow >= sControlData(sCurrentWindow, i).row AND mouseRow <= sControlData(sCurrentWindow, i).lastRow THEN
                        IF mouseCol >= sControlData(sCurrentWindow, i).col AND mouseCol <= sControlData(sCurrentWindow, i).lastCol THEN
                            fOnControl = cTrue
                            EXIT FOR
                        END IF
                    END IF
                END IF
            NEXT
            IF NOT fOnControl THEN
                sfMouseLock = cTrue
                fLockedMouse = cTrue
            END IF
        END IF
        fMouseChecked = cTrue
    ELSE
        IF fLockedMouse THEN
            sfMouseLock = cFalse
            fLockedMouse = cFalse
        END IF
        fMouseChecked = cFalse
        fOnControl = cFalse
    END IF

    IF sfDiscardKey OR callerIndex = startIndex THEN
        
        ' Read the keystroke information into local variables.
        GetKey localKey$, localKeyCode

        ' If we got here, then we must be displaying
        ' the screen page of the current window.
        IF sfEnableToggle THEN

            ' If the user pressed the F4 key, then toggle to
            ' screen page 0, set the page 0 flag, and return
            ' to the caller.
            IF localKeyCode = cKeyF4 THEN
                HideMouseCursor
                SCREEN , , 0, 0
                ShowMouseCursor
                fPage0 = cTrue
                EXIT SUB
            END IF
        END IF

        ' Set variables so we can detect when we need
        ' to read the keyboard again.
        sfDiscardKey = cFalse
        startIndex = callerIndex
    END IF

    ' Return the values stored in the local variables.
    key$ = localKey$
    keyCode = localKeyCode

END SUB

SUB HideTextCursor

    LOCATE , , 0

END SUB

SUB InitModule

    ' Checks for a VGA compatible display adapter, sets the screen mode
    ' to 80x25 text, sets the active and visible screen page to page 0,
    ' enables high-intensity background colors, and sets a shared flag
    ' to indicate that all of this has been done.
    '
    ' Displays an error message and terminates the program if the
    ' display adapter is not VGA compatible.
    '
    ' Note that the procedures in this module will actually work with
    ' EGA (32KB minimum), MCGA, and several other display adapter types.
    ' However, since such adapters are now relatively rare, there is no
    ' real need to complicate the test.

    IF NOT IsVGA THEN
        PRINT
        PRINT "This program requires a VGA compatible display adapter."
        END
    END IF
  
    SCREEN 0
    WIDTH 80
    SCREEN , , 0, 0

    EnableHiBackColors
    
    sfInitModuleOK = cTrue

END SUB

FUNCTION KeyPressed (keyCode%)

    ' Returns true if the user pressed the key corresponding
    ' to <keyCode%>.
    '
    ' Other modules can use this procedure to trap keys by
    ' calling it from the main loop of a window procedure.
    '
    ' The value in <keyCode%> can be one of the key code
    ' constants defined in GENERAL.BI, or the ASCII code
    ' for a character key (case sensitive).
    '
    ' The value in <keyCode%> is used to generate a unique
    ' callerIndex argument for the GetUserInput procedure.
    ' The constant 1000 ensures that the argument will not
    ' match the index for any control. Without the unique
    ' value, this procedure could not be used to (reliably)
    ' trap more than a single key because the GetUserInput
    ' procedure discards the current key whenever the value
    ' in it's callerIndex argument repeats.
    '
    ' If <keyCode%> specifies a key that is trapped by any
    ' of the controls called from the window procedure that
    ' called this procedure, then one or more of the controls,
    ' or this procedure, may have problems trapping the key.
    
    GetUserInput key$, kCode, 0, 0, 0, 0, keyCode + 1000
   
    IF kCode = keyCode THEN
        KeyPressed = cTrue
        sfDiscardKey = cTrue
    END IF

END FUNCTION

FUNCTION MessageBox (win AS WindowType, style%, default%, caption$, msg$())

    ' Displays and operates a message dialog.
    '
    ' Closes the dialog window when the user selects a command button
    ' and returns the number of the selected button.
    '
    ' The dialog window is automatically sized and centered on the
    ' screen. It contains from 1 to 3 predefined command buttons
    ' depending on the value of <style%>. The initial default command
    ' button must be specified by <default%>.
    '
    ' Constants for the possible values of <style%> are defined in
    ' DIALOG.BI.
    '   CONST cMbOk = 1
    '   CONST cMbOkCancel = 2
    '   CONST cMbRetryCancel = 4
    '   CONST cMbYesNo = 8
    '   CONST cMbYesNoCancel = 16
    '
    ' The command buttons are numbered left to right starting at 1.
    '
    ' The size and location elements of <win> are initialized by this
    ' procedure. All other elements are used as is.
    '
    ' The horizontal centering of the command buttons relative to the
    ' window title can be fine-tuned by padding the title with spaces.
    '
    ' Note that <msg$()> is assumed to contain 1 line per element and
    ' that the number of elements must not exceed 19. Elements
    ' containing more than 74 characters are truncated.
   
    ' Declare a local variable for each command button.
    DIM cmd1 AS CommandButtonType
    DIM cmd2 AS CommandButtonType
    DIM cmd3 AS CommandButtonType

    IF LBOUND(msg$) <> 1 THEN
        LogError "DIALOG.BAS", "MessageBox", "LBOUND(msg$) <> 1"
        ERROR cErrorIllegalFunctionCall
    END IF

    lineCount = UBOUND(msg$)

    IF lineCount > 19 THEN
        LogError "DIALOG.BAS", "MessageBox", "lineCount > 19"
        ERROR cErrorIllegalFunctionCall
    END IF

    FOR i = 1 TO lineCount
        length = LEN(msg$(i))
        IF length > maxLineLength THEN maxLineLength = length
    NEXT
    
    IF maxLineLength > 74 THEN maxLineLength = 74
    
    IF win.border THEN
        win.rows = lineCount + 6
        win.row = (25 - win.rows) \ 2 + 1
        buttonRow = win.row + win.rows - 3
    ELSE
        win.rows = lineCount + 5
        win.row = (25 - win.rows) \ 2 + 1
        buttonRow = win.row + win.rows - 2
    END IF

    ' Initialize the button parameters.
    SELECT CASE style
        CASE cMbOk
            IF default <> 1 THEN
                LogError "DIALOG.BAS", "MessageBox", "style = cMbOk, default <> 1"
                ERROR cErrorIllegalFunctionCall
            END IF
            cmd1.row = buttonRow
            cmd1.col = 35
            cmd1.fCancel = cTrue
            cmd1Caption$ = "    OK    "
            IF maxLineLength < 29 THEN maxLineLength = 29
        CASE cMbOkCancel
            IF (default <> 1) AND (default <> 2) THEN
                LogError "DIALOG.BAS", "MessageBox", "style = cMbOkCancel, (default <> 1) AND (default <> 2)"
                ERROR cErrorIllegalFunctionCall
            END IF
            cmd1.row = buttonRow
            cmd1.col = 28
            cmd2.row = buttonRow
            cmd2.col = 42
            cmd2.fCancel = cTrue
            IF maxLineLength < 31 THEN maxLineLength = 31
            cmd1Caption$ = "    OK    "
            cmd2Caption$ = "  Cancel  "
        CASE cMbRetryCancel
            IF (default <> 1) AND (default <> 2) THEN
                LogError "DIALOG.BAS", "MessageBox", "style = cMbRetryCancel, (default <> 1) AND (default <> 2)"
                ERROR cErrorIllegalFunctionCall
            END IF
            cmd1.row = buttonRow
            cmd1.col = 29
            cmd2.row = buttonRow
            cmd2.col = 42
            cmd2.fCancel = cTrue
            IF maxLineLength < 30 THEN maxLineLength = 30
            cmd1Caption$ = "  &Retry  "
            cmd2Caption$ = "  Cancel  "
        CASE cMbYesNo
            IF (default <> 1) AND (default <> 2) THEN
                LogError "DIALOG.BAS", "MessageBox", "style = cMbYesNo, (default <> 1) AND (default <> 2)"
                ERROR cErrorIllegalFunctionCall
            END IF
            cmd1.row = buttonRow
            cmd1.col = 29
            cmd2.row = buttonRow
            cmd2.col = 42
            cmd2.fCancel = cTrue
            IF maxLineLength < 30 THEN maxLineLength = 30
            cmd1Caption$ = "   &Yes   "
            cmd2Caption$ = "   &No   "
        CASE cMbYesNoCancel
            IF (default < 1) OR (default > 3) THEN
                LogError "DIALOG.BAS", "MessageBox", "style = cMbYesNoCancel, (default < 1) OR (default > 3)"
                ERROR cErrorIllegalFunctionCall
            END IF
            cmd1.row = buttonRow
            cmd1.col = 22
            cmd2.row = buttonRow
            cmd2.col = 35
            cmd3.row = buttonRow
            cmd3.col = 49
            cmd3.fCancel = cTrue
            IF maxLineLength < 40 THEN maxLineLength = 40
            cmd1Caption$ = "   &Yes   "
            cmd2Caption$ = "    &No    "
            cmd3Caption$ = "  Cancel  "
        CASE ELSE
            LogError "DIALOG.BAS", "MessageBox", "bad style value"
            ERROR cErrorIllegalFunctionCall
    END SELECT

    win.cols = maxLineLength + 6
    win.col = (80 - win.cols) \ 2 + 1
    
    OpenWindow win, caption$

    HideTextCursor
    SetColor sWindowData(sCurrentWindow).fore, sWindowData(sCurrentWindow).back
  
    ' Print the message.
    FOR i = 1 TO lineCount
        HideMouseCursorIf win.row + i + 1, win.col + 3, 1, maxLineLength
        LOCATE win.row + i + 1, win.col + 3
        PRINT LEFT$(msg$(i), maxLineLength);
    NEXT

    ShowMouseCursor

    ' Start the main loop. Because the dialog contains only command
    ' button controls, the initial default command button will be the
    ' first command button in the tab order regardless of which
    ' command button has its default flag set. Because the tab order
    ' is established by the order in which the controls are called,
    ' we must call the default command button first.
    '
    ' If the user clicks a command button, then set the return value and
    ' exit from the loop.
    DO
        SELECT CASE default
            CASE 1
                CpCommandButton cmd1, cmd1Caption$
                IF cmd2.row THEN CpCommandButton cmd2, cmd2Caption$
                IF cmd3.row THEN CpCommandButton cmd3, cmd3Caption$
                IF cmd1.fClicked THEN
                    MessageBox = 1
                    EXIT DO
                END IF
                IF cmd2.fClicked THEN
                    MessageBox = 2
                    EXIT DO
                END IF
                IF cmd3.fClicked THEN
                    MessageBox = 3
                    EXIT DO
                END IF
            CASE 2
                IF cmd2.row THEN CpCommandButton cmd2, cmd2Caption$
                IF cmd3.row THEN CpCommandButton cmd3, cmd3Caption$
                CpCommandButton cmd1, cmd1Caption$
                IF cmd1.fClicked THEN
                    MessageBox = 1
                    EXIT DO
                END IF
                IF cmd2.fClicked THEN
                    MessageBox = 2
                    EXIT DO
                END IF
                IF cmd3.fClicked THEN
                    MessageBox = 3
                    EXIT DO
                END IF
            CASE 3
                IF cmd3.row THEN CpCommandButton cmd3, cmd3Caption$
                CpCommandButton cmd1, cmd1Caption$
                IF cmd2.row THEN CpCommandButton cmd2, cmd2Caption$
                IF cmd1.fClicked THEN
                    MessageBox = 1
                    EXIT DO
                END IF
                IF cmd2.fClicked THEN
                    MessageBox = 2
                    EXIT DO
                END IF
                IF cmd3.fClicked THEN
                    MessageBox = 3
                    EXIT DO
                END IF
        END SELECT
    LOOP

    CloseWindow

END FUNCTION

SUB OpenScreen (fore%, back%, patternCharacter$)

    ' Preserves the background and creates a full screen backdrop using
    ' the character specified by the first byte of <patternCharacter$>
    ' (or a space character if <patternCharacter$> is a null string) and
    ' the colors specified by <fore%> and <back%>.
    '
    ' The preserved background can be restored by calling the CloseWindow
    ' procedure.
   
    DIM win AS WindowType
    win.row = 1
    win.col = 1
    win.rows = 25
    win.cols = 80
    win.fore = fore
    win.back = back
    OpenWindow win, ""

    c$ = LEFT$(patternCharacter$, 1)
    IF c$ = "" THEN c$ = " "

    SetColor fore, back
    HideTextCursor
    HideMouseCursor
   
    FOR row = 1 TO 25
        LOCATE row, 1
        PRINT STRING$(80, c$);
    NEXT
    
    ShowMouseCursor

    ' This statement solves a problem with the drop down list box
    ' procedure causing the module to ignore keyboard input under
    ' specific conditions.
    sfDiscardKey = cTrue

END SUB

SUB OpenWindow (win AS WindowType, caption$) STATIC

    ' Preserves the current screen image, copies the window information
    ' to the shared window data array, and displays a text mode window
    ' as specified by the elements of <win>. The window is centered on
    ' the screen vertically if the row element is 0 and centered on the
    ' screen horizontally if the column element is 0.
    '
    ' Setting the fTitleBar parameter to true causes the window to be
    ' displayed with a title bar (the window title bar is independent
    ' of the window title, which must specified separately).
    '
    ' Setting the fNoShadow parameter to true causes the window to be
    ' displayed without a shadow.
    '
    ' Setting the fNoModColors parameter to true causes this procedure
    ' to NOT modify the title bar and control caption colors for the
    ' *image* of the previous window. See below for more information.
    '
    ' The visible screen image is preserved by leaving it intact on the
    ' currently visible screen page and displaying the new window on the
    ' next screen page. The visible screen image is first copied to the
    ' next screen page so the window background will appear unchanged
    ' when that screen page is made visible.
    '
    ' The next screen page is not made visible until all operations
    ' on that screen page are complete. This takes advantage of the
    ' automatic synchronization with vertical retrace that (most)
    ' current display adapters perform before switching the visible
    ' screen page. Synchronization with vertical retrace ensures that
    ' the entire screen will be updated in a single vertical frame
    ' (assuming the system is fast enough to update the screen within
    ' the vertical frame time - a very reasonable assumption for
    ' recent systems and text mode). Without synchronization, most
    ' screen updates will require two vertical frames, and most users
    ' will be able to perceive the window appearing in two steps.
    '
    ' Refer to the CloseWindow procedure for more information.
    
    IF NOT sfInitModuleOK THEN InitModule

    ' Simulate an Illegal function call error if no more screen pages.
    IF sCurrentWindow = 7 THEN
        LogError "DIALOG.BAS", "OpenWindow", "sCurrentWindow = 7"
        ERROR cErrorIllegalFunctionCall
    END IF
   
    IF win.rows < 1 THEN
        LogError "DIALOG.BAS", "OpenWindow", "win.rows < 1"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF win.cols < 4 THEN
        LogError "DIALOG.BAS", "OpenWindow", "win.cols < 4"
        ERROR cErrorIllegalFunctionCall
    END IF

    visiblePage = sCurrentWindow
    sCurrentWindow = sCurrentWindow + 1

    ' Center the window as necessary.
    IF win.row THEN row = win.row ELSE row = 13 - win.rows \ 2
    IF win.col THEN col = win.col ELSE col = 41 - win.cols \ 2

    lastRow = row + win.rows - 1
    lastCol = col + win.cols - 1

    IF row < 1 THEN
        LogError "DIALOG.BAS", "OpenWindow", "row < 1"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF lastRow > 25 THEN
        LogError "DIALOG.BAS", "OpenWindow", "lastRow > 25"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF col < 1 THEN
        LogError "DIALOG.BAS", "OpenWindow", "col < 1"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF lastCol > 80 THEN
        LogError "DIALOG.BAS", "OpenWindow", "lastCol > 80"
        ERROR cErrorIllegalFunctionCall
    END IF

    ' Copy the window colors to the shared window data array
    ' and the title bar colors to local variables. Substitute
    ' the default color as necessary.
    IF win.fore THEN
        sWindowData(sCurrentWindow).fore = win.fore
    ELSE
        sWindowData(sCurrentWindow).fore = cWindowFore
    END IF
    IF win.back THEN
        sWindowData(sCurrentWindow).back = win.back
    ELSE
        sWindowData(sCurrentWindow).back = cWindowBack
    END IF
    IF win.titleBarFore THEN
        titleBarFore = win.titleBarFore
    ELSE
        titleBarFore = cTitleBarFore
    END IF
    IF win.titleBarBack THEN
        titleBarBack = win.titleBarBack
    ELSE
        titleBarBack = cTitleBarBack
    END IF

    ' Copy the data for this window to the shared array.
    sWindowData(sCurrentWindow).row = row
    sWindowData(sCurrentWindow).col = col
    sWindowData(sCurrentWindow).rows = win.rows
    sWindowData(sCurrentWindow).cols = win.cols
    sWindowData(sCurrentWindow).fTitleBar = win.fTitleBar

    sWindowData(sCurrentWindow).activeControl = 0
    sWindowData(sCurrentWindow).controlCount = 0

    ' Set the active screen page.
    SCREEN , , sCurrentWindow, visiblePage

    HideMouseCursor
 
    ' Copy the contents of the visible screen page to the
    ' active screen page.
    PCOPY visiblePage, sCurrentWindow

    IF NOT win.fNoModColors THEN
   
        ' If there is a previous window, then modify the title bar
        ' and control caption colors for the *image* of the previous
        ' window to provide a visual indication that the previous
        ' window is not active.
        IF sCurrentWindow > 1 THEN

            previousWindow = sCurrentWindow - 1

            IF sWindowData(previousWindow).fTitleBar THEN
                SetAttribute sWindowData(previousWindow).row, sWindowData(previousWindow).col, 1, sWindowData(previousWindow).cols, sWindowData(previousWindow).fore, sWindowData(previousWindow).back
            END IF
        
            FOR i = 1 TO sWindowData(previousWindow).controlCount
                SELECT CASE sControlData(previousWindow, i).controlType
                    CASE cCtCommandButton
                        cols = sControlData(previousWindow, i).lastCol - sControlData(previousWindow, i).col + 1
                        SetAttribute sControlData(previousWindow, i).row, sControlData(previousWindow, i).col, 1, cols, cWhite, cCmdBack
                    CASE cCtLabel, cCtOptionButton, cCtCheckBox
                        fore = sWindowData(previousWindow).fore
                        back = sWindowData(previousWindow).back
                        cols = sControlData(previousWindow, i).lastCol - sControlData(previousWindow, i).col + 1
                        SetAttribute sControlData(previousWindow, i).row, sControlData(previousWindow, i).col, 1, cols, fore, back
                END SELECT
            NEXT
        END IF
    END IF

    ' Truncate the caption to fit.
    caption$ = LEFT$(caption$, win.cols - 4)

    ' Initialize the border characters.
    SELECT CASE win.border
        CASE 0
            upperLeft$ = " "
            upperRight$ = " "
            lowerLeft$ = " "
            lowerRight$ = " "
            horz$ = " "
            vert$ = " "
        CASE 1
            upperLeft$ = ""
            upperRight$ = ""
            lowerLeft$ = ""
            lowerRight$ = ""
            horz$ = ""
            vert$ = ""
        CASE 2
            upperLeft$ = ""
            upperRight$ = ""
            lowerLeft$ = ""
            lowerRight$ = ""
            horz$ = ""
            vert$ = ""
        CASE ELSE
            LogError "DIALOG.BAS", "OpenWindow", "bad win.border value"
            ERROR cErrorIllegalFunctionCall

    END SELECT

    ' Position the cursor at the upper left corner of the window.
    LOCATE row, col, 0

    ' Print the window on the active screen page.
    leftLength = (win.cols - LEN(caption$) - 4) \ 2
    rightLength = (win.cols - LEN(caption$) - leftLength - 4)
    IF win.fTitleBar THEN
        SetColor titleBarBack, titleBarBack
        PRINT upperLeft$; STRING$(leftLength, horz$);
        SetColor titleBarFore, titleBarBack
        PRINT " "; caption$; " ";
        SetColor titleBarBack, titleBarBack
        PRINT STRING$(rightLength, horz$); upperRight$;
        SetColor sWindowData(sCurrentWindow).fore, sWindowData(sCurrentWindow).back
    ELSE
        SetColor sWindowData(sCurrentWindow).fore, sWindowData(sCurrentWindow).back
        IF caption$ <> "" THEN
            PRINT upperLeft$; STRING$(leftLength, horz$); " "; caption$; " "; STRING$(rightLength, horz$); upperRight$;
        ELSE
            PRINT upperLeft$; STRING$(win.cols - 2, horz$); upperRight$
        END IF
    END IF
    FOR r = row + 1 TO lastRow - 1
        LOCATE r, col
        PRINT vert$; SPACE$(win.cols - 2); vert$;
    NEXT
    LOCATE lastRow, col
    PRINT lowerLeft$; STRING$(win.cols - 2, horz$); lowerRight$;

    ' If the shadow is enabled and it will fit on an 80x25 screen,
    ' then add a shadow.
    IF win.fNoShadow = cFalse AND lastRow < 25 AND lastCol < 79 THEN
        Shadow row + 1, lastCol + 1, win.rows, 2
        Shadow lastRow + 1, col + 2, 1, win.cols - 2
    END IF

    ' Set the visible screen page.
    SCREEN , , , sCurrentWindow
 
    ShowMouseCursor

    ' This statement solves a problem with the drop down list box
    ' procedure causing the module to ignore keyboard input under
    ' specific conditions.
    sfDiscardKey = cTrue
  
END SUB

SUB PrintControlCap (row%, col%, fore%, back%, accessFore%, accessCol%, caption$)

    ' Displays a control caption as specified by the arguments.

    HideTextCursor
    LOCATE row, col
    SetColor fore, back
    HideMouseCursorIf row, col, 2, LEN(caption$)
    PRINT caption$;
    IF accessCol THEN SetAttribute row, accessCol, 1, 1, accessFore, back
    ShowMouseCursor

END SUB

SUB PrintDebug (debug$) STATIC

    ' Prints <debug$> to screen page 0 and sets a shared flag to
    ' enable toggling of the display between the screen page of
    ' the current window and screen page 0 with the F4 key.
    '
    ' Note that you can control the printing process to a limited
    ' extent by adding control characters to <debug$>.

    ' Set the active screen page to page 0 without changing
    ' the visible page.
    HideMouseCursor
    SCREEN , , 0, sCurrentWindow

    ' If this is the first call, then clear screen page 0,
    ' initialize the cursor coordinates, and set the shared
    ' flag.
    IF cursorRow = 0 THEN
        CLS
        cursorRow = 1
        cursorCol = 1
        sfEnableToggle = cTrue
    END IF
    
    LOCATE cursorRow, cursorCol

    ' Print and move the cursor to the next line.
    PRINT debug$

    ' Save the new cursor location.
    cursorRow = CSRLIN
    cursorCol = POS(0)

    ' Restore the active and visible page.
    SCREEN , , sCurrentWindow, sCurrentWindow
    ShowMouseCursor

END SUB

SUB PrintFrame (row%, col%, rows%, cols%, border%, fFill%, caption$)

    ' Displays a frame as specified by the arguments.
    '
    ' The valid values for <border%> are 0, 1, and 2.
    '
    ' The interior of the frame is filled with spaces if
    ' <fFill%> is true.

    IF rows < 2 THEN
        LogError "DIALOG.BAS", "PrintFrame", "rows < 2"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF cols < 4 THEN
        LogError "DIALOG.BAS", "PrintFrame", "cols < 4"
        ERROR cErrorIllegalFunctionCall
    END IF

    lastRow = row + rows - 1
    lastCol = col + cols - 1

    IF row < 1 THEN
        LogError "DIALOG.BAS", "PrintFrame", "row < 1"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF lastRow > 25 THEN
        LogError "DIALOG.BAS", "PrintFrame", "lastRow > 25"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF col < 1 THEN
        LogError "DIALOG.BAS", "PrintFrame", "col < 1"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF lastCol > 80 THEN
        LogError "DIALOG.BAS", "PrintFrame", "lastCol > 80"
        ERROR cErrorIllegalFunctionCall
    END IF

    ' Truncate the caption to fit.
    caption$ = LEFT$(caption$, cols - 4)
  
    ' Initialize the border characters.
    SELECT CASE border
        CASE 1
            upperLeft$ = ""
            upperRight$ = ""
            lowerLeft$ = ""
            lowerRight$ = ""
            horz$ = ""
            vert$ = ""
        CASE 2
            upperLeft$ = ""
            upperRight$ = ""
            lowerLeft$ = ""
            lowerRight$ = ""
            horz$ = ""
            vert$ = ""
        CASE ELSE
            LogError "DIALOG.BAS", "PrintFrame", "bad border value"
            ERROR cErrorIllegalFunctionCall

    END SELECT
    
    ' If any windows are open, then set the display
    ' colors to colors of the current window.
    IF sCurrentWindow THEN
        SetColor sWindowData(sCurrentWindow).fore, sWindowData(sCurrentWindow).back
    END IF

    HideTextCursor
    LOCATE row, col
    HideMouseCursor

    ' Print the frame.
    IF caption$ <> "" THEN
        rightLength = (cols - LEN(caption$) - 4)
        PRINT upperLeft$; " "; caption$; " "; STRING$(rightLength, horz$); upperRight$;
    ELSE
        PRINT upperLeft$; STRING$(cols - 2, horz$); upperRight$;
    END IF
    IF fFill THEN
        FOR r = row + 1 TO lastRow - 1
            LOCATE r, col
            PRINT vert$; SPACE$(cols - 2); vert$;
        NEXT
    ELSE
        FOR r = row + 1 TO lastRow - 1
            LOCATE r, col
            PRINT vert$;
            LOCATE r, lastCol
            PRINT vert$;
        NEXT
    END IF
    LOCATE lastRow, col
    PRINT lowerLeft$; STRING$(cols - 2, horz$); lowerRight$;

    ShowMouseCursor

END SUB

SUB PrintStaticText (row%, col%, text$)

    ' Prints <text$> at <row%>, <col%> using the colors for
    ' the current window.

    ' If any windows are open, then set the display colors
    ' to colors of the current window.
    IF sCurrentWindow THEN
        SetColor sWindowData(sCurrentWindow).fore, sWindowData(sCurrentWindow).back
    END IF

    HideTextCursor
    LOCATE row, col
    HideMouseCursorIf row, col, 1, LEN(text$)
    PRINT text$;
    ShowMouseCursor
   
END SUB

SUB ProgressBox (row%, col%, cols%, percent%) STATIC

    ' Displays a progress box as specified by the arguments.
    '
    ' For the first call, <percent%> must be set to 0.
    '
    ' The progress box is centered on the screen horizontally if
    ' the value of <col%> is 0.
    '
    ' The height of the progress box is fixed at 3 rows.
    '
    ' The progress box is displayed using the colors of the
    ' current window.
    '
    ' The progress bar highlight is created by using CHR$(219).
    ' The percent field highlight is created by reversing the
    ' foreground and background colors.

    ' Convert the current value of <percent%> to a 4 byte string.
    percent$ = RIGHT$("  " + STR$(percent) + "%", 4)

    SELECT CASE percent
        CASE 0

            ' Center the box as necessary.
            IF col = 0 THEN col = 41 - cols \ 2

            ' Simulate an illegal function call error if
            ' no windows are open.
            IF sCurrentWindow < 1 THEN
                LogError "DIALOG.BAS", "ProgressBox", "sCurrentWindow < 1"
                ERROR cErrorIllegalFunctionCall
            END IF

            IF cols < 6 THEN
                LogError "DIALOG.BAS", "ProgressBox", "cols < 6"
                ERROR cErrorIllegalFunctionCall
            END IF
            IF row < 1 THEN
                LogError "DIALOG.BAS", "ProgressBox", "row < 1"
                ERROR cErrorIllegalFunctionCall
            END IF
            IF row > 23 THEN
                LogError "DIALOG.BAS", "ProgressBox", "row > 23"
                ERROR cErrorIllegalFunctionCall
            END IF
            IF col < 1 THEN
                LogError "DIALOG.BAS", "ProgressBox", "col < 1"
                ERROR cErrorIllegalFunctionCall
            END IF
            IF col + cols > 81 THEN
                LogError "DIALOG.BAS", "ProgressBox", "col + cols > 81"
                ERROR cErrorIllegalFunctionCall
            END IF
  
            barRow = row + 1
            barCols = cols - 4
            barStartCol = col + 2
            barEndCol = col + cols - 3
            percentFieldStartCol = col + cols \ 2 - 2
          
            ' Print an empty frame and the percent field.
            PrintFrame row, col, 3, cols, 1, cTrue, ""
            HideTextCursor
            LOCATE barRow, percentFieldStartCol
            HideMouseCursor
            PRINT percent$;

            fInitialized = cTrue

        CASE 1 TO 100

            ' If <percent%> is between 1 and 100 inclusive, then update
            ' the progress bar and the percent field.

            IF NOT fInitialized THEN
                LogError "DIALOG.BAS", "ProgressBox", "is first call and percent <> 0"
                ERROR cErrorIllegalFunctionCall
            END IF

            ' Calculate the end column for the highlighted
            ' section of the progress bar.
            hiLiteEndCol = barStartCol + percent * barCols \ 100
            
            SetColor sWindowData(sCurrentWindow).fore, sWindowData(sCurrentWindow).back
            HideTextCursor
            HideMouseCursorIf barRow, barStartCol, 1, barCols

            ' Print the progress bar and the percent field.
            FOR c = barStartCol TO percentFieldStartCol - 1
                IF c <= hiLiteEndCol THEN
                    LOCATE barRow, c
                    PRINT CHR$(219);
                END IF
            NEXT
            FOR c = percentFieldStartCol TO percentFieldStartCol + 3
                LOCATE barRow, c

                ' If the current character of the percent field falls
                ' within the highlighted section of the bar, then reverse
                ' the display colors, print the character, and change
                ' the colors back. Otherwise, just print the character.
                IF c <= hiLiteEndCol THEN
                    SetColor sWindowData(sCurrentWindow).back, sWindowData(sCurrentWindow).fore
                    PRINT MID$(percent$, c - percentFieldStartCol + 1, 1);
                    SetColor sWindowData(sCurrentWindow).fore, sWindowData(sCurrentWindow).back
                ELSE
                    PRINT MID$(percent$, c - percentFieldStartCol + 1, 1);
                END IF
            NEXT
            FOR c = percentFieldStartCol + 4 TO barEndCol
                IF c <= hiLiteEndCol THEN
                    LOCATE barRow, c
                    PRINT CHR$(219);
                END IF
            NEXT

        CASE ELSE
            LogError "DIALOG.BAS", "ProgressBox", "bad percent value"
            ERROR cErrorIllegalFunctionCall

    END SELECT

    ShowMouseCursor

END SUB

FUNCTION ScreenAscii (row%, col%) STATIC

    ' Returns the ASCII code for the character at <row%>, <col%>
    ' on the current (active and visible) screen page.
    '
    ' The procedures in this module call this function and/or the
    ' ScreenAttribute function in place of the SCREEN function
    ' because these functions are much faster than the SCREEN
    ' function and because these functions avoid moving the text
    ' cursor.
    '
    ' Note that this is not a general purpose function - it assumes
    ' a VGA compatible display adapter operating in 80x25 text mode.

    IF bufferSeg = 0 THEN

        ' Get the segment address of the display buffer.
        ' The address depends on the current video mode,
        ' which is stored in the BIOS data area at offset
        ' 49h. The address is B000h for mode 7 (the VGA
        ' monochrome alphanumeric mode), or B800h for
        ' any other alphanumeric mode.
        IF MemByte(&H40, &H49) = 7 THEN
            bufferSeg& = &HB000&
        ELSE
            bufferSeg& = &HB800&
        END IF
    END IF

    ' Calculate the offset address of the current screen page.
    pageOffset = sCurrentWindow * 4096

    ' Get the character code for the specified character.
    ' Refer to the SetAttribute procedure for more information.
    ScreenAscii = MemByte(bufferSeg&, UInt(pageOffset + (row - 1) * 160 + (col - 1) * 2))

END FUNCTION

FUNCTION ScreenAttribute (row%, col%) STATIC

    ' Returns the attribute value for the character at <row%>, <col%>
    ' on the current (active and visible) screen page.
    '
    ' The procedures in this module call this function and/or the
    ' ScreenAscii function in place of the SCREEN function because
    ' these functions are much faster than the SCREEN function and
    ' because these functions avoid moving the text cursor.
    '
    ' Note that this is not a general purpose function - it assumes
    ' a VGA compatible display adapter operating in 80x25 text mode.

    IF bufferSeg = 0 THEN

        ' Get the segment address of the display buffer.
        ' The address depends on the current video mode,
        ' which is stored in the BIOS data area at offset
        ' 49h. The address is B000h for mode 7 (the VGA
        ' monochrome alphanumeric mode), or B800h for
        ' any other alphanumeric mode.
        IF MemByte(&H40, &H49) = 7 THEN
            bufferSeg& = &HB000&
        ELSE
            bufferSeg& = &HB800&
        END IF
    END IF

    ' Calculate the offset address of the current screen page.
    pageOffset = sCurrentWindow * 4096

    ' Get the attribute value for the specified character.
    ' Refer to the SetAttribute procedure for more information.
    ScreenAttribute = MemByte(bufferSeg&, UInt(pageOffset + (row - 1) * 160 + (col - 1) * 2 + 1))

END FUNCTION

SUB SetActiveControl (callerIndex%, direction%)

    ' Sets the active control for the current window.
    '
    ' The active control is set relative to the caller (identified
    ' by the value in <callerIndex%>) as specified by the value in
    ' <direction%>. Constants for the possible <direction%> values
    ' are defined in the module level code of this module.
    '
    ' The term "caller" refers to the control instance being processed
    ' by the control procedure that called this procedure.
    '
    ' This procedure automatically bypasses labels.

    ' Set the variables used to store the minimum distance
    ' between controls to the maximum value for an integer.
    minDistOnAxis = 32767
    minDistCrossing = 32767
    minDist = 32767

    ' Set the active control.
    SELECT CASE direction

        CASE cAcThis
            ' Set the active control to the caller.
            sWindowData(sCurrentWindow).activeControl = callerIndex

        CASE cAcNext
            ' Set the active control to the control following the caller
            ' in the tab order. If there is no following control, then
            ' set the active control to the first control in the tab order.
            IF sWindowData(sCurrentWindow).controlCount > callerIndex THEN
                sWindowData(sCurrentWindow).activeControl = callerIndex + 1
            ELSE
                sWindowData(sCurrentWindow).activeControl = 1
            END IF

        CASE cAcPrev
            ' Set the active control to the control preceeding the caller
            ' in the tab order. If there is no preceeding control, then
            ' set the active control to the last control in the tab order.
            IF callerIndex > 1 THEN
                sWindowData(sCurrentWindow).activeControl = callerIndex - 1
            ELSE
                sWindowData(sCurrentWindow).activeControl = sWindowData(sCurrentWindow).controlCount
            END IF
        CASE ELSE
            ' This complex block of code allows the user to navigate
            ' between controls with the arrow keys. In the following,
            ' "direction axis" refers to the line that originates at
            ' the upper left corner of the current active control and
            ' extends in the direction specified by <direction%>.
            '
            ' If the current window contains any controls other than
            ' labels and disabled command buttons for which the upper
            ' left corner of the control is on the direction axis,
            ' then the active control is set to the nearest one.
            '
            ' Otherwise, if the current window contains any controls
            ' other than labels and disabled command buttons for which
            ' the near boundary of the control crosses the direction
            ' axis, then the active control is set to the nearest one.
            '
            ' Otherwise, the active control is set to the nearest
            ' control that is not a label or a disabled command
            ' button. The relationship used to calculate the distance
            ' between controls is biased to favor controls that are
            ' closer to the direction axis. This relationship assumes
            ' an 80x25 screen with an aspect ratio of 4:3. For maximum
            ' execution speed, this code uses integer operations only.
            ' The constant "4700&" minimizes the effect of round-off
            ' errors.

            callerRow = sControlData(sCurrentWindow, callerIndex).row
            callerCol = sControlData(sCurrentWindow, callerIndex).col
            FOR i = 1 TO sWindowData(sCurrentWindow).controlCount
                iRow = sControlData(sCurrentWindow, i).row
                iCol = sControlData(sCurrentWindow, i).col
                iLastRow = sControlData(sCurrentWindow, i).lastRow
                iLastCol = sControlData(sCurrentWindow, i).lastCol
                SELECT CASE direction
                    CASE cAcUp
                        IF iRow < callerRow THEN
                            dx = ABS(callerCol - iCol)
                            dy = callerRow - iRow
                            IF iCol = callerCol THEN
                                IF dy < minDistOnAxis THEN
                                    IF sControlData(sCurrentWindow, i).controlType <> cCtLabel AND sControlData(sCurrentWindow, i).controlType <> cCtDisabledCommandButton THEN
                                        minDistOnAxis = dy
                                        indexOnAxis = i
                                    END IF
                                END IF
                            ELSEIF callerCol >= iCol AND callerCol <= iLastCol THEN
                                IF dy < minDistOnAxis THEN
                                    IF sControlData(sCurrentWindow, i).controlType <> cCtLabel AND sControlData(sCurrentWindow, i).controlType <> cCtDisabledCommandButton THEN
                                        minDistCrossing = dy
                                        indexCrossing = i
                                    END IF
                                END IF
                            ELSE
                                dist = (4700& * dx \ 20) + (4700& * 3 * dy \ 25)
                                IF dist < minDist THEN
                                    IF sControlData(sCurrentWindow, i).controlType <> cCtLabel AND sControlData(sCurrentWindow, i).controlType <> cCtDisabledCommandButton THEN
                                        minDist = dist
                                        indexNearest = i
                                    END IF
                                END IF
                            END IF
                        END IF

                    CASE cAcDown
                        IF iRow > callerRow THEN
                            dx = ABS(callerCol - iCol)
                            dy = iRow - callerRow
                            IF iCol = callerCol THEN
                                IF dy < minDistOnAxis THEN
                                    IF sControlData(sCurrentWindow, i).controlType <> cCtLabel AND sControlData(sCurrentWindow, i).controlType <> cCtDisabledCommandButton THEN
                                        minDistOnAxis = dy
                                        indexOnAxis = i
                                    END IF
                                END IF
                            ELSEIF callerCol >= iCol AND callerCol <= iLastCol THEN
                                IF dy < minDistCrossing THEN
                                    IF sControlData(sCurrentWindow, i).controlType <> cCtLabel AND sControlData(sCurrentWindow, i).controlType <> cCtDisabledCommandButton THEN
                                        minDistCrossing = dy
                                        indexCrossing = i
                                    END IF
                                END IF
                            ELSE
                                dist = (4700& * dx \ 20) + (4700& * 3 * dy \ 25)
                                IF dist < minDist THEN
                                    IF sControlData(sCurrentWindow, i).controlType <> cCtLabel AND sControlData(sCurrentWindow, i).controlType <> cCtDisabledCommandButton THEN
                                        minDist = dist
                                        indexNearest = i
                                    END IF
                                END IF
                            END IF
                        END IF

                    CASE cAcLeft
                        IF iCol < callerCol THEN
                            dx = callerCol - iCol
                            dy = ABS(callerRow - iRow)
                            IF iRow = callerRow THEN
                                IF dx < minDistOnAxis THEN
                                    IF sControlData(sCurrentWindow, i).controlType <> cCtLabel AND sControlData(sCurrentWindow, i).controlType <> cCtDisabledCommandButton THEN
                                        minDistOnAxis = dx
                                        indexOnAxis = i
                                    END IF
                                END IF
                            ELSEIF callerRow >= iRow AND callerRow <= iLastRow THEN
                                IF dx < minDistCrossing THEN
                                    IF sControlData(sCurrentWindow, i).controlType <> cCtLabel AND sControlData(sCurrentWindow, i).controlType <> cCtDisabledCommandButton THEN
                                        minDistCrossing = dx
                                        indexCrossing = i
                                    END IF
                                END IF
                            ELSE
                                dist = (4700& * dx \ 20) + (4700& * 3 * dy \ 25)
                                IF dist < minDist THEN
                                    IF sControlData(sCurrentWindow, i).controlType <> cCtLabel AND sControlData(sCurrentWindow, i).controlType <> cCtDisabledCommandButton THEN
                                        minDist = dist
                                        indexNearest = i
                                    END IF
                                END IF
                            END IF
                        END IF

                    CASE cAcRight
                        IF iCol > callerCol THEN
                            dx = iCol - callerCol
                            dy = ABS(callerRow - iRow)
                            IF iRow = callerRow THEN
                                IF dx < minDistOnAxis THEN
                                    IF sControlData(sCurrentWindow, i).controlType <> cCtLabel AND sControlData(sCurrentWindow, i).controlType <> cCtDisabledCommandButton THEN
                                        minDistOnAxis = dx
                                        indexOnAxis = i
                                    END IF
                                END IF
                            ELSEIF callerRow >= iRow AND callerRow <= iLastRow THEN
                                IF dx < minDistCrossing THEN
                                    IF sControlData(sCurrentWindow, i).controlType <> cCtLabel AND sControlData(sCurrentWindow, i).controlType <> cCtDisabledCommandButton THEN
                                        minDistCrossing = dx
                                        indexCrossing = i
                                    END IF
                                END IF
                            ELSE
                                dist = (4700& * dx \ 20) + (4700& * 3 * dy \ 25)
                                IF dist < minDist THEN
                                    IF sControlData(sCurrentWindow, i).controlType <> cCtLabel AND sControlData(sCurrentWindow, i).controlType <> cCtDisabledCommandButton THEN
                                        minDist = dist
                                        indexNearest = i
                                    END IF
                                END IF
                            END IF
                        END IF

                    CASE ELSE
                        LogError "DIALOG.BAS", "SetActiveControl", "bad direction value"
                        ERROR cErrorIllegalFunctionCall

                END SELECT
            NEXT

            IF indexOnAxis THEN
                sWindowData(sCurrentWindow).activeControl = indexOnAxis
            ELSEIF indexCrossing THEN
                sWindowData(sCurrentWindow).activeControl = indexCrossing
            ELSEIF indexNearest THEN
                sWindowData(sCurrentWindow).activeControl = indexNearest
            END IF

    END SELECT

    ' If the active control for the current window is a label or a
    ' disabled command button, then bypass it by setting the active
    ' control to the next or previous control in the tab order.
    ' If the active control for the current window is a label or
    ' a disabled command button, then <direction%> must be set to
    ' cAcNext or cAcPrev because these controls will never call
    ' this procedure with <direction%> set to cAcThis and this
    ' procedure skips labels and disabled command buttons if
    ' <direction%> is set to cAcUp, cAcDown, cAcLeft, or cAcRight.
    ' A recursive approach provides a compact way to bypass multiple
    ' adjacent (in the tab order) labels and/or disabled command
    ' buttons. The CpCommandButton procedure prevents any problems
    ' with unlimited recursion if the current window contains only
    ' disabled command buttons by generating a run-time error if
    ' the first command button in the tab order for the current
    ' window is disabled.
    '
    ' Note that unlimited recursion can be a problem if the
    ' current window contains only label controls.
    ac = sWindowData(sCurrentWindow).activeControl
    IF sControlData(sCurrentWindow, ac).controlType = cCtLabel OR sControlData(sCurrentWindow, ac).controlType = cCtDisabledCommandButton THEN
        SetActiveControl ac, direction
    END IF

END SUB

SUB SetAttribute (row%, col%, rows%, cols%, fore%, back%) STATIC

    ' Sets the attribute for the rectangular block of characters
    ' specified by the arguments.

    IF fore < 0 OR fore > cHiWhite AND fore <> cBlack THEN
        LogError "DIALOG.BAS", "SetAttribute", "fore < 0 OR fore > cHiWhite AND fore <> cBlack"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF back < 0 OR back > cHiWhite AND back <> cBlack THEN
        LogError "DIALOG.BAS", "SetAttribute", "back < 0 OR back > cHiWhite AND back <> cBlack"
        ERROR cErrorIllegalFunctionCall
    END IF

    lastRow = row + rows - 1
    lastCol = col + cols - 1

    IF row < 1 THEN
        LogError "DIALOG.BAS", "SetAttribute", "row < 1"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF lastRow > 25 THEN
        LogError "DIALOG.BAS", "SetAttribute", "lastRow > 25"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF col < 1 THEN
        LogError "DIALOG.BAS", "SetAttribute", "col < 1"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF lastCol > 80 THEN
        LogError "DIALOG.BAS", "SetAttribute", "lastCol > 80"
        ERROR cErrorIllegalFunctionCall
    END IF

    IF bufferSeg = 0 THEN

        ' Get the segment address of the display buffer.
        ' The address depends on the current video mode,
        ' which is stored in the BIOS data area at offset
        ' 49h. The address is B000h for mode 7 (the VGA
        ' monochrome alphanumeric mode), or B800h for
        ' any other alphanumeric mode.
        IF MemByte(&H40, &H49) = 7 THEN
            bufferSeg = &HB000
        ELSE
            bufferSeg = &HB800
        END IF
    END IF

    ' Calculate the offset address of the current screen page.
    pageOffset = sCurrentWindow * 4096

    ' Adjust the color values as necessary and
    ' combine them into an attribute value.
    IF fore = cBlack THEN fore = 0
    IF back = cBlack THEN back = 0
    attribute = back
    ShiftLeft attribute, 4
    attribute = attribute OR fore

    FOR r = row TO lastRow
        FOR c = col TO lastCol

            ' Calculate the offset address in the display
            ' buffer for the character at r, c. For each
            ' character on the screen, the display buffer
            ' contains a character code byte followed by
            ' an attribute byte. The first character in
            ' the buffer is at row 0, column 0.
            offset = pageOffset + (r - 1) * 160 + (c - 1) * 2

            ' Write the attribute to the display buffer.
            DEF SEG = bufferSeg
            POKE offset + 1, attribute
            DEF SEG
        NEXT
    NEXT

END SUB

SUB SetColor (fore%, back%) STATIC

    ' Sets the foreground and background colors for the display.
    '
    ' Unlike the COLOR statement, this procedure can be used to select
    ' high-intensity background colors.
   
    IF NOT sfInitModuleOK THEN InitModule

    foreTemp = fore
    backTemp = back

    ' Convert the special value for black to the actual value.
    IF foreTemp = cBlack THEN foreTemp = 0
    IF backTemp = cBlack THEN backTemp = 0

    IF foreTemp < 0 OR foreTemp > cHiWhite THEN
        LogError "DIALOG.BAS", "SetColor", "fore < 0 OR fore > cHiWhite AND fore <> cBlack"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF backTemp < 0 OR backTemp > cHiWhite THEN
        LogError "DIALOG.BAS", "SetColor", "back < 0 OR back > cHiWhite AND back <> cBlack"
        ERROR cErrorIllegalFunctionCall
    END IF

    ' The COLOR statement normally selects blinking foreground colors when
    ' bit 4 of the foreground color argument is set. With high-intensity
    ' background colors enabled, bit 4 of the foreground color argument
    ' controls the background color intensity in the same manner as bit 3
    ' of the foreground color argument controls foreground color intensity.
    '
    ' If backTemp specifies a high-intensity background color, then set
    ' bit 4 of foreTemp.
    IF backTemp > 7 THEN foreTemp = foreTemp OR 16

    COLOR foreTemp, backTemp

END SUB

SUB SetFocus (tabindex%)

    ' Sets the active control for the current window to the
    ' non-label control specified by <tabIndex%>.
    '
    ' Refer to comments at the end of the SetActiveControl
    ' procedure for more information.

    IF sControlData(sCurrentWindow, tabindex).controlType <> cCtLabel THEN
        SetActiveControl tabindex, cAcThis
    END IF
    
END SUB

SUB Shadow (row%, col%, rows%, cols%) STATIC

    ' Creates a shadow effect by altering the attribute for the
    ' rectangular block of characters specified by the arguments.

    lastRow = row + rows - 1
    lastCol = col + cols - 1

    IF row < 1 THEN
        LogError "DIALOG.BAS", "Shadow", "row < 1"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF lastRow > 25 THEN
        LogError "DIALOG.BAS", "Shadow", "lastRow > 25"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF col < 1 THEN
        LogError "DIALOG.BAS", "Shadow", "col < 1"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF lastCol > 80 THEN
        LogError "DIALOG.BAS", "Shadow", "lastCol > 80"
        ERROR cErrorIllegalFunctionCall
    END IF

    FOR r = row TO lastRow
        FOR c = col TO lastCol
           
            charCode = ScreenAscii(r, c)
            charAttribute = ScreenAttribute(r, c)
            fore = charAttribute AND &HF
            back = charAttribute \ &H10

            ' Use intensified black for the brighter color and black for
            ' the other color. Assume that the brighter color is the one
            ' with the highest color code. Although this simple assumption
            ' is not correct for all color combinations, it works OK for
            ' the most likely combinations.
            IF fore >= back THEN
                shadedFore = cHiBlack
                shadedBack = cBlack
            ELSE
                shadedFore = cBlack
                shadedBack = cHiBlack
            END IF

            SetColor shadedFore, shadedBack
            LOCATE r, c, 0
            PRINT CHR$(charCode);
        NEXT
    NEXT

END SUB

SUB ShowTextCursor

    ' Makes the text cursor visible and sets it size as per
    ' the current insert/typeover state.

    IF sfTypeOver THEN
        LOCATE , , 1, cFullStart, cFullStop
    ELSE
        LOCATE , , 1, cUnderlineStart, cUnderlineStop
    END IF

END SUB

