' =============================================================
' This is a special version with the declarations for the
' DropDownListBox and TextBox controls removed to eliminate
' out of memory problems when compiling from the command
' line under Windows 2000 (no problems under Windows 98 SE).
' =============================================================

' DIALOG.BI, v1.1
'
' Copyright (C) 2002 Michael Webster
'
' This program is distributed under the terms of the GNU General
' Public License. Refer to the BAS module for the complete notice.

' This file contains the constant, data, and procedure declarations
' for DIALOG.BAS.

' Declare constants for the 16 text mode colors. Note that these
' constants will not work with the COLOR statement because of the
' special value for black.
CONST cUseDefault = 0
CONST cBlue = 1
CONST cGreen = 2
CONST cCyan = 3
CONST cRed = 4
CONST cMagenta = 5
CONST cYellow = 6
CONST cWhite = 7
CONST cHiBlack = 8
CONST cHiBlue = 9
CONST cHiGreen = 10
CONST cHiCyan = 11
CONST cHiRed = 12
CONST cHiMagenta = 13
CONST cHiYellow = 14
CONST cHiWhite = 15
CONST cBlack = 32

' Declare constants for the default colors.
CONST cWindowFore = cBlack
CONST cWindowBack = cWhite
CONST cTitleBarFore = cBlack
CONST cTitleBarBack = cHiWhite
CONST cCmdFore = cBlack
CONST cCmdBack = cHiWhite
CONST cCmdDisabledFore = cWhite
CONST cHighlightFore = cWhite
CONST cHighlightBack = cBlack
CONST cAccessFore = cHiWhite
CONST cCmdAccessFore = cRed

' Declare constants for the MessageBox style parameter.
CONST cMbOk = 1
CONST cMbOkCancel = 2
CONST cMbRetryCancel = 4
CONST cMbYesNo = 8
CONST cMbYesNoCancel = 16

' Define data types for the dialog components.

TYPE WindowType
    row                 AS INTEGER
    col                 AS INTEGER
    rows                AS INTEGER
    cols                AS INTEGER
    border              AS INTEGER
    fore                AS INTEGER
    back                AS INTEGER
    titleBarFore        AS INTEGER
    titleBarBack        AS INTEGER
    fNoShadow           AS INTEGER
    fNoModColors        AS INTEGER
    fTitleBar           AS INTEGER
END TYPE

' The elements of type ...InternalType are used by the control procedures
' to store control instance data. These elements must be altered only by
' the control procedures and must be initialized to 0 before use.

TYPE CheckBoxInternalType
    tabindex            AS INTEGER
    accessCode          AS INTEGER
    accessKey           AS STRING * 1
    lastCol             AS INTEGER
    fChecked            AS INTEGER
    fMouseLeft          AS INTEGER
END TYPE
TYPE CheckBoxType
    row                 AS INTEGER
    col                 AS INTEGER
    fChecked            AS INTEGER
    internal            AS CheckBoxInternalType
END TYPE

TYPE CommandButtonInternalType
    tabindex            AS INTEGER
    accessCode          AS INTEGER
    accessKey           AS STRING * 1
    lastCol             AS INTEGER
    fActive             AS INTEGER
    fCurrentDefault     AS INTEGER
    fDisabled           AS INTEGER
    fActuated           AS INTEGER
END TYPE
TYPE CommandButtonType
    row                 AS INTEGER
    col                 AS INTEGER
    fClicked            AS INTEGER
    fDefault            AS INTEGER
    fDisabled           AS INTEGER
    fCancel             AS INTEGER
    fUpdate             AS INTEGER
    internal            AS CommandButtonInternalType
END TYPE

TYPE LabelInternalType
    tabindex            AS INTEGER
    accessCode          AS INTEGER
    accessKey           AS STRING * 1
END TYPE
TYPE LabelType
    row                 AS INTEGER
    col                 AS INTEGER
    internal            AS LabelInternalType
END TYPE

TYPE ListBoxInternalType
    tabindex            AS INTEGER
    firstVisibleItem    AS INTEGER
    firstRow            AS INTEGER
    lastRow             AS INTEGER
    firstCol            AS INTEGER
    lastCol             AS INTEGER
    rows                AS INTEGER
    cols                AS INTEGER
    selectedItem        AS INTEGER
    sliderRow           AS INTEGER
    itemCount           AS INTEGER
    prevSelectedItem    AS INTEGER
    fMouseLeft          AS INTEGER
    fMouseLeftOnSlider  AS INTEGER
    fScrollBar          AS INTEGER
    fSelect             AS INTEGER
    timeOut             AS SINGLE
    doubleClickTimeout  AS SINGLE
END TYPE
TYPE ListBoxType
    row                 AS INTEGER
    col                 AS INTEGER
    rows                AS INTEGER
    cols                AS INTEGER
    selectedItem        AS INTEGER
    itemCount           AS INTEGER
    fDoubleClick        AS INTEGER
    fSearch             AS INTEGER
    fText               AS INTEGER
    fUpdate             AS INTEGER
    internal            AS ListBoxInternalType
END TYPE

TYPE OptionButtonInternalType
    tabindex            AS INTEGER
    accessCode          AS INTEGER
    accessKey           AS STRING * 1
    lastCol             AS INTEGER
    fMouseLeft          AS INTEGER
    fSelected           AS INTEGER
END TYPE
TYPE OptionButtonType
    row                 AS INTEGER
    col                 AS INTEGER
    group               AS INTEGER
    fSelected           AS INTEGER
    internal            AS OptionButtonInternalType
END TYPE

' Declare the windowing procedures.
DECLARE SUB CloseWindow ()
DECLARE SUB OpenWindow (win AS WindowType, caption$)

' Declare the control procedures.
DECLARE SUB CpCheckBox (chk AS CheckBoxType, caption$)
DECLARE SUB CpCommandButton (cmd AS CommandButtonType, caption$)
DECLARE SUB CpLabel (lbl AS LabelType, caption$)
DECLARE SUB CpListBox (lst AS ListBoxType, item$())
DECLARE SUB CpOptionButton (opt AS OptionButtonType, caption$)

' Declare the support procedures.
DECLARE FUNCTION ActiveControlIndex% ()
DECLARE SUB EnableHiBackColors ()
DECLARE SUB GetUserInput (key$, keyCode%, mouseRow%, mouseCol%, fMouseLeft%, fMouseRight%, callerIndex%)
DECLARE SUB HideTextCursor ()
DECLARE FUNCTION KeyPressed% (keyCode%)
DECLARE SUB PrintFrame (row%, col%, rows%, cols%, border%, fFill%, caption$)
DECLARE SUB PrintStaticText (row%, col%, text$)
DECLARE FUNCTION ScreenAscii% (row%, col%)
DECLARE FUNCTION ScreenAttribute% (row%, col%)
DECLARE SUB SetAttribute (row%, col%, rows%, cols%, fore%, back%)
DECLARE SUB SetColor (fore%, back%)
DECLARE SUB SetFocus (tabindex%)
DECLARE SUB Shadow (row%, col%, rows%, cols%)
DECLARE SUB ShowTextCursor ()

' Declare the other procedures.
DECLARE FUNCTION MessageBox% (win AS WindowType, style%, default%, caption$, msg$())
DECLARE SUB OpenScreen (fore%, back%, patternCharacter$)
DECLARE SUB PrintDebug (debug$)
DECLARE SUB ProgressBox (row%, col%, cols%, percent%)


