' FONTVIEW.BAS, v1.0
'
' Copyright (C) 2002 Michael Webster
'
' This program is free software; you can redistribute it and/or
' modify it under the terms of the GNU General Public License as
' published by the Free Software Foundation; either version 2 of
' the License, or (at your option) any later version.
'
' This program is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public
' License along with this program; if not, write to the Free
' Software Foundation, Inc., 59 Temple Place, Suite 330,
' Boston, MA  02111-1307  USA

DECLARE SUB ReadBitPatterns (bitPattern%())
DECLARE SUB MainWin ()
DEFINT A-Z

' This file contains a quick and dirty program that allows the user to
' view the character font bit patterns for a VGA subsystem.
'
' This could be expanded to function as a font editor.
'
' Note that the GetFontBlock procedure switches the VGA to graphics
' mode 11h and then back to text mode. Calling this procedure before
' the DIALOG.BAS module is initialized, as is done here, should cause
' no problems. Calling this procedure after the DIALOG.BAS module is
' initialized without taking steps to save and restore the video state
' around the call will disturb the video state and corrupt the display.
' Refer to the SaveVideoState and RestoreVideoState procedures in
' VGA.BAS and to VGATEST.BAS for more information.
'
' This program provides no method of viewing the 9x14 supplement, and
' no method of directly viewing the 9x16 supplement. To indirectly
' view the 9x16 supplement, compare the default character set (which
' includes the 9x16 supplement) to the 8x16 character set.
'
' Note that this program will not display correctly in a window.
'
' Refer the VGA.BAS for more information.

' $INCLUDE: 'INTX.BI'
' $INCLUDE: 'GENERAL.BI'
' $INCLUDE: 'MOUSE.BI'
' $INCLUDE: 'ERROR.BI'
' $INCLUDE: 'DIALOG.BI'
' $INCLUDE: 'VGA.BI'

CONST cFalse = 0, cTrue = NOT cFalse

CONST cGrid0Left = 208
CONST cGrid0Right = 128
CONST cGrid1Left = 209
CONST cGrid1Right = 129

' Force a mode set so the VGA BIOS will load the full default character
' font bit patterns (the 8x16 character font and the 9x16 supplement).
' Note that QB, for the default mode, does not perform a full mode set
' at run time, even if the program includes a SCREEN 0 statement.
SCREEN 11
SCREEN 0

' Declare a shared dynamic array and load it with
' the default character font bit patterns.
DIM SHARED sDefaultBitPatterns(x)
GetFontBlock 0, 16, sDefaultBitPatterns()

' We will use character font block 0 to hold the primary character
' set and character font block 1 to hold the secondary character set.
' Because this program uses no components from DIALOG.BAS that have
' a high-intensity forground, we can assume that DIALOG.BAS will
' always use characters from character set 0.
'
' Load special characters for the character pixel grid into character font
' block 0. The character codes for the special characters avoid any codes
' that DIALOG.BAS might use. The grid elements are one character high and
' two characters wide so the aspect ratio of the grid will reasonably match
' the aspect ratio of the default character box. The character codes for
' the left character of each pair are from the range for which the ninth
' pixel is a copy of the eight pixel, allowing the horizontal lines to be
' continuous across the character pair.
DIM bp(1 TO 16)
RESTORE CharacterData
ReadBitPatterns bp()
LoadUserCharacterSet 0, cGrid0Left, 16, bp()
ReadBitPatterns bp()
LoadUserCharacterSet 0, cGrid0Right, 16, bp()
ReadBitPatterns bp()
LoadUserCharacterSet 0, cGrid1Left, 16, bp()
ReadBitPatterns bp()
LoadUserCharacterSet 0, cGrid1Right, 16, bp()

' We will use character font block 1 to store the bit patterns for the
' character font being viewed. The VGA will use the characters from
' character font block 1 when bit 3 of the character's attribute (which
' controls foreground intensity) is set. To display these characters
' with a black foreground (rather than high-intensity black), we need to
' assign the color black to an otherwise unused high-intensity attribute.
' High-intensity red is a good choice because the mouse driver will
' invert the color to high-intensity cyan, which will provide good
' contrast with the mouse cursor background.
PALETTE cHiRed, 0

MainWin

' Force a mode set so QB will re-program the VGA, overwriting any of
' the non-standard settings that this program uses. Without a forced
' mode set, the next QB program that attempts to use the default mode
' will display a corrupted screen.
SCREEN 11
SCREEN 0

END

CharacterData:

DATA 00000000
DATA 01111111
DATA 01000000
DATA 01000000
DATA 01000000
DATA 01000000
DATA 01000000
DATA 01000000
DATA 01000000
DATA 01000000
DATA 01000000
DATA 01000000
DATA 01000000
DATA 01000000
DATA 01111111
DATA 00000000

DATA 00000000
DATA 11111111
DATA 00000001
DATA 00000001
DATA 00000001
DATA 00000001
DATA 00000001
DATA 00000001
DATA 00000001
DATA 00000001
DATA 00000001
DATA 00000001
DATA 00000001
DATA 00000001
DATA 11111111
DATA 00000000

DATA 00000000
DATA 01111111
DATA 01111111
DATA 01111111
DATA 01111111
DATA 01111111
DATA 01111111
DATA 01111111
DATA 01111111
DATA 01111111
DATA 01111111
DATA 01111111
DATA 01111111
DATA 01111111
DATA 01111111
DATA 00000000

DATA 00000000
DATA 11111111
DATA 11111111
DATA 11111111
DATA 11111111
DATA 11111111
DATA 11111111
DATA 11111111
DATA 11111111
DATA 11111111
DATA 11111111
DATA 11111111
DATA 11111111
DATA 11111111
DATA 11111111
DATA 00000000

SUB MainWin

    DIM regX AS RegTypeX

    OpenScreen 7, 1, ""

    DIM win AS WindowType
    win.rows = 22
    win.cols = 66
    win.fTitleBar = cTrue
    OpenWindow win, "FontView"

    DIM cmdPrevChar AS CommandButtonType
    cmdPrevChar.row = 22
    cmdPrevChar.col = 13

    DIM cmdNextChar AS CommandButtonType
    cmdNextChar.row = 22
    cmdNextChar.col = 36

    DIM cmdClose AS CommandButtonType
    cmdClose.row = 22
    cmdClose.col = 59
    cmdClose.fCancel = cTrue

    DIM optDefault AS OptionButtonType
    optDefault.row = 6
    optDefault.col = 36
    optDefault.fSelected = cTrue
    DIM opt8x8 AS OptionButtonType
    opt8x8.row = 8
    opt8x8.col = 36
    DIM opt8x14 AS OptionButtonType
    opt8x14.row = 10
    opt8x14.col = 36
    DIM opt8x16 AS OptionButtonType
    opt8x16.row = 12
    opt8x16.col = 36

    IF win.fore = cUseDefault THEN winFore = cWindowFore
    IF win.back = cUseDefault THEN winBack = cWindowBack
   
    ' Use character font block 0 for the primary block and character
    ' font block 1 for the secondary block. Because DIALOG.BAS always
    ' uses high-intensity background colors for title bars and command
    ' buttons, we need to leave high intensity enabled.
    SetBlockSpecifier 0, 1, cFalse

    charCode = 0

    ' Set the mouse driver's software text cursor masks to invert
    ' the colors and leave the blink/intensity and the character
    ' code unchanged. This will ensure that the character displayed
    ' in the mouse cursor is from the proper character set. Note
    ' that we must reload the masks after each mode set because the
    ' mode set will cause the mouse driver to revert to the default
    ' masks.
    SetCursorMasks &HFFFF, &H7700

    DO
        CpOptionButton optDefault, "Default"
        CpOptionButton opt8x8, "8x8"
        CpOptionButton opt8x14, "8x14"
        CpOptionButton opt8x16, "8x16"
        CpCommandButton cmdPrevChar, "  &Prev Character  "
        CpCommandButton cmdNextChar, "  &Next Character  "
        CpCommandButton cmdClose, "  Close  "

        IF optDefault.fSelected AND currentTabIndex <> optDefault.internal.tabindex THEN
            REDIM bitPatterns(1 TO 16 * 256)
            FOR i = 1 TO 16 * 256
                bitPatterns(i) = sDefaultBitPatterns(i)
            NEXT
            LoadUserCharacterSet 1, 0, 16, bitPatterns()
            GOSUB UpdateAll
            ShowMouseCursor
            currentTabIndex = optDefault.internal.tabindex
            SetCursorMasks &HFFFF, &H7700
        END IF

        IF opt8x8.fSelected AND currentTabIndex <> opt8x8.internal.tabindex THEN
            GetRomFont c8x8low128, 16, bitPatterns()
            LoadUserCharacterSet 1, 0, 16, bitPatterns()
            GOSUB UpdateAll
            ShowMouseCursor
            currentTabIndex = opt8x8.internal.tabindex
            SetCursorMasks &HFFFF, &H7700
        END IF

        IF opt8x14.fSelected AND currentTabIndex <> opt8x14.internal.tabindex THEN
            GetRomFont c8x14, 16, bitPatterns()
            LoadUserCharacterSet 1, 0, 16, bitPatterns()
            GOSUB UpdateAll
            currentTabIndex = opt8x14.internal.tabindex
            SetCursorMasks &HFFFF, &H7700
        END IF

        IF opt8x16.fSelected AND currentTabIndex <> opt8x16.internal.tabindex THEN
            GetRomFont c8x16, 16, bitPatterns()
            LoadUserCharacterSet 1, 0, 16, bitPatterns()
            GOSUB UpdateAll
            currentTabIndex = opt8x16.internal.tabindex
            SetCursorMasks &HFFFF, &H7700
        END IF

        IF cmdPrevChar.fClicked THEN
            charCode = charCode - 1
            IF charCode < 0 THEN charCode = 255
            GOSUB UpdateGrid
            GOSUB UpdateHighlight
        END IF
      
        IF cmdNextChar.fClicked THEN
            charCode = charCode + 1
            IF charCode > 255 THEN charCode = 0
            GOSUB UpdateGrid
            GOSUB UpdateHighlight
        END IF

        GetMouseStatus mRow, mCol, fLeft, 0
        IF fLeft THEN
            firstRow = 4
            lastRow = firstRow + 255 \ 16
            firstCol = 50
            lastCol = firstCol + 255 MOD 16
            IF mRow >= firstRow AND mRow <= lastRow THEN
                IF mCol >= firstCol AND mCol <= lastCol THEN
                    charCode = (mRow - 4) * 16 + mCol - 50
                    GOSUB UpdateHighlight
                    GOSUB UpdateGrid
                END IF
            END IF
        END IF

    LOOP UNTIL cmdClose.fClicked = cTrue

    SetBlockSpecifier 0, 0, cFalse

    CloseWindow
    CloseWindow

EXIT SUB

WriteChars:
   
    FOR i = 0 TO 255
        r = i \ 16
        c = i MOD 16
        LOCATE r + 4, c + 50
        PRINT " ";
        LOCATE r + 4, c + 50
        WriteCharacter r * 16 + c
    NEXT
    RETURN

UpdateHighlight:
    
    row = charCode \ 16 + 4
    col = charCode MOD 16 + 50
    IF prevRow THEN SetAttribute prevRow, prevCol, 1, 1, cHiRed, winBack
    HideMouseCursorIf row, col, 1, 1
    SetAttribute row, col, 1, 1, cHiRed, cHiWhite
    ShowMouseCursor
    prevRow = row
    prevCol = col
    RETURN

UpdateGrid:

    start = charCode * 16 + 1
    FOR row = 0 TO 15
        byte = bitPatterns(start + row)
        FOR col = 0 TO 7
            bit = BitRangeValue(byte, col, col)
            IF bit = 0 THEN
                pixel$ = CHR$(cGrid0Left) + CHR$(cGrid0Right)
            ELSE
                pixel$ = CHR$(cGrid1Left) + CHR$(cGrid1Right)
            END IF
            r = row + 4
            c = 2 * (8 - col) + 14
            SetColor winFore, winBack
            LOCATE r, c
            HideMouseCursorIf r, c, 1, 2
            HideTextCursor
            PRINT pixel$;
            ShowMouseCursor
        NEXT
    NEXT
    RETURN

UpdateAll:
    
    SetColor cHiRed, winBack
    GOSUB WriteChars
    GOSUB UpdateHighlight
    GOSUB UpdateGrid
    RETURN

END SUB

SUB ReadBitPatterns (bitPattern%())

    ' Loads the elements of <bitPattern%()> with character font bit
    ' patterns that are stored in DATA statements as "binary" string
    ' constants. Each string constant contains the bit pattern for
    ' one scan line. The caller must dimension <bitPattern%()> with
    ' a lower bound of 1 and an upper bound equal to the number of
    ' bytes per character for the font.

    IF LBOUND(bitPattern) <> 1 THEN ERROR cErrorIllegalFunctionCall
    bytesPerCharacter = UBOUND(bitPattern)
    IF bytesPerCharacter > 32 THEN ERROR cErrorIllegalFunctionCall
   
    FOR i = 1 TO bytesPerCharacter
        READ bp$
        bitPattern(i) = Value(bp$)
    NEXT

END SUB

