' GENERAL.BAS, v1.2
'
' Copyright (C) 2002 Michael Webster
'
' This program is free software; you can redistribute it and/or
' modify it under the terms of the GNU General Public License as
' published by the Free Software Foundation; either version 2 of
' the License, or (at your option) any later version.
'
' This program is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public
' License along with this program; if not, write to the Free
' Software Foundation, Inc., 59 Temple Place, Suite 330,
' Boston, MA  02111-1307  USA

DEFINT A-Z

' This file contains a QuickBASIC module consisting of various general-
' purpose procedures.
'
' To use these procedures within the QuickBASIC or QuickBASIC Extended
' programming environment, you must load the default Quick library
' (QB.QLB or QBX.QLB) when the programming environment is started.
' Refer to QUICKLIB.TXT for more information.
'
' IMPORTANT:
'   The FarAddress FUNCTION procedure must be modified for compatibility
'   with QuickBASIC Extended (PDS 7.x) far strings. Refer to comments in
'   the procedure for more information.
'
' Most of the byte and word manipulation procedures in this module
' (HighByte, HighWord, LowWord, MakeInt, and MakeLong) use PEEK
' and/or POKE instead of more traditional methods so they can handle
' a full range of *unsigned* values without complex code.
'
' This module contains several memory related components (MemByte,
' MemWord, MemDword, FarAddress, and GetMemDescriptor) and three
' (or four if you count the GetKeyStatus procedure) BIOS keyboard
' procedures (ReadEnhancedKeyboardStatus, ReadEnhancedKeyboard,
' and PushKey) that were placed in this module rather than in
' MEMORY.BAS or BIOS.BAS to avoid unnecessary dependencies.
'
' ===================================================================
' Changes in v1.1:
'
'   Added the Value% function.
'
'   Renamed TempFileName to TempFileSpec and modified the format of
'   the returned value so the user can reasonably associate the
'   temporary file with the original.
'
'   Fixed an error in the EnTab procedure where it failed to check
'   for a non-zero value in spaceCnt before adding a tab to temp$.
'
' Changes in v1.2:
'
'   Added LogError procedure.
'
'   Added error logging feature to parameter check code.
'
' ===================================================================

' $INCLUDE: 'INTX.BI'
' $INCLUDE: 'ERROR.BI'
' $INCLUDE: 'GENERAL.BI'

CONST cFalse = 0, cTrue = NOT cFalse

DIM SHARED sErrorCode

' *** START TEST CODE ***
CLS
DO
	DO
		ReadEnhancedKeyboardStatus fKeyWaiting, scanCode, asciiCode
	LOOP UNTIL fKeyWaiting
	ReadEnhancedKeyboard scanCode, asciiCode
	LOCATE 1, 1
	PRINT scanCode, asciiCode
LOOP UNTIL asciiCode = 27


S! = TIMER
FOR n& = 1 TO 100000
	i% = RND * 32767
	IF i% < 0 THEN L& = 65536 + i% ELSE L& = i%
NEXT
F! = TIMER
PRINT F! - S!
S! = TIMER
FOR n& = 1 TO 100000
	i% = RND * 32767
	IF i% AND 32768 THEN L& = 65536 + i% ELSE L& = i%
	'IF I% < 0 THEN L& = 65536 + I% ELSE L& = I%
NEXT
F! = TIMER
PRINT F! - S!
S! = TIMER
FOR n& = 1 TO 100000
	i% = RND * 32767
	POKE VARPTR(L&), PEEK(VARPTR(i%))
	POKE VARPTR(L&) + 1, PEEK(VARPTR(i%) + 1)
NEXT
F! = TIMER
PRINT F! - S!
END


StartTimer
FOR n& = 1 TO 100000
	i = RND * 32768
	L& = UInt(i)
NEXT
PRINT ElapsedTime
' *** END TEST CODE ***

END

GeneralErrorHandler:
	sErrorCode = ERR
	RESUME NEXT

SUB AddListItem (item$(), newItem$)

	' Increments the number of elements allocated for <items$()> and
	' copies <newItem$> to the last element. Note that the caller
	' must declare <item$()> as an empty $DYNAMIC array and that this
	' procedure sets the lower bound of <item$()> to 1.
	'
	' If <item$()> is declared in a procedure without a STATIC attribute,
	' then the declaration can be of the form DIM arrayname$(0). If the
	' default lower bound for array subscripts has been changed with
	' OPTION BASE, then use DIM arrayname$(0 TO 0). If <item$()> is
	' declared in module level code, or in a procedure with a STATIC
	' attribute, then use the $DYNAMIC metacommand or dimension it with
	' variable subscripts to ensure that it will be a $DYNAMIC array.
	' Refer to the on-line help for more information.
	'
	' This procedure uses QB code to effectively duplicate the function
	' of the ReDim Preserve statement supported by VB.

	lb = LBOUND(item$)
	ub = UBOUND(item$)

	IF lb = 0 AND ub = 0 THEN
		REDIM item$(1 TO 1)
		item$(1) = newItem$
		EXIT SUB
	END IF

	DIM temp$(lb TO ub)
	FOR i = lb TO ub
		temp$(i) = item$(i)
	NEXT
	REDIM item$(lb TO ub + 1)
	FOR i = lb TO ub
		item$(i) = temp$(i)
	NEXT
	item$(ub + 1) = newItem$
	ERASE temp$

END SUB

FUNCTION AltPlusKeyCode% (key$) STATIC
   
	' Returns the same value that the GetKey procedure would return in
	' its keyCode argument if the user pressed Alt + the alpha-numeric
	' key specified by <key$>.
	'
	' Returns 0 if <key$> does not specify an alpha-numeric key.

	IF NOT fLoaded THEN
		DIM altPlusKeyCodes(1 TO 36)
		altPlusKeyCodes(1) = 376
		altPlusKeyCodes(2) = 377
		altPlusKeyCodes(3) = 378
		altPlusKeyCodes(4) = 379
		altPlusKeyCodes(5) = 380
		altPlusKeyCodes(6) = 381
		altPlusKeyCodes(7) = 382
		altPlusKeyCodes(8) = 383
		altPlusKeyCodes(9) = 384
		altPlusKeyCodes(10) = 385
		altPlusKeyCodes(11) = 272
		altPlusKeyCodes(12) = 273
		altPlusKeyCodes(13) = 274
		altPlusKeyCodes(14) = 275
		altPlusKeyCodes(15) = 276
		altPlusKeyCodes(16) = 277
		altPlusKeyCodes(17) = 278
		altPlusKeyCodes(18) = 279
		altPlusKeyCodes(19) = 280
		altPlusKeyCodes(20) = 281
		altPlusKeyCodes(21) = 286
		altPlusKeyCodes(22) = 287
		altPlusKeyCodes(23) = 288
		altPlusKeyCodes(24) = 289
		altPlusKeyCodes(25) = 290
		altPlusKeyCodes(26) = 291
		altPlusKeyCodes(27) = 292
		altPlusKeyCodes(28) = 293
		altPlusKeyCodes(29) = 294
		altPlusKeyCodes(30) = 300
		altPlusKeyCodes(31) = 301
		altPlusKeyCodes(32) = 302
		altPlusKeyCodes(33) = 303
		altPlusKeyCodes(34) = 304
		altPlusKeyCodes(35) = 305
		altPlusKeyCodes(36) = 306
		fLoaded = cTrue
	END IF

	IF key$ <> "" THEN
		i = INSTR("1234567890QWERTYUIOPASDFGHJKLZXCVBNM", UCASE$(key$))
	END IF

	IF i THEN
		AltPlusKeyCode = altPlusKeyCodes(i)
	END IF

END FUNCTION

FUNCTION Bin$ (i%)

	' Returns a string that represents the value of <i%> as a
	' 16-digit binary number (0's and 1's).

	' Convert <i%> to an unsigned value so we can avoid having
	' to deal with the sign in the subtraction operation below.
	i& = UInt(i)

	FOR j = 15 TO 0 STEP -1
		IF i& \ 2 ^ j THEN
			b$ = b$ + "1"
			i& = i& - 2 ^ j
		ELSE
			b$ = b$ + "0"
		END IF
	NEXT
	Bin$ = b$

END FUNCTION

FUNCTION BitRangeValue% (i%, lowBit%, highBit%)

	' Extracts bits <lowBit%> to <highBit%> from <i%> and
	' returns the value of the extracted bits.
	'
	' Note that the extracted bits are shifted to the least
	' significant position of the return value (the least
	' significant bit is bit 0).

	IF lowBit > highBit THEN SWAP lowBit, highBit
	temp = i
	ShiftRight temp, lowBit
	mask& = 2 ^ (highBit - lowBit + 1) - 1

	BitRangeValue = LowWord(temp AND mask&)

END FUNCTION

FUNCTION Choice (prompt$, keys$, default%)

	' Prints a prompt, a list of choices, and a default choice, and waits
	' for the user to press a key. Returns the default choice if the user
	' pressed the Enter key or Spacebar. Returns the selected choice if
	' the user pressed a valid choice key. Produces a mild warning beep
	' if the user presses any other key.

	IF default > LEN(keys$) THEN
		LogError "GENERAL.BAS", "Choice", "default > LEN(keys$)"
		ERROR cErrorIllegalFunctionCall
	END IF

	keys$ = UCASE$(keys$)

	IF default THEN defaultKey$ = MID$(keys$, default, 1)

	PRINT prompt$; " [";
	FOR i = 1 TO LEN(keys$)
		PRINT MID$(keys$, i, 1);
		IF i < LEN(keys$) THEN PRINT "/";
	NEXT
	PRINT "]: "; defaultKey$;
	
	IF default THEN
		LOCATE , POS(0) - 1, 1, cFullStart, cFullStop
	ELSE
		LOCATE , , 1, cUnderlineStart, cUnderlineStop
	END IF

	DO
		GetKey key$, keyCode
		SELECT CASE keyCode
			CASE 0
				' Do nothing - no key was pressed.

			CASE cKeyEnter, cKeySpace
				Choice = default
				EXIT DO

			CASE ELSE
				position = INSTR(keys$, UCASE$(key$))
				IF position THEN
					PRINT key$;
					Choice = position
					EXIT DO
				ELSE
					SOUND 100, 1
				END IF
		END SELECT
	LOOP
	
	LOCATE , , 0, cUnderlineStart, cUnderlineStop

END FUNCTION

FUNCTION CtrlKey%

	' Returns true if either Ctrl key is pressed.
	'
	' Reads the Ctrl key status from the Keyboard
	' Shift Flags Byte in the BIOS Data Area.

	DEF SEG = &H40
	shiftFlagsByte = PEEK(&H17)
	IF shiftFlagsByte AND 4 THEN CtrlKey = cTrue
	DEF SEG

END FUNCTION

SUB Delay (milliseconds%) STATIC

	' Delays for <milliseconds%> milliseconds before returning.
	'
	' This procedure uses system timer 2 to produce a short programmable
	' delay. This method was chosen because it produces better accuracy
	' than would be possible with the TIMER function and because it does
	' not depend on a hardware interrupt (so it will continue to function
	' with the system timer tick interrupt disabled).
	'
	' IMPORTANT:
	'   Because system timer 2 is normally used to produce speaker
	'   sounds, this procedure will conflict with the QB SOUND and
	'   PLAY statements.

	' Exit now if <milliseconds%> is 0. The timer decrements the count
	' value before testing it for 0 so a starting value of 0 would cause
	' the timer to cycle through the full range.
	IF milliseconds = 0 THEN EXIT SUB
  
	' Set the gate for timer 2 (bit 0 at I/O port 61h) to off. To avoid
	' changing other bits in the register, this statement reads the
	' current value, sets bit 0, and writes the altered value back.
	OUT &H61, INP(&H61) AND NOT 1

	' Program timer 2 for LSB then MSB, mode 0, binary. The system timer
	' control word register is at I/O port 43h. The system timer control
	' word is set as follows:
	'   bit 7-6: 10     = timer 2
	'   bit 5-4: 11     = R/W LSB then MSB
	'   bit 3-1: 000    = single timeout
	'   bit 0:   0      = binary
	OUT &H43, &HB0

	' Calculate the number of full range cycles and the count for the
	' last cycle. Each full range cycle has a duration of ~54.9 ms.
	totalCount& = milliseconds * 1193&
	cycleCount = totalCount& \ 65536
	lastCount& = totalCount& MOD 65536

	' Do the full range cycles.
	FOR i = 1 TO cycleCount

		' Load the starting value, LSB then MSB.
		OUT &H42, 0
		OUT &H42, 0

		' Set the gate for timer 2 (bit 0 at I/O port 61h) to on.
		OUT &H61, INP(&H61) OR 1

		' Wait until the output bit (bit 5 at I/O port 61h) is set.
		WAIT &H61, &H20
	NEXT

	' Do the last cycle.
	IF lastCount& THEN
   
		' Set the gate for timer 2 (bit 0 at I/O port 61h) to off.
		OUT &H61, INP(&H61) AND NOT 1
	
		' Load the starting value, LSB then MSB.
		OUT &H42, LowByte(LowWord(lastCount&))
		OUT &H42, HighByte(LowWord(lastCount&))

		' Set the gate for timer 2 to on.
		OUT &H61, INP(&H61) OR 1

		' Wait until the output bit (bit 5 at I/O port 61h) is set.
		WAIT &H61, &H20
	END IF

	' Set the gate for timer 2 (bit 0 at I/O port 61h) to off.
	OUT &H61, INP(&H61) AND NOT 1

END SUB

FUNCTION ElapsedTime!

	' Ends the current timing cycle and returns the time, in seconds,
	' since the most recent call to the StartTimer procedure.

	finish! = TIMER
	ElapsedTime! = finish! - gStartTime!
   
END FUNCTION

SUB EnTab (line$, tabWidth%)

	' Replaces runs of spaces in <line$> with tabs using <tabWidth%>
	' to determine the location of the tab stops.

	' Expand any existing tabs.
	UnTab line$, tabWidth%

	FOR i = 1 TO LEN(line$)
		c$ = MID$(line$, i, 1)
		IF c$ = " " THEN
			IF i MOD tabWidth = 0 AND spaceCnt THEN
				temp$ = temp$ + CHR$(9)
				spaceCnt = 0
			ELSE
				spaceCnt = spaceCnt + 1
			END IF
		ELSE
			IF spaceCnt THEN
				temp$ = temp$ + SPACE$(spaceCnt)
				spaceCnt = 0
			END IF
			temp$ = temp$ + c$
		END IF
	NEXT
   
	line$ = temp$

END SUB

FUNCTION FarAddress& (S$)

	' Returns a far address (a 32-bit memory address) that has the
	' segment address of <s$> in the most significant word and the
	' offset address of <s$> in the least significant word.
	'
	' The use of this procedure to get the address of a variable-
	' length string limits the modifications required to switch
	' between QuickBASIC near strings and QuickBASIC Extended
	' (PDS 7.x) far strings to this one procedure. To switch to
	' far strings, comment out the next 14 statements and un-
	' comment the last statement.
	
	' Switch to screen page 0, print an error message, and exit
	' if we are running under PDS 7.x. QuickBASIC should return
	' the size of the next free block of string storage for
	' FRE(-4), and PDS 7.x should return an Illegal function call
	' error. Note that this has not been tested under PDS 7.1.
	ON ERROR GOTO GeneralErrorHandler
	i& = FRE(-4)
	IF sErrorCode THEN
		SCREEN , , 0, 0
		CLS
		PRINT
		PRINT "The FarAddress procedure in GENERAL.BAS must be modified for"
		PRINT "use with QuickBASIC Extended (PDS 7.x) and far strings."
		Pause "exit"
		END
	END IF
	ON ERROR GOTO 0
	
	' The following two statements are for QuickBASIC near strings.
	' Under PDS 7.x with far strings, these statements will silently
	' return the wrong address.
	MakeLong i&, VARSEG(S$), SADD(S$)
	FarAddress& = i&

	' The following statement is for PDS 7.x far strings. Note that
	' far strings are always used by programs running in the QBX
	' programming environment, and by programs that were compiled
	' with a Quick library loaded.
	'FarAddress& = SSEGADD(s$)

END FUNCTION

SUB GetKey (key$, keyCode%) STATIC

	' Reads a character from the keyboard. Returns with <key$> set to
	' the key character and <keyCode%> set to a value that identifies
	' the key or key combination, or to 0 if no key is available.
	' Refer to GENERAL.BI for more information.
	'
	' On systems that support the BIOS enhanced keyboard functions
	' (which means virtually all systems produced after the late
	' 1980s), this procedure gets keystroke information by calling
	' these functions. This allows it to return keystroke information
	' for keys and key combinations that the normal BIOS keyboard
	' functions (and the INKEY$ function) would ignore.
	'
	' The code that determines if the BIOS supports the enhanced
	' keyboard functions was adapted from the KEYTYPE routine in
	' Frank van Gilluwe's book _The_Undocumented_PC_ (first
	' edition), Addison-Wesley 1994. ISBN 0-201-62277-7.

	IF NOT fKeyboardChecked THEN

		' Determine if the BIOS supports the enhanced keyboard functions.

		' Empty the keyboard buffer.
		DO
		LOOP WHILE INKEY$ <> ""

		' Push an invalid key into the keyboard buffer.
		scanCode = &HFF
		asciiCode = &HFF
		PushKey scanCode, asciiCode, fBufferFull

		' If the above call did not fail...
		IF NOT fBufferFull THEN

			' If we can read the key just pushed with the
			' BIOS Read Enhanced Keyboard functions, then
			' they must be supported.
			FOR i = 1 TO 16
				ReadEnhancedKeyboardStatus fKeyWaiting, scanCode, asciiCode
				IF NOT fKeyWaiting THEN EXIT FOR
				ReadEnhancedKeyboard scanCode, asciiCode
				IF scanCode = &HFF AND asciiCode = &HFF THEN
					fExtendedKeyboard = cTrue
					EXIT FOR
				END IF
			NEXT
		END IF
		fKeyboardChecked = cTrue
	END IF
	
	IF fExtendedKeyboard THEN
		ReadEnhancedKeyboardStatus fKeyWaiting, scanCode, asciiCode
		IF NOT fKeyWaiting THEN
			key$ = ""
			keyCode = 0
			EXIT SUB
		END IF
		ReadEnhancedKeyboard scanCode, asciiCode
		IF asciiCode = 0 OR asciiCode = &HE0 THEN
			key$ = CHR$(0) + CHR$(scanCode)
			keyCode = scanCode + 256
		ELSE
			key$ = CHR$(asciiCode)
			keyCode = asciiCode
		END IF
	ELSE
		key$ = INKEY$
		SELECT CASE LEN(key$)
			CASE 0
				keyCode = 0
			CASE 1
				keyCode = ASC(key$)
			CASE 2
				keyCode = ASC(RIGHT$(key$, 1)) + 256
		END SELECT
	END IF

END SUB

SUB GetKeyStatus (key$, keyCode%) STATIC

	' Reads the keyboard status. If a key is available in the
	' keyboard buffer, then this procedure returns with <key$>
	' set to the key character and <keyCode%> set to a value
	' that identifies the key or key combination. Otherwise,
	' it returns with <keyCode%> set to 0. Refer to GENERAL.BI
	' for more information.
	'
	' This procedure returns the same values as the GetKey
	' procedure, but does NOT remove the key from the keyboard
	' buffer.
	'
	' On systems that support the BIOS enhanced keyboard functions
	' (which means virtually all systems produced after the late
	' 1980s), this procedure calls the BIOS Read Enhanced Keyboard
	' Status function. This allows it to return keystroke
	' information for keys and key combinations that the normal
	' BIOS Read Keyboard Status function would ignore.
	'
	' The code that determines if the BIOS supports the enhanced
	' keyboard functions was adapted from the KEYTYPE routine in
	' Frank van Gilluwe's book _The_Undocumented_PC_ (first
	' edition), Addison-Wesley 1994. ISBN 0-201-62277-7.

	DIM regX AS RegTypeX

	IF NOT fKeyboardChecked THEN

		' Determine if the BIOS supports the enhanced keyboard functions.

		' Empty the keyboard buffer.
		DO
		LOOP WHILE INKEY$ <> ""

		' Push an invalid key into the keyboard buffer.
		scanCode = &HFF
		asciiCode = &HFF
		PushKey scanCode, asciiCode, fBufferFull

		' If the above call did not fail...
		IF NOT fBufferFull THEN

			' If we can read the key just pushed with the
			' BIOS Read Enhanced Keyboard functions, then
			' they must be supported.
			FOR i = 1 TO 16
				ReadEnhancedKeyboardStatus fKeyWaiting, scanCode, asciiCode
				IF NOT fKeyWaiting THEN EXIT FOR
				ReadEnhancedKeyboard scanCode, asciiCode
				IF scanCode = &HFF AND asciiCode = &HFF THEN
					fExtendedKeyboard = cTrue
					EXIT FOR
				END IF
			NEXT
		END IF
		fKeyboardChecked = cTrue
	END IF

	IF fExtendedKeyboard THEN
		ReadEnhancedKeyboardStatus fKeyWaiting, scanCode, asciiCode
		IF NOT fKeyWaiting THEN
			key$ = ""
			keyCode = 0
		ELSE
			IF asciiCode = 0 OR asciiCode = &HE0 THEN
				key$ = CHR$(0) + CHR$(scanCode)
				keyCode = scanCode + 256
			ELSE
				key$ = CHR$(asciiCode)
				keyCode = asciiCode
			END IF
		END IF
	ELSE
		' Call the normal BIOS Read Keyboard Status Function.
		regX.ax = &H100
		InterruptX &H16, regX, regX
		IF regX.flags AND cZeroFlag THEN
			key$ = ""
			keyCode = 0
		ELSE
			scanCode = HighByte(regX.ax)
			asciiCode = LowByte(regX.ax)
			IF asciiCode = 0 THEN
				keyCode = scanCode + 256
				key$ = CHR$(scanCode)
			ELSE
				keyCode = asciiCode
				key$ = CHR$(asciiCode)
			END IF
		END IF
	END IF

END SUB

SUB GetMemDescriptor (memDescriptor AS MemDescriptorType, S$)

	' Returns with the elements of <memDescriptor> set to the address
	' and length of <s$>.
	'
	' Note that the FarAddress FUNCTION procedure must be modified for
	' compatibility with QuickBASIC Extended (PDS 7.x) far strings.
	' Refer to comments in the procedure for more information.

	memDescriptor.length = LEN(S$)
	fa& = FarAddress(S$)
	memDescriptor.segment = HighWord(fa&)
	memDescriptor.offset = LowWord(fa&)
	
END SUB

FUNCTION GroupedDigits$ (digits$)

	' Returns <digits$> with commas inserted between each 3 digit group
	' and the 1-3 digit group preceeding it.
	'
	' Assumes that <digits$> will contain no characters other than the
	' decimal digits 0-9 and leading spaces.
	'
	' Unlike the PRINT USING statement, this procedure can be used to
	' display a numeric value without leading spaces.
	
	d$ = LTRIM$(digits$)
	dLen = LEN(d$)
	FOR i = dLen TO 1 STEP -1
		gd$ = MID$(d$, i, 1) + gd$
		revPos = dLen - i + 1
		IF (i > 1) AND (revPos MOD 3 = 0) THEN gd$ = "," + gd$
	NEXT
	GroupedDigits$ = gd$

END FUNCTION

FUNCTION HexByte$ (i%)
	
	' Returns the least significant byte of <i%> as a 2-digit hex string.

	h$ = HEX$(LowByte(i))
	HexByte$ = STRING$(2 - LEN(h$), "0") + h$

END FUNCTION

FUNCTION HexWord$ (i&)

	' Returns the least significant word of <i&> as a 4-digit hex string.
	'
	' For an INTEGER argument, use HexWord$(UInt(i%)).

	h$ = HEX$(LowWord(i&))
	HexWord$ = STRING$(4 - LEN(h$), "0") + h$

END FUNCTION

FUNCTION HighByte% (i%)
  
	' Returns the most significant byte of <i%>.

	' Make sure the current segment is the default data segment.
	DEF SEG
  
	HighByte = PEEK(VARPTR(i) + 1)

END FUNCTION

FUNCTION HighWord% (i&)

	' Returns the most significant word of <i&>.

	' Make sure the current segment is the default data segment.
	DEF SEG
 
	POKE VARPTR(HighWord), PEEK(VARPTR(i&) + 2)
	POKE VARPTR(HighWord) + 1, PEEK(VARPTR(i&) + 3)

END FUNCTION

SUB InputString (prompt$, in$)
  
	' Prints a prompt and allows the user to input/edit a single line of
	' characters.
	'
	' The input/edit field begins immediately after the prompt on the
	' line where the prompt ends and extends to column 80 on that same
	' line. The input/edit field is scrolled horizontally as necessary
	' to accommodate a maximum input string length of 32,767 characters.
	' The editing features are similar to those of the INPUT statement
	' (without the WordStar keys, etc). Returns with <in$> set to the
	' current value of the input string when the user presses Enter.
	' Clears the entire input string, if not clear, or returns with
	' <in$> set to a null string when the user presses Escape.
	'
	' The insert/typeover state is maintained locally and is initialized
	' to insert on each call.
	'
	' This procedure, unlike the INPUT and LINE INPUT statements, uses
	' the initial value of <in$> as a default input string.

	PRINT
	PRINT prompt$;
	firstCol = POS(0)

	' Avoid conditions where either the input
	' or the prompt would not be visible.
	IF firstCol > 79 THEN
		firstCol = 1
		LOCATE CSRLIN + 1, firstCol
	END IF
	
	length = 80 - firstCol
	temp$ = in$
	firstVisibleChar = 1
	lastCol = 80
	cursorCol = firstCol

	DO
		LOCATE , firstCol, 0
		tempLength = LEN(temp$)
		PRINT MID$(temp$, firstVisibleChar, length);

		' If the text cursor is not at the end of the input/edit
		' field, then fill the remainder of the input/edit field
		' with spaces to overwrite any old characters.
		cursorColTemp = POS(0)
		IF cursorColTemp <= lastCol THEN PRINT SPACE$(lastCol - cursorColTemp + 1);

		IF fTypeOver THEN
			LOCATE , cursorCol, 1, cFullStart, cFullStop
		ELSE
			LOCATE , cursorCol, 1, cUnderlineStart, cUnderlineStop
		END IF

		DO
			GetKey key$, keyCode
		LOOP UNTIL keyCode

		SELECT CASE keyCode
			CASE cKeyBackspace
				IF cursorCol > firstCol OR firstVisibleChar > 1 THEN
					leftPart$ = LEFT$(temp$, firstVisibleChar + cursorCol - firstCol - 2)
					rightPart$ = MID$(temp$, firstVisibleChar + cursorCol - firstCol)
					temp$ = leftPart$ + rightPart$
					IF cursorCol > firstCol THEN
						cursorCol = cursorCol - 1
					ELSEIF firstVisibleChar > 1 THEN
						firstVisibleChar = firstVisibleChar - 1
					END IF
				END IF

			CASE cKeyEnter
				EXIT DO

			CASE cKeyEscape
				IF temp$ = "" THEN
					EXIT DO
				ELSE
					temp$ = ""
					cursorCol = firstCol
					firstVisibleChar = 1
				END IF

			CASE cKeyHome
				cursorCol = firstCol
				firstVisibleChar = 1

			CASE cKeyLeft
				IF cursorCol > firstCol THEN
					cursorCol = cursorCol - 1
				ELSEIF firstVisibleChar > 1 THEN
					firstVisibleChar = firstVisibleChar - 1
				END IF

			CASE cKeyRight
				IF cursorCol < lastCol THEN
					cursorCol = cursorCol + 1
				ELSE
					firstVisibleChar = firstVisibleChar + 1
				END IF

			CASE cKeyEnd
				IF tempLength > length THEN
					cursorCol = lastCol
					firstVisibleChar = tempLength - length + 1
				ELSE
					cursorCol = firstCol + tempLength
				END IF

			CASE cKeyInsert
				fTypeOver = NOT fTypeOver

			CASE cKeyDelete
				leftPart$ = LEFT$(temp$, firstVisibleChar + cursorCol - firstCol - 1)
				rightPart$ = MID$(temp$, firstVisibleChar + cursorCol - firstCol + 1)
				temp$ = leftPart$ + rightPart$

			CASE 32 TO 255
				' If the cursor is past the end of temp$,
				' then add spaces to fill the gap.
				spaceCount = firstVisibleChar + cursorCol - firstCol - tempLength - 1
				IF spaceCount > 0 THEN
					temp$ = temp$ + SPACE$(spaceCount)
				END IF

				leftPart$ = LEFT$(temp$, firstVisibleChar + cursorCol - firstCol - 1)
				IF fTypeOver THEN
					rightPart$ = MID$(temp$, firstVisibleChar + cursorCol - firstCol + 1)
				ELSE
					rightPart$ = MID$(temp$, firstVisibleChar + cursorCol - firstCol)
				END IF
				temp$ = leftPart$ + key$ + rightPart$
				
				IF cursorCol < lastCol THEN
					cursorCol = cursorCol + 1
				ELSE
					firstVisibleChar = firstVisibleChar + 1
				END IF
		END SELECT
	LOOP
	
	in$ = temp$
	
	LOCATE , , 0, cUnderlineStart, cUnderlineStop

END SUB

FUNCTION InStrL (string1$, string2$)

	' Returns the character position of the last occurrence of <string2$>
	' in <string1$), or 0 if there are no occurrences of <string2$> in
	' <string1$>.

	FOR i = LEN(string1$) TO 1 STEP -1
		IF MID$(string1$, i, LEN(string2$)) = string2$ THEN
			InStrL = i
			EXIT FUNCTION
		END IF
	NEXT

END FUNCTION

FUNCTION IsVGA
  
	' Returns true if the active display adapter is VGA compatible.
	'
	' Checks for a VGA compatible display adapter by calling the VGA
	' BIOS Read Display Combination Code function.

	DIM regX AS RegTypeX
  
	regX.ax = &H1A00
	InterruptX &H10, regX, regX

	' If a VGA compatible display adapter is present, then the function
	' should return with AL = 1Ah and the code for the active display
	' adapter in BL. Because a VGA adapter can coexist with other display
	' adapters, we need to verify that the code for the active adapter
	' is 7 (VGA with analog monochrome) or 8 (VGA with analog color).
	IF (regX.ax AND &HFF) = &H1A AND (regX.bx AND &HFF) = 7 OR (regX.bx AND &HFF) = 8 THEN
		IsVGA = cTrue
	END IF

END FUNCTION

SUB LogError (module$, procedure$, error$)

	' Copies the arguments, formatted as comments, to ERRORLOG.BAS.
	'
	' Whenever a procedure in any of the modules that are likley to
	' be used as part of a Quick library detects a parameter error,
	' it calls this procedure with the arguments set to values that
	' indicate the location and the specific cause of the error,
	' before it simulates an Illegal function call error.
	'
	' Because the debugger in the programming environment cannot
	' trace into library code, when a procedure in a Quick library
	' generates an error, the programming environment is limited to
	' highlighting the line where the procedure was called and
	' displaying a standard QB error message (Illegal function call
	' in this case). To get more specific information on the error,
	' load ERRORLOG.BAS into the programming environment and view
	' the contents.
	'
	' Note that, for subsequent errors, you must unload and reload
	' ERRORLOG.BAS to refresh the contents.

	file = FREEFILE
	OPEN "errorlog.bas" FOR OUTPUT AS file
	PRINT #file, "' "; module$
	PRINT #file, "' "; procedure$
	PRINT #file, "' "; error$
	CLOSE file

END SUB

FUNCTION LowByte% (i%)
  
	' Returns the least significant byte of <i%>.

	LowByte = i AND &HFF

END FUNCTION

FUNCTION LowWord% (i&)

	' Returns the least significant word of <i&>.

	' Make sure the current segment is the default data segment.
	DEF SEG
  
	POKE VARPTR(LowWord), PEEK(VARPTR(i&))
	POKE VARPTR(LowWord) + 1, PEEK(VARPTR(i&) + 1)

END FUNCTION

SUB MakeInt (i%, hiByte%, loByte%)

	' Returns with <i%> set to the integer value that has
	' <hiByte%> in the most significant byte and <loByte%>
	' in the least significant byte.
  
	' Make sure the current segment is the default data segment.
	DEF SEG

	POKE VARPTR(i), loByte
	POKE VARPTR(i) + 1, hiByte

END SUB

SUB MakeLong (i&, hiWord&, loWord&)

	' Returns with <i&> set to the long integer value that has
	' <hiWord&> in the most significant word and <loWord&> in
	' the least significant word.

	' Make sure the current segment is the default data segment.
	DEF SEG

	POKE VARPTR(i&), PEEK(VARPTR(loWord&))
	POKE VARPTR(i&) + 1, PEEK(VARPTR(loWord&) + 1)
	POKE VARPTR(i&) + 2, PEEK(VARPTR(hiWord&))
	POKE VARPTR(i&) + 3, PEEK(VARPTR(hiWord&) + 1)

END SUB

FUNCTION MemByte% (segment&, offset&)

	' Returns the byte value stored in memory at the
	' location specified by <segment&>:<offset&>.

	DEF SEG = segment&

	MemByte = PEEK(offset&)
   
	DEF SEG

END FUNCTION

FUNCTION MemDword& (segment&, offset&)

	' Returns the 32-bit value (double word) stored in memory
	' at the location specified by <segment&>:<offset&>.

	DEF SEG = segment&

	b0 = PEEK(offset& + 0)
	b1 = PEEK(offset& + 1)
	b2 = PEEK(offset& + 2)
	b3 = PEEK(offset& + 3)

	DEF SEG

	POKE VARPTR(MemDword) + 0, b0
	POKE VARPTR(MemDword) + 1, b1
	POKE VARPTR(MemDword) + 2, b2
	POKE VARPTR(MemDword) + 3, b3

END FUNCTION

FUNCTION MemWord% (segment&, offset&)

	' Returns the 16-bit value stored in memory at the
	' location specified by <segment&>:<offset&>.

	DEF SEG = segment&

	b0 = PEEK(offset& + 0)
	b1 = PEEK(offset& + 1)

	DEF SEG

	POKE VARPTR(MemWord) + 0, b0
	POKE VARPTR(MemWord) + 1, b1

END FUNCTION

SUB Pause (action$)

	PRINT
	PRINT "Press any key to "; action$; "...";
	DO
	LOOP UNTIL INKEY$ <> ""

END SUB

SUB PushKey (scanCode%, asciiCode%, fBufferFull%)

	' Calls the BIOS Push Character and Scan Code function.
	' Stores the key data specified by <scanCode%> and
	' <asciiCode%> into the keyboard buffer. Returns with
	' <fBufferFull%) set to true if the keyboard buffer is
	' full.
	'
	' The scan code and ascii code values for a particular
	' key can be obtained from available BIOS references,
	' or by running KEYCODE.BAS.
	'
	' This procedure was moved from the BIOS.BAS module to
	' this module to avoid unnecessary module dependencies.

	DIM regX AS RegTypeX

	regX.ax = &H500
	MakeInt regX.cx, scanCode, asciiCode
	InterruptX &H16, regX, regX
	IF LowByte(regX.ax) THEN fBufferFull = cTrue

END SUB

SUB ReadEnhancedKeyboard (scanCode%, asciiCode%)

	' Calls the BIOS Read Enhanced Keyboard function. If a key is
	' available in the keyboard buffer, then this procedure returns
	' with the key scan code in <scanCode%> and the key character
	' code in <asciiCode%>. If no key is available, then the BIOS
	' will wait (indefinitely) until a key is available.
	'
	' To avoid the wait, you should first determine if a key is
	' available by calling the ReadEnhancedKeyboardStatus procedure.
	'
	' Note that the BIOS removes the returned key from the keyboard
	' buffer.
	'
	' This procedure was moved from the BIOS.BAS module to this
	' module to avoid unnecessary module dependencies.

	DIM regX AS RegTypeX

	regX.ax = &H1000
	InterruptX &H16, regX, regX
	scanCode = HighByte(regX.ax)
	asciiCode = LowByte(regX.ax)

END SUB

SUB ReadEnhancedKeyboardStatus (fKeyWaiting%, scanCode%, asciiCode%)

	' Calls the BIOS Read Enhanced Keyboard Status function. If a
	' key is available in the keyboard buffer, then this procedure
	' returns with <fKeyWaiting%> set to true, the key scan code
	' in <scanCode%>, and the key character code in <asciiCode%>.
	' If no key is available, then this procedure returns with
	' <fKeyWaiting%> set to false.
	'
	' Note that the BIOS does NOT remove the returned key from the
	' keyboard buffer.
	'
	' This procedure was moved from the BIOS.BAS module to this
	' module to avoid unnecessary module dependencies.
  
	DIM regX AS RegTypeX

	regX.ax = &H1100
	InterruptX &H16, regX, regX
	IF regX.flags AND cZeroFlag THEN
		fKeyWaiting = cFalse
		scanCode = 0
		asciiCode = 0
	ELSE
		fKeyWaiting = cTrue
		scanCode = HighByte(regX.ax)
		asciiCode = LowByte(regX.ax)
	END IF

END SUB

FUNCTION ShiftKey%

	' Returns true if either Shift key is pressed.
	'
	' Reads the Shift key status from the Keyboard
	' Shift Flags Byte in the BIOS Data Area.

	DEF SEG = &H40
	shiftFlagsByte = PEEK(&H17)
	IF shiftFlagsByte AND 1 OR shiftFlagsByte AND 2 THEN ShiftKey = cTrue
	DEF SEG

END FUNCTION

SUB ShiftLeft (i%, n%)

	' Returns with <i%> shifted left by <n%> bit positions.
	'
	' Note that <i%> is manipulated as an unsigned integer - no
	' attempt is made to preserve the sign.

	' If <n%> is greater the 15, then the return value must be 0.
	IF n > 15 THEN
		i = 0
		EXIT SUB
	END IF

	' Copy <i%> to a long integer so we can avoid overflow problems
	' in the multiply operation below.
	i& = i

	FOR j = 1 TO n
		i& = i& * 2
	NEXT
   
	i = LowWord(i&)

END SUB

SUB ShiftRight (i%, n%)

	' Returns with <i%> shifted right by <n%> bit positions.
	'
	' Note that <i%> is manipulated as an unsigned integer - empty
	' bit positions at the left end of <i%> are filled with zeros.

	' If <n%> is greater the 15, then the return value must be 0.
	IF n > 15 THEN
		i = 0
		EXIT SUB
	END IF
	
	i& = UInt(i)
	
	FOR j = 1 TO n
		i& = i& \ 2
	NEXT
   
	i = LowWord(i&)

END SUB

SUB SplitPath (path$, drive$, directory$, fileName$, extension$)

	' Breaks the path specification in <path$> into its component parts
	' and returns those components in the other arguments. A null string
	' is returned for any component that is not present. Note that
	' <path$> is assumed to contain a properly formed path.

	drive$ = ""
	directory$ = ""
	fileName$ = ""
	extension$ = ""

	pathTemp$ = path$

	IF MID$(pathTemp$, 2, 1) = ":" THEN
		drive$ = MID$(pathTemp$, 1, 2)
		pathTemp$ = MID$(pathTemp$, 3)
	END IF

	i = InStrL(pathTemp$, "\")
	IF i THEN
		directory$ = LEFT$(pathTemp$, i)
		pathTemp$ = MID$(pathTemp$, i + 1)
	END IF

	i = INSTR(pathTemp$, ".")
	IF i THEN
		fileName$ = LEFT$(pathTemp$, i - 1)
		IF LEN(pathTemp$) > i THEN extension$ = MID$(pathTemp$, i)
	ELSE
		fileName$ = pathTemp$
	END IF

END SUB

SUB StartTimer

	' Sychronizes with the system timer tick and sets the global variable
	' gStartTime to the current elapsed time. Refer to the ElapsedTime!
	' function for more information.
	'
	' The return value for the TIMER function is based on the system
	' timer tick. Because the system timer tick has a period of ~55 ms,
	' the uncertainty in the return value is ~55 ms. The timing of an
	' interval requires two calls to the TIMER function for a total
	' uncertainty of ~110 ms. This uncertainty can be cut in half by
	' sychronizing the start of the interval with the system timer
	' tick.

	' The return value for the TIMER function updates after each system
	' timer tick. To sychronize with the system timer tick, we need to
	' wait for the return value to change.
	t! = TIMER
	DO
	LOOP WHILE TIMER = t!

	' Set the global variable.
	gStartTime! = TIMER

END SUB

FUNCTION TempFileSpec$ (fileSpec$)

	' Returns the file specification from <fileSpec$> with the filename
	' component modified such that the returned value forms a unique
	' file specification.
	'
	' The modifications consist of replacing as many characters from the
	' right end of the filename as necessary with an underscore followed
	' by one or more decimal digits.
	'
	' Note that a directory name must be followed with a "\" to be
	' recognized as such.
	'
	' Assuming this procedure is working as intended (judge for
	' yourself), the returned string can be used to open a temporary
	' file for OUTPUT without concern that an existing file may be
	' destroyed.
	'
	' This procedure does not support long filenames.
	'
	' Note that this procedure disables error handling.

	SplitPath fileSpec$, drive$, directory$, fileName$, extension$

	fileNumber = FREEFILE

	ON ERROR GOTO GeneralErrorHandler

	charCount = 6
	number& = 0
	triggerValue& = 10
	DO
		trialFileName$ = LEFT$(fileName$, charCount) + "_" + LTRIM$(STR$(number&))

		trialSpec$ = drive$ + directory$ + trialFileName$
		IF LEFT$(extension$, 1) = "." THEN
			trialSpec$ = trialSpec$ + LEFT$(extension$, 4)
		ELSE
			trialSpec$ = trialSpec$ + "." + LEFT$(extension$, 3)
		END IF
	 
		sErrorCode = 0
		OPEN trialSpec$ FOR INPUT AS fileNumber
		CLOSE fileNumber

		' If the OPEN FOR INPUT statement failed, then sErrorCode will
		' contain a non-zero value. If the value is cErrorFileNotFound,
		' then trialSpec$ contains a unique file specification so exit
		' from the loop. If the error code is any other non-zero value,
		' then disable error handling and simulate the corresponding
		' run-time error.
		IF sErrorCode THEN
			IF sErrorCode = cErrorFileNotFound THEN
				EXIT DO
			ELSE
				ON ERROR GOTO 0
				ERROR sErrorCode
		   END IF
		END IF

		number& = number& + 1
		IF number& = triggerValue& THEN
			triggerValue& = triggerValue& * 10
			charCount = charCount - 1
		END IF

		' This should never happen.
		IF charCount < 1 THEN EXIT FUNCTION
	LOOP
 
	ON ERROR GOTO 0

	TempFileSpec$ = trialSpec$

END FUNCTION

FUNCTION UInt& (i%)

	' Converts <i%> to a LONG without extending the sign bit and
	' returns the result.
	'
	' This provides a convenient way to manipulate an INTEGER as
	' an unsigned value. The more obvious coding, UInt& = i, would
	' extend the sign bit (for example, 8000h would convert to
	' FFFF8000h instead of 00008000h).

	' Make sure the current segment is the default data segment.
	DEF SEG

	POKE VARPTR(UInt), PEEK(VARPTR(i))
	POKE VARPTR(UInt) + 1, PEEK(VARPTR(i) + 1)

END FUNCTION

SUB UnTab (line$, tabWidth%)

	' Expands any tabs in <line$> to spaces using <tabWidth%>
	' to determine the location of the tab stops.
 
	' Note that currPos holds the current position in
	' the expanded string and that the first position
	' in the expanded string is position 0.
	FOR i = 1 TO LEN(line$)
		c$ = MID$(line$, i, 1)
		IF c$ = CHR$(9) THEN
			spaceCnt = tabWidth - (currPos MOD tabWidth)
			temp$ = temp$ + SPACE$(spaceCnt)
			currPos = currPos + spaceCnt
		ELSE
			temp$ = temp$ + c$
			currPos = currPos + 1
		END IF
	NEXT

	line$ = temp$

END SUB

FUNCTION Value (binary$)

	' Returns the numeric value of <binary$>, which is assumed
	' to contain a valid binary representation of an integer.

	length = LEN(binary$)
	FOR i = length TO 1 STEP -1
		IF MID$(binary$, i, 1) = "1" THEN v& = v& + 2 ^ (length - i)
	NEXT
	Value = LowWord(v&)

END FUNCTION

FUNCTION WindowsIsRunning

	' Returns true if Windows is running.
	'
	' Calls the Identify Windows Version and Type function. If the
	' function returns with AX set to 0, then the function is supported
	' and Windows is running. The other return values for the function
	' are: BH = major version, BL = minor version, CX = 2 for standard
	' mode (obsolete), CX = 3 for enhanced mode.

	DIM regX AS RegTypeX

	regX.ax = &H160A
	InterruptX &H2F, regX, regX
   
	IF regX.ax = 0 THEN WindowsIsRunning = cTrue

END FUNCTION

