' GENERAL.BI, v1.1
'
' Copyright (C) 2002 Michael Webster
'
' This program is distributed under the terms of the GNU General
' Public License. Refer to the BAS module for the complete notice.

' This file contains the constant, data, and procedure declarations
' for GENERAL.BAS.

' Declare constants for the text cursor start and stop values.
CONST cUnderlineStart = 7, cUnderlineStop = 8
CONST cFullStart = 0, cFullStop = 7

' Declare a global variable to store the current elapsed time. Note the
' block name "/general/" which reduces the possibility of problems with
' mismatched COMMON statements if this include file is mixed with other
' include files and/or COMMON statements.
COMMON SHARED /general/ gStartTime!

' Declare constants for the non-character keys, for selected key
' combinations, for the character keys that return ASCII control
' codes (ASCII codes < 32), and for the Spacebar. The values for
' the non-character keys and for the key combinations are the
' corresponding keyboard scan codes + 256. This prevents the
' values for these keys from overlapping the ASCII codes, thus
' eliminating the need for separate keyboard scan code and ASCII
' code values. Note that the Shift keys do not affect the return
' values for the cursor control keys.
CONST cKeyF1 = 59 + 256
CONST cKeyF2 = 60 + 256
CONST cKeyF3 = 61 + 256
CONST cKeyF4 = 62 + 256
CONST cKeyF5 = 63 + 256
CONST cKeyF6 = 64 + 256
CONST cKeyF7 = 65 + 256
CONST cKeyF8 = 66 + 256
CONST cKeyF9 = 67 + 256
CONST cKeyF10 = 68 + 256
CONST cKeyF11 = 133 + 256
CONST cKeyF12 = 134 + 256

CONST cKeyUp = 72 + 256
CONST cKeyDown = 80 + 256
CONST cKeyLeft = 75 + 256
CONST cKeyRight = 77 + 256

CONST cKeyInsert = 82 + 256
CONST cKeyDelete = 83 + 256
CONST cKeyHome = 71 + 256
CONST cKeyEnd = 79 + 256
CONST cKeyPageUp = 73 + 256
CONST cKeyPageDown = 81 + 256

CONST cKeyAltUp = 152 + 256
CONST cKeyAltDown = 160 + 256

CONST cKeyShiftTab = 15 + 256

CONST cKeyCtrlLeft = 115 + 256
CONST cKeyCtrlRight = 116 + 256
CONST cKeyCtrlHome = 119 + 256
CONST cKeyCtrlEnd = 117 + 256
CONST cKeyCtrlPageUp = 132 + 256
CONST cKeyCtrlPageDown = 118 + 256

CONST cKeyCtrlInsert = 146 + 256

CONST cKeyEnter = 13
CONST cKeyEscape = 27
CONST cKeyTab = 9
CONST cKeyBackspace = 8

CONST cKeySpace = 32

' Define constants for the carry flag and zero flag bits of the
' CPU (80x86) flags word. The value specified is the value of the
' corresponding bit. This allows us to test the flag with an AND
' operation and set the flag with an OR operation.
CONST cCarryFlag = 1
CONST cZeroFlag = &H40

' Define a data type to hold the address and length of a memory object.
TYPE MemDescriptorType
    segment AS LONG
    offset  AS LONG
    length  AS LONG
END TYPE

' Declare the procedures.
DECLARE FUNCTION Bin$ (i%)
DECLARE FUNCTION BitRangeValue% (i%, lowBit%, highBit%)
DECLARE FUNCTION Value% (binary$)

DECLARE FUNCTION HighByte% (i%)
DECLARE FUNCTION HighWord% (i&)
DECLARE FUNCTION LowByte% (i%)
DECLARE FUNCTION LowWord% (i&)

DECLARE SUB MakeInt (i%, hiByte%, loByte%)
DECLARE SUB MakeLong (i&, hiWord&, loWord&)
DECLARE FUNCTION UInt& (i%)

DECLARE SUB ShiftLeft (i%, n%)
DECLARE SUB ShiftRight (i%, n%)

DECLARE FUNCTION MemByte% (segment&, offset&)
DECLARE FUNCTION MemDword& (segment&, offset&)
DECLARE FUNCTION MemWord% (segment&, offset&)

DECLARE FUNCTION HexByte$ (i%)
DECLARE FUNCTION HexWord$ (i&)

DECLARE FUNCTION FarAddress& (s$)
DECLARE SUB GetMemDescriptor (memDescriptor AS MemDescriptorType, s$)

DECLARE FUNCTION IsVGA% ()

DECLARE SUB Delay (milliseconds%)

DECLARE FUNCTION AltPlusKeyCode% (key$)
DECLARE SUB GetKey (key$, keyCode%)
DECLARE SUB GetKeyStatus (key$, keyCode%)

DECLARE SUB PushKey (scanCode%, asciiCode%, fBufferFull%)
DECLARE SUB ReadEnhancedKeyboard (scanCode%, asciiCode%)
DECLARE SUB ReadEnhancedKeyboardStatus (fKeyWaiting%, scanCode%, asciiCode%)

DECLARE FUNCTION CtrlKey% ()
DECLARE FUNCTION AltKey% ()
DECLARE FUNCTION ShiftKey% ()

DECLARE SUB SplitPath (path$, drive$, directory$, fileName$, extension$)
DECLARE FUNCTION TempFileSpec$ (fileSpec$)

DECLARE FUNCTION Choice% (prompt$, keys$, default%)
DECLARE SUB InputString (prompt$, in$)
DECLARE SUB Pause (action$)

DECLARE FUNCTION InStrL% (string1$, string2$)
DECLARE FUNCTION StrTok$ (search$, delimiter$)
DECLARE FUNCTION GroupedDigits$ (digits$)

DECLARE SUB EnTab (line$, tabWidth%)
DECLARE SUB UnTab (line$, tabWidth%)

DECLARE SUB StartTimer ()
DECLARE FUNCTION ElapsedTime! ()

DECLARE SUB AddListItem (item$(), newItem$)

DECLARE FUNCTION WindowsIsRunning% ()

DECLARE SUB LogError (module$, procedure$, error$)

