' VGA.BAS, v1.2
'
' Copyright (C) 2002 Michael Webster
'
' This program is free software; you can redistribute it and/or
' modify it under the terms of the GNU General Public License as
' published by the Free Software Foundation; either version 2 of
' the License, or (at your option) any later version.
'
' This program is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public
' License along with this program; if not, write to the Free
' Software Foundation, Inc., 59 Temple Place, Suite 330,
' Boston, MA  02111-1307  USA

DEFINT A-Z

' This file contains a QuickBASIC module consisting of various VGA
' related procedures.
'
' The VGA has 256 KB of display memory that is logically divided into
' four 64 KB bit planes. In the graphics modes, the bit planes are
' used to store pixel color information. Depending on the mode,
' anywhere from one to all of the bit planes may be utilized. In the
' text (alphanumeric) modes, bit plane 0 is used to store the character
' codes, bit plane 1 the character attributes, and bit plane 2 the
' character font bit patterns.
'
' For the VGA, the Palette Registers are used to associate an attribute
' with the Color Register that contains the actual colors values that
' are displayed.
'
' For the VGA operating in 80x25 text mode (and assuming the default
' setting for the total number of scan lines), the character box is
' 9 pixels wide and 16 pixels (scan lines) high, so the display
' resolution is 720x400.
'
' Although the VGA can utilize character font bit patterns with up to
' 32 bytes (scan lines) per character, the use of bit patterns with
' more than 16 bytes per character requires special techniques and/or
' a reduction in the number of character rows.
'
' Most of the procedures in this module require that the active display
' adapter be VGA compatible, and many require that it be operating in
' a text (alphanumeric) mode. Refer to the IsVGA function in GENERAL.BAS
' for more information.
'
' The Load____CharacterSet procedures are intended for use with an 80x25
' text mode, regardless of the number of bytes per character of the font.
'
' Note that most of the VGA adapters that were tested exhibited behavior
' that deviated in some way from what is documented as "standard" in the
' available references. For example, in the state information returned
' by the VGA BIOS Return Functionality/State Information function,
' several of the adapters failed to properly set the bit that specifies
' the size of the active character set, and two always returned zero for
' the secondary block (the secondary character set) even when a secondary
' block other than block zero was active AND functional. For another
' example, for two of the adapters, the VGA BIOS function that is supposed
' to load the ROM 8x14 character set actually loaded the 8x16 character
' set. Refer to the GetVgaStateInfo and GetFontBlock procedures for more
' examples.
'
' The source for most of the information on which this module was
' based was Richard F. Ferraro's book Programmer's_Guide_to_the_EGA,
' _VGA,_and_Super_VGA_Cards (third edition), Addison-Wesley 1994.
' ISBN 0-201-62490-7. In addition to a large amount of information,
' the book contains many programming examples in C and Assembly.
'
' ===================================================================
' Changes in v1.2:
'
'   Added error logging feature to parameter check code.
'
' ===================================================================

' $INCLUDE: 'INTX.BI'
' $INCLUDE: 'ERROR.BI'
' $INCLUDE: 'GENERAL.BI'
' $INCLUDE: 'VGA.BI'

CONST cFalse = 0, cTrue = NOT cFalse

' *** START TEST CODE ***
' For dithered primaries, fraction each should be 192/(c+1)?
' random pixel placement probably better
' get values from color reg
' create 192 pixel block, or whatever size
' using value for number of randomly placed pixels
r = 3
SCREEN 12
GetColorValues r, red, green, blue
FOR i = 0 TO (red - 1) * 200
    x = RND * 100
    y = RND * 100
    PSET (x, y), 4
NEXT
FOR i = 0 TO (green - 1) * 200
    x = RND * 100
    y = RND * 100
    PSET (x, y), 2
NEXT
FOR i = 0 TO (blue - 1) * 200
    x = RND * 100
    y = RND * 100
    PSET (x, y), 1
NEXT



END
' cycle 124
c = 7
FOR y = 0 TO 97
    FOR x = 0 TO 97
        SELECT CASE c
            CASE 7
                c = 9
            CASE 9
                c = 12
            CASE 12
                c = 7
        END SELECT
        'c = x MOD 3 + 1
        'IF c = 3 THEN c = 4
        IF y AND 1 THEN PSET (x, y), c ELSE PSET (x + 1, y), c
    NEXT
NEXT


END
CLS
FOR r = 0 TO 15
    GetColorValues r, red, green, blue
    PRINT red, green, blue
NEXT

END

DIM biosMode(5 TO 13)
' 1 2 7 8 9 10 11 12 13
' VideoMode = MemByte(&H40, &H49)
FOR i = 7 TO 13
    SCREEN i
    biosMode(i) = VideoMode
NEXT
SCREEN 0
WIDTH 80
FOR i = 7 TO 13
    PRINT i, HEX$(biosMode(i))
NEXT

' *** END TEST CODE ***

FUNCTION BaseAddress

    ' Returns the base I/O port address for the active
    ' display adapter, as read from the BIOS data area.
   
    BaseAddress = MemWord(&H40, &H63)

END FUNCTION

FUNCTION CharacterWidth

    ' Returns the current character width in pixels, as determined from
    ' the Dot Clocks field of the Sequencer Clocking Mode Register.
    '
    ' This procedure requires that the active display adapter be VGA
    ' compatible.

    ' Write the index to the Sequencer Address register.
    OUT &H3C4, 1

    ' Read the value from Clocking Mode Register.
    cmrValue = INP(&H3C5)

    IF cmrValue AND 1 THEN CharacterWidth = 8 ELSE CharacterWidth = 9

END FUNCTION

FUNCTION CharScanLines

    ' Returns the current scan lines per character,
    ' as read from the BIOS data area.
  
    CharScanLines = MemByte(&H40, &H85)

END FUNCTION

SUB ClearFontBlock (block%)

    ' Clears the entire character font block specified by <block%>.
    '
    ' Refer to the LoadUserCharacterSet and Load8x8CharacterSet
    ' procedures for more information.
    '
    ' This procedure requires that the active display adapter be VGA
    ' compatible and operating in text mode.

    IF block < 0 OR block > 7 THEN
        LogError "VGA.BAS", "ClearFontBlock", "block < 0 OR block > 7"
        ERROR cErrorIllegalFunctionCall
    END IF

    ' The bit patterns occupy 32 bytes per character.
    DIM bp(1 TO 32 * 256)

    LoadUserCharacterSet block, 0, 32, bp()

    ERASE bp

END SUB

FUNCTION DisplayPage

    ' Returns the current display page, as read from the BIOS data area.
    '
    ' Note that this value corresponds to the visible screen page
    ' (vpage) argument for the SCREEN statement.

    DisplayPage = MemByte(&H40, &H62)

END FUNCTION

SUB GetColorValues (register%, red%, green%, blue%)

    ' Returns with <red%>, <green%>, and <blue%> set to
    ' the color intensity values from the Color Register
    ' specified by <register%>.
    '
    ' This procedure requires that the active display
    ' adapter be VGA compatible.

    DIM regX AS RegTypeX

    IF register < 0 OR register > 255 THEN
        LogError "VGA.BAS", "GetColorValues", "register < 0 OR register > 255"
        ERROR cErrorIllegalFunctionCall
    END IF

    regX.ax = &H1015
    regX.bx = register
    InterruptX &H10, regX, regX
    red = HighByte(regX.dx)
    green = HighByte(regX.cx)
    blue = LowByte(regX.cx)

END SUB

SUB GetFontBlock (block%, bytesPerCharacter%, bitPatterns%()) STATIC

    ' REDIM's <bitPatterns%()> to (1 to <bytesPerCharacter%> * 256)
    ' and loads the elements with the character font bit patterns
    ' for the entire 256-character character font block specified by
    ' <block%>.
    '
    ' Note that <bitPatterns%()> must be dimensioned as a DYNAMIC array.
    '
    ' Note that the value in <bytesPerCharacter%> should be >= the
    ' number of bytes per character for the font.
    '
    ' This procedure reads the bit patterns directly from bit plane 2.
    ' To avoid problems with the odd/even address shuffle that the
    ' VGA performs for all text modes, this procedure temporarily
    ' changes to graphics mode 11h. This mode was selected because
    ' read operations access only a single bit plane. This allows us
    ' to avoid having to filter the data to extract the relevant bits.
    '
    ' This procedure loads a block translation array on the first call.
    ' The element numbers for this array correspond to the character
    ' font block numbers for the VGA BIOS functions and the VGA hardware
    ' registers, and the value loaded into each element is the block in
    ' bit plane 2 that the VGA maps to the character font block number.
    '
    ' Note that there is some uncertainty as to what the "standard"
    ' mapping is. Several of the available references specify that the
    ' standard mapping is 0.0, 1.4, 2.1, 3.5, 4.2, 5.6, 6.3, 7.7, and
    ' one that the standard mapping is 0.0, 1.2, 2.4, 3.6, 4.1, 5.3,
    ' 6.5, 7.7. This procedure uses the latter because it duplicates
    ' the mapping implemented in all but one of the VGA adapters that
    ' were tested.
    '
    ' Refer to the LoadBitPattern procedure for more information.
    '
    ' All mode sets are performed without clearing the display memory.
    '
    ' This procedure requires that the active display adapter be VGA
    ' compatible and operating in text mode.

    DIM regX AS RegTypeX

    IF block < 0 OR block > 7 THEN
        LogError "VGA.BAS", "GetFontBlock", "block < 0 OR block > 7"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF bytesPerCharacter < 8 OR bytesPerCharacter > 32 THEN
        LogError "VGA.BAS", "GetFontBlock", "bytesPerCharacter < 8 OR bytesPerCharacter > 32"
        ERROR cErrorIllegalFunctionCall
    END IF

    REDIM bitPatterns(1 TO bytesPerCharacter * 256)

    ' Save the current video mode so we can restore it.
    vMode = VideoMode

    ' Simulate a run-time error if the current
    ' mode is one of the graphics modes.
    IF vMode > 3 AND vMode <> 7 THEN
        LogError "VGA.BAS", "GetFontBlock", "current video mode is graphics"
        ERROR cErrorIllegalFunctionCall
    END IF

    IF NOT fInitialized THEN
        DIM bitPlane2Block(0 TO 7)
        bitPlane2Block(1) = 2
        bitPlane2Block(2) = 4
        bitPlane2Block(3) = 6
        bitPlane2Block(4) = 1
        bitPlane2Block(5) = 3
        bitPlane2Block(6) = 5
        bitPlane2Block(7) = 7
        fInitialized = cTrue
    END IF

    ' Save the value from the Graphics Controller Read Map Select
    ' Register (index = 4). The Graphics Controller Address Register
    ' is at I/O port 3CEh and the Data Register at I/O port 3CFh.
    OUT &H3CE, 4
    rms = INP(&H3CF)

    ' Change to mode 11h. Set bit 7 of AL to specify
    ' that the display memory should not be cleared.
    regX.ax = &H11 OR &H80
    InterruptX &H10, regX, regX

    ' Set the Read Map Select Register so the next
    ' read operation will access bit plane 2.
    OUT &H3CE, 4
    OUT &H3CF, 2

    ' This segment address of the buffer is A000h for all graphics modes.
    segment& = &HA000&

    FOR charCode = 0 TO 255

        ' The blocks are 8 KB each and the bit
        ' patterns occupy 32 bytes per character.
        offset& = bitPlane2Block(block) * &H2000& + charCode * 32

        FOR i = 1 TO bytesPerCharacter
            bitPatterns(charCode * bytesPerCharacter + i) = MemByte(segment&, offset& + i - 1)
        NEXT
    NEXT

    ' Restore the original video mode without
    ' clearing the display memory.
    regX.ax = vMode OR &H80
    InterruptX &H10, regX, regX

    ' Restore the Read Map Select Register.
    OUT &H3CE, 4
    OUT &H3CF, rms

END SUB

SUB GetFontInfo (fontInfo AS FontInfoType, pointerSpecifier%)

    ' Returns character font information for the *current* character
    ' font and the address of the character set specified by
    ' <pointerSpecifier%> in the elements of <fontInfo>.
    '
    ' Constants for the value in <pointerSpecifier%> are declared in
    ' VGA.BI. Each constant corresponds to one of the six character
    ' fonts that are stored in the VGA BIOS ROM. The 8x14 and 8x16
    ' fonts are full (256-character) fonts. The 8x8 font is a full
    ' font that is split into two parts. The 9x14 and 9x16 supplement
    ' fonts are partial fonts that are intended to replace specific
    ' characters in the full fonts with versions that have a better
    ' appearance when displayed in a 9 pixel wide character box.
    '
    ' This procedure requires that the active display adapter be
    ' VGA compatible.

    DIM regX AS RegTypeX

    regX.ax = &H1130
    MakeInt regX.bx, pointerSpecifier, 0
    InterruptX &H10, regX, regX
    fontInfo.bytesPerCharacter = regX.cx
    fontInfo.characterRows = regX.dx + 1
    fontInfo.pointerSegment = regX.es
    fontInfo.pointerOffset = regX.bp

END SUB

SUB GetRomFont (pointerSpecifier%, bytesPerCharacter%, bitPatterns%())

    ' REDIM's <bitPatterns%()> and loads the elements with the character
    ' font bit patterns from the ROM resident character set specified by
    ' <pointerSpecifier%>. On return, the lower bound of <bitPatterns%()>
    ' will be 1, and the upper bound will be <bytesPerCharacter%> times
    ' 256 for the full character sets, or (<bytesPerCharacter%> + 1)
    ' times the number of characters for the supplements. Note that the
    ' first element for each character in the supplements will be used
    ' to store the character code.
    '
    ' Note that <bitPatterns%()> must be dimensioned as a DYNAMIC array.
    '
    ' Note that the value in <bytesPerCharacter%> should be >= the number
    ' of bytes per character for the font.
    '
    ' Note that <pointerSpecifier%> values of c8x8low128 or c8x8high128
    ' will cause this procedure to load the entire 8x8 character set.
    '
    ' Constants for the value in <pointerSpecifier%> are defined in
    ' VGA.BI.
    '
    ' Refer to the GetFontInfo procedure for more information.
    '
    ' This procedure requires that the active display adapter be
    ' VGA compatible.

    IF pointerSpecifier < c8x14 OR pointerSpecifier > c9x16supplement THEN
        LogError "VGA.BAS", "GetRomFont", "pointerSpecifier < c8x14 OR pointerSpecifier > c9x16supplement"
        ERROR cErrorIllegalFunctionCall
    END IF

    DIM fontInfo AS FontInfoType

    SELECT CASE pointerSpecifier
        CASE c8x8low128, c8x8high128
            REDIM bitPatterns(1 TO bytesPerCharacter * 256)
            GOSUB Get8x8low128
            GOSUB Get8x8high128
        CASE c8x14
            REDIM bitPatterns(1 TO bytesPerCharacter * 256)
            GOSUB Get8x14
        CASE c9x14supplement
            GOSUB Get9x14supplement
        CASE c8x16
            REDIM bitPatterns(1 TO bytesPerCharacter * 256)
            GOSUB Get8x16
        CASE c9x16supplement
            GOSUB Get9x16supplement
    END SELECT

EXIT SUB

Get8x8low128:
    GetFontInfo fontInfo, c8x8low128
    FOR c = 0 TO 127
        element = c * bytesPerCharacter + 1
        FOR b = 0 TO 7
            bitPatterns(element + b) = MemByte(UInt(fontInfo.pointerSegment), UInt(fontInfo.pointerOffset + c * 8 + b))
        NEXT
    NEXT
    RETURN

Get8x8high128:
    GetFontInfo fontInfo, c8x8high128
    FOR c = 128 TO 255
        element = c * bytesPerCharacter + 1
        FOR b = 0 TO 7
            bitPatterns(element + b) = MemByte(UInt(fontInfo.pointerSegment), UInt(fontInfo.pointerOffset + (c - 128) * 8 + b))
        NEXT
    NEXT
    RETURN

Get8x14:
    GetFontInfo fontInfo, pointerSpecifier
    FOR c = 0 TO 255
        element = c * bytesPerCharacter + 1
        FOR b = 0 TO 13
            bitPatterns(element + b) = MemByte(UInt(fontInfo.pointerSegment), UInt(fontInfo.pointerOffset + c * 14 + b))
        NEXT
    NEXT
    RETURN

Get8x16:
    GetFontInfo fontInfo, pointerSpecifier
    FOR c = 0 TO 255
        element = c * bytesPerCharacter + 1
        FOR b = 0 TO 15
            bitPatterns(element + b) = MemByte(UInt(fontInfo.pointerSegment), UInt(fontInfo.pointerOffset + c * 16 + b))
        NEXT
    NEXT
    RETURN

Get9x14supplement:
    GetFontInfo fontInfo, pointerSpecifier
    ' We need to count the number of characters in the set.
    ' The bit patterns are stored with the character code in
    ' the first byte. A character code of zero marks the end
    ' of the character set.
    offset = fontInfo.pointerOffset
    DO
        charCode = MemByte(UInt(fontInfo.pointerSegment), UInt(offset))
        IF charCode = 0 THEN EXIT DO
        cnt = cnt + 1
        offset = offset + 15
    LOOP
    REDIM bitPatterns(1 TO (bytesPerCharacter + 1) * cnt)
    FOR c = 0 TO cnt - 1
        element = c * bytesPerCharacter + 1
        FOR b = 0 TO 14
            bitPatterns(element + b) = MemByte(UInt(fontInfo.pointerSegment), UInt(fontInfo.pointerOffset + c * 15 + b))
        NEXT
    NEXT
    RETURN

Get9x16supplement:
    GetFontInfo fontInfo, pointerSpecifier
    ' We need to count the number of characters in the set.
    ' The bit patterns are stored with the character code in
    ' the first byte. A character code of zero marks the end
    ' of the character set.
    offset = fontInfo.pointerOffset
    DO
        charCode = MemByte(UInt(fontInfo.pointerSegment), UInt(offset))
        IF charCode = 0 THEN EXIT DO
        cnt = cnt + 1
        offset = offset + 17
    LOOP
    REDIM bitPatterns(1 TO (bytesPerCharacter + 1) * cnt)
    FOR c = 0 TO cnt - 1
        element = c * bytesPerCharacter + 1
        FOR b = 0 TO 16
            bitPatterns(element + b) = MemByte(UInt(fontInfo.pointerSegment), UInt(fontInfo.pointerOffset + c * 17 + b))
        NEXT
    NEXT
    RETURN

END SUB

SUB GetVgaStateInfo (state AS FunctionalityStateType)

    ' Calls the VGA BIOS Return Functionality/State Information
    ' function to load the VGA state information into the elements
    ' of <state>.
    '
    ' The function is supposed to return with AL set to 1Bh as an
    ' indication that the function is supported. Because several
    ' of the VGA adapters that were tested failed to do so even
    ' though the function was supported, this procedure ignores
    ' the return value.
    '
    ' This procedure requires that the active display adapter be
    ' VGA compatible.

    DIM regX AS RegTypeX

    regX.ax = &H1B00
    regX.es = VARSEG(state)
    regX.di = VARPTR(state)
    InterruptX &H10, regX, regX

END SUB

SUB Load8x14CharacterSet (block%)

    ' Loads the ROM 8x14 character set into the character font block
    ' specified by <block%>.
    '
    ' This procedure requires that the active display adapter be VGA
    ' compatible and operating in text mode.
  
    DIM regX AS RegTypeX

    IF block < 0 OR block > 7 THEN
        LogError "VGA.BAS", "Load8x14CharacterSet", "block < 0 OR block > 7"
        ERROR cErrorIllegalFunctionCall
    END IF

    regX.ax = &H1101
    regX.bx = block
    InterruptX &H10, regX, regX

END SUB

SUB Load8x16CharacterSet (block%)

    ' Loads the ROM 8x16 character set into the character font
    ' block specified by <block%>.
    '
    ' This procedure requires that the active display adapter be
    ' VGA compatible and operating in text mode.
   
    DIM regX AS RegTypeX

    IF block < 0 OR block > 7 THEN
        LogError "VGA.BAS", "Load8x16CharacterSet", "block < 0 OR block > 7"
        ERROR cErrorIllegalFunctionCall
    END IF

    regX.ax = &H1104
    regX.bx = block
    InterruptX &H10, regX, regX

END SUB

SUB Load8x8CharacterSet (block%)

    ' Loads the ROM 8x8 "eye-strain" character set into the character
    ' font block specified by <block%>.
    '
    ' Most of the VGA adapters that were tested had a problem with
    ' loading an 8 scan line character set when the adapter was
    ' programmed for 16 scan lines (bytes) per character. The loaded
    ' characters would display correctly in the upper 8 scan lines of
    ' the 16 scan line character box, and any residual bit patterns
    ' in the lower 8 scan lines of the characters would display as
    ' garbage. To correct this problem, this procedure clears the
    ' destination block before it loads the character set.
    '
    ' This procedure requires that the active display adapter be VGA
    ' compatible and operating in text mode.
  
    DIM regX AS RegTypeX

    IF block < 0 OR block > 7 THEN
        LogError "VGA.BAS", "Load8x8CharacterSet", "block < 0 OR block > 7"
        ERROR cErrorIllegalFunctionCall
    END IF

    ClearFontBlock block

    regX.ax = &H1102
    regX.bx = block
    InterruptX &H10, regX, regX

END SUB

SUB LoadUserCharacterSet (block%, charCode%, bytesPerCharacter%, bitPatterns%())
   
    ' Loads the character font bit patterns stored in the elements
    ' of <bitPatterns%()> into the character font block specified
    ' by <block%>, starting at the character offset specified by
    ' <charCode%>. The caller must dimension <bitPatterns%()> with
    ' a lower bound of 1 and an upper bound equal to the value in
    ' <bytesPerCharacter%> times the total number of characters.
    ' Note that the value in <bytesPerCharacter%> must be even.
    '
    ' For the VGA operating in text mode, bit plane 2 is reserved for
    ' the storage of character font bit patterns. This 64 KB plane is
    ' segmented into eight blocks of 8KB each. Each block can store
    ' the character font bit patterns for one 256-character "character
    ' set". The first block (the one that is loaded and used by
    ' default) is block 0. Refer to the GetFontBlock procedure for
    ' more information.
    '
    ' For the VGA operating in 80x25 text mode (and assuming the default
    ' setting for the total number of scan lines), the character box is
    ' 9 pixels wide and 16 pixels (scan lines) high.
    '
    ' Each element of <bitPatterns()> contains the bit pattern for
    ' one scan line.
    '
    ' The most significant bit of the bit pattern corresponds to the
    ' left-most pixel and the least significant bit corresponds to the
    ' eight pixel from the left.
    '
    ' A set bit in the bit pattern causes the corresponding pixel to
    ' be displayed in the foreground color, and a cleared bit causes
    ' the corresponding pixel to be displayed in the background color.
    '
    ' For most characters, the ninth pixel from the left is displayed
    ' in the background color. For the characters with ASCII codes
    ' from 192 to 223 inclusive, the ninth pixel is displayed in the
    ' same color as the eight pixel.
    '
    ' The VGA BIOS User Alpha Load with Reset function performs a
    ' mode set before the character bit patterns are loaded, but
    ' does so without clearing the display.
    '
    ' Note that a mode set will reload the default character set
    ' into block 0.
    '
    ' This procedure requires that the active display adapter be
    ' VGA compatible and operating in text mode.

    IF block < 0 OR block > 7 THEN
        LogError "VGA.BAS", "LoadUserCharacterSet", "block < 0 OR block > 7"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF bytesPerCharacter > 32 THEN
        LogError "VGA.BAS", "LoadUserCharacterSet", "bytesPerCharacter > 32"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF bytesPerCharacter MOD 2 THEN
        LogError "VGA.BAS", "LoadUserCharacterSet", "bytesPerCharacter not even"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF LBOUND(bitPatterns) <> 1 THEN
        LogError "VGA.BAS", "LoadUserCharacterSet", "LBOUND(bitPatterns) <> 1"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF UBOUND(bitPatterns) MOD bytesPerCharacter THEN
        LogError "VGA.BAS", "LoadUserCharacterSet", "UBOUND(bitPatterns) not integral multiple of bytesPerCharacter"
        ERROR cErrorIllegalFunctionCall
    END IF
    charCount = UBOUND(bitPatterns) \ bytesPerCharacter
    IF charCount > 256 THEN
        LogError "VGA.BAS", "LoadUserCharacterSet", "charCount > 256"
        ERROR cErrorIllegalFunctionCall
    END IF

    DIM regX AS RegTypeX
    
    ' To avoid using a significant amount of space in the default
    ' data segment, pack the bit patterns into an integer array
    ' (instead of placing them in a string).
    ub = charCount * bytesPerCharacter \ 2
    DIM bp(1 TO ub)
    FOR i = 1 TO ub
        MakeInt bp(i), bitPatterns(i * 2), bitPatterns(i * 2 - 1)
    NEXT
   
    regX.ax = &H1100
    MakeInt regX.bx, bytesPerCharacter, block
    regX.cx = charCount
    regX.dx = charCode
    regX.es = VARSEG(bp(1))
    regX.bp = VARPTR(bp(1))

    ' Attempt to hide any disruptions of the displayed image by
    ' synchronizing the function call with vertical retrace.
    ' If the function can complete everything within the vertical
    ' retrace time (which depends on the system speed, the vertical
    ' refresh rate, and whether the program is running in the
    ' programming environment and/or under Windows), then the
    ' disruptions should be invisible. Bit 3 of the Input Status #1
    ' Register is set when the display is in vertical retrace.
    ' The register read port is 3BAh for monochrome and 3DAh for
    ' color.
    inputStatus1 = BaseAddress AND &HFF0 OR &HA
    ' Wait until not VR.
    WAIT inputStatus1, 8, 8
    ' Wait until VR.
    WAIT inputStatus1, 8

    InterruptX &H10, regX, regX

    ERASE bp

END SUB

FUNCTION PaletteValue (register%)

    ' Returns the value from the Palette Register specified
    ' by <register%>.
    '
    ' This procedure requires that the active display adapter
    ' be VGA compatible.

    DIM regX AS RegTypeX

    IF register < 0 OR register > 15 THEN
        LogError "VGA.BAS", "PaletteValue", "register < 0 OR register > 15"
        ERROR cErrorIllegalFunctionCall
    END IF

    regX.ax = &H1007
    regX.bx = register
    InterruptX &H10, regX, regX
    PaletteValue = HighByte(regX.bx)

END FUNCTION

SUB RestoreVideoState (videoState%())
   
    ' Calls the VGA Alternate Select Save/Restore Video State
    ' function to restore the current video state from the
    ' elements of <videoState%()>.
    '
    ' Note that <videoState%()> MUST contain video state data
    ' that was saved by the SaveVideoState procedure.
    '
    ' This procedure requires that the active display adapter
    ' be VGA compatible.

    DIM regX AS RegTypeX

    regX.ax = &H1C02
    regX.bx = VARPTR(videoState(0))
    regX.cx = 7
    regX.es = VARSEG(videoState(0))
    InterruptX &H10, regX, regX

END SUB

SUB SaveVideoState (videoState%())

    ' REDIM's <videoState%()> and calls the VGA Alternate Select Save/
    ' Restore Video State function to save the current video state to
    ' the elements of <videoState%()>.
    '
    ' Note that <videoState%()> must be dimensioned as a DYNAMIC array.
    '
    ' Note that this procedure does not (and probably could not) save
    ' the state of the internal variables that QB uses to manage and
    ' operate the display subsystem.
    '
    ' Note that this procedure does not save the display memory.
    ' Doing so in the 80x25 text modes would require 4 KB of storage
    ' per display (screen) page and 8 KB of storage per character
    ' font block. Doing so in the graphics modes would require,
    ' depending on the mode, 64-256 KB of storage. By saving and
    ' restoring the display memory in addition to the state, it
    ' should be possible, at least in theory :), to switch between
    ' graphics mode and text mode without having to redraw anything.
    ' Note, however, that the QB display statements will generally
    ' not work correctly when the display subsystem is in a mode
    ' (or state) that was not programmed by QB.
    '
    ' Note that this procedure does not save any graphics mode fonts
    ' that may be stored in host memory.
    '
    ' This procedure requires that the active display adapter be VGA
    ' compatible.

    DIM regX AS RegTypeX

    ' Determine the size of the required buffer. Subfunction 0
    ' returns the size in 64-byte blocks. Set bits 2-0 of CX to
    ' specify that we want to save everything (the video hardware
    ' state, the video data areas, and the Palette and Color
    ' Register state.
    regX.ax = &H1C00
    regX.cx = 7
    InterruptX &H10, regX, regX

    ' Set the buffer size.
    REDIM videoState(0 TO regX.bx * 32 - 1)

    ' Call subfunction 1 to save the state.
    regX.ax = &H1C01
    regX.bx = VARPTR(videoState(0))
    regX.cx = 7
    regX.es = VARSEG(videoState(0))
    InterruptX &H10, regX, regX

    ' Per Ralf Brown's x86 Interrupt List, many BIOSes corrupt
    ' the video state in the process of storing it, so the state
    ' should be restored immediately after it is stored.
    RestoreVideoState videoState()

END SUB

SUB SetBlockSpecifier (primaryBlock%, secondaryBlock%, fDisableIntensity%)

    ' Loads the values specified by <primaryBlock%> and <secondaryBlock%>
    ' into the VGA's Character Map Select Register, and disables high-
    ' intensity colors if <fDisableIntensity%> is set to true.
    '
    ' These primary and secondary block fields of the Character Map
    ' Select Register specify which of the loaded character sets should
    ' be active. If the two values are the same, then they specify the
    ' block number of the active character set. If the two values are
    ' different, then each specifies one of two active character sets.
    ' When two character sets are active, bit 3 of a character's
    ' attribute byte, which normally controls foreground intensity,
    ' determines from which character set the character's bit patterns
    ' will be read. The bit patterns will be read from the primary block
    ' when bit 3 of the attribute byte is cleared, or from the secondary
    ' block when bit 3 of the attribute byte is set.
    '
    ' The value in <fDisableIntensity%> controls the translation of
    ' attribute values to color values. If the value is set to false,
    ' then the translation works normally and the attribute values
    ' 8-15 are translated to high-intensity colors. If the value is
    ' set to true, then this procedure loads the Color Plane Enable
    ' Register with a value that disables the most significant bit of
    ' the palette address, causing the attribute values 8-15 to be
    ' translated to the same colors as for attribute values 0-7.
    '
    ' If two character sets are active and <fDisableIntensity%> is
    ' set to false, then characters from the primary character set
    ' are displayed with a normal foreground and characters from the
    ' secondary character set are displayed with a high-intensity
    ' foreground.
    '
    ' If two character sets are active and <fDisableIntensity%> is
    ' set to true, then characters from both character sets are
    ' displayed with a normal foreground.
    '
    ' To display high-intensity colors for attribute values in the
    ' range 0-7, or normal-intensity colors for attribute values in
    ' the range 8-15, use the PALETTE statement to reassign colors.
    '
    ' Note that a mode set will restore the Color Plane Enable and
    ' Character Map Select registers to their default settings.
    '
    ' This procedure requires that the active display adapter be VGA
    ' compatible and operating in text mode.
    '
    ' Refer to the LoadBitPattern procedure for more information.

    CONST cColorPlaneEnableRegister = &H12

    DIM regX AS RegTypeX

    IF primaryBlock < 0 OR primaryBlock > 7 THEN
        LogError "VGA.BAS", "SetBlockSpecifier", "primaryBlock < 0 OR primaryBlock > 7"
        ERROR cErrorIllegalFunctionCall
    END IF
    IF secondaryBlock < 0 OR secondaryBlock > 7 THEN
        LogError "VGA.BAS", "SetBlockSpecifier", "secondaryBlock < 0 OR secondaryBlock > 7"
        ERROR cErrorIllegalFunctionCall
    END IF

    ' The following block of code calls the VGA BIOS SetPaletteRegister
    ' function to load the appropriate value into the Color Plane Enable
    ' Register.
    IF fDisableIntensity THEN cpeValue = 7 ELSE cpeValue = &HF
    regX.ax = &H1000
    MakeInt regX.bx, cpeValue, cColorPlaneEnableRegister
    InterruptX &H10, regX, regX

    ' The bits of the block numbers must be copied to the
    ' Character Map Select Register as follows:
    '
    '   primaryBlock bits 2, 1, 0 to register bits 4, 1, 0
    '
    '   secondaryBlock bits 2, 1, 0 to register bits 5, 3, 2
    '
    pb = primaryBlock
    IF pb AND 4 THEN pb = pb XOR &H14
    sb = secondaryBlock
    ShiftLeft sb, 2
    IF sb AND &H10 THEN sb = sb XOR &H30

    ' Call the VGA BIOS Set Block Specifier function.
    regX.bx = pb OR sb
    regX.ax = &H1103
    InterruptX &H10, regX, regX

END SUB

SUB ShowVgaStateInfo

    ' Displays selected elements of the state information returned by
    ' the VGA BIOS Return Functionality/State Information function,
    ' and the current character width in pixels.
    '
    ' This procedure requires that the active display adapter be VGA
    ' compatible.
    
    DIM state AS FunctionalityStateType

    GetVgaStateInfo state

    IF ASC(state.activeVideoMode) < 10 THEN
        PRINT "Active video mode            : "; HEX$(ASC(state.activeVideoMode))
    ELSE
        PRINT "Active video mode            : "; HEX$(ASC(state.activeVideoMode)); "h"
    END IF
    PRINT "Character columns            :"; state.characterColumns
    PRINT "Bytes per page               :"; UInt(state.bytesPerPage)
    PRINT "Current page offset          :"; state.currentPageOffset
    PRINT "Cursor start and end         : "; HEX$(state.cursorStartAndEnd); "h"
    PRINT "Current display page         :"; ASC(state.currentDisplayPage)
    PRINT "Base I/O port address        : "; HEX$(state.baseIoPortAddress); "h"
    PRINT "Character rows               :"; ASC(state.characterRows)
    PRINT "Bytes per character          :"; state.bytesPerCharacter
    PRINT "Active display code          : ";
    SELECT CASE ASC(state.activeDisplayCode)
        CASE 7
            PRINT "VGA with analog monochrome"
        CASE 8
            PRINT "VGA with analog color"
        CASE ELSE
            PRINT "Non-VGA"
    END SELECT
    PRINT "Current number of colors     :"; state.currentNumberOfColors
    PRINT "Current number of pages      :"; ASC(state.currentNumberOfPages)
    PRINT "Current number of scan lines : ";
    SELECT CASE ASC(state.currentNumberOfScanLines)
        CASE 0
            PRINT "200"
        CASE 1
            PRINT "350"
        CASE 2
            PRINT "400"
        CASE 3
            PRINT "480"
    END SELECT
    PRINT "Primary character set        :"; ASC(state.primaryCharacterSet)
    PRINT "Secondary character set      :"; ASC(state.secondaryCharacterSet)
    PRINT "Size of active character set : ";
    IF ASC(state.savePointerStateInfo) AND 1 THEN
        PRINT "512"
    ELSE
        PRINT "256"
    END IF
    
    PRINT "Character width in pixels    :"; CharacterWidth

END SUB

FUNCTION VideoMode

    ' Returns the current video mode, as read from the BIOS data area.

    VideoMode = MemByte(&H40, &H49)

END FUNCTION

SUB WriteCharacter (charCode%)

    ' Calls the VGA BIOS Write Character function to write the single
    ' character specified by <charCode%> to the current display page
    ' at the current cursor position. The character attribute at the
    ' current cursor position is not altered.
    '
    ' Unlike the PRINT statement, this procedure can be used to write
    ' *any* character to the screen (including CR, LF, BS, etc).
    '
    ' Note that this procedure does not update the cursor position.
    '
    ' To set the current cursor position, use the LOCATE statement.
    '
    ' To set the current display page, use the SCREEN statement with
    ' an appropriate vpage argument.
    '
    ' This procedure requires that the active display adapter be VGA
    ' compatible.

    IF charCode < 0 OR charCode > 255 THEN
        LogError "VGA.BAS", "WriteCharacter", "charCode < 0 OR charCode > 255"
        ERROR cErrorIllegalFunctionCall
    END IF

    DIM regX AS RegTypeX
  
    MakeInt regX.ax, &HA, charCode
    MakeInt regX.bx, DisplayPage, 0
    regX.cx = 1
    InterruptX &H10, regX, regX

END SUB

SUB WriteString (s$, fore%)

    ' Calls the VGA BIOS Write TTY function to write each character
    ' in <s$> to the current display page starting at the current
    ' cursor position.
    '
    ' After each character is written, the BIOS maintained cursor
    ' position is updated and the screen is scrolled as necessary.
    ' Note that QB's cursor position is NOT updated.
    '
    ' For the graphics modes, <fore%> specifies the foreground
    ' attribute for the characters. For the text modes, <fore%>
    ' is ignored and the character attributes are not altered.
    '
    ' The Write TTY function handles the CR, LF, BS, and Bell (ASCII
    ' code 7) characters similar to the way that the PRINT statement
    ' handles them.
    '
    ' To set the current display page, use the SCREEN statement with
    ' an appropriate vpage argument.
    '
    ' To set the current cursor position, use the LOCATE statement.

    DIM regX AS RegTypeX

    FOR i = 1 TO LEN(s$)
        MakeInt regX.ax, &HE, ASC(MID$(s$, i, 1))
        regX.bx = fore
        InterruptX &H10, regX, regX
    NEXT

END SUB

