' VGA.BI, v1.0
'
' Copyright (C) 2002 Michael Webster
'
' This program is distributed under the terms of the GNU General
' Public License. Refer to the BAS module for the complete notice.

' This file contains the constant, data, and procedure declarations
' for VGA.BAS.

' Define a data type for the video state table returned by the VGA BIOS
' Return Functionality/State Information function. The first two elements
' are the address of a static functionality table in the VGA BIOS ROM.
' The cursor position elements contain the row in the high byte and the
' column in the low byte. Note that the currentNumberOfScanLines,
' miscStateInfo, amountOfVideoMemory, and savePointerStateInfo elements
' either contain codes or are bit-encoded. Refer to the ShowStateInfo
' procedure for more information.
TYPE FunctionalityStateType
    functionalityTableOffset    AS INTEGER
    functionalityTableSegment   AS INTEGER
    activeVideoMode             AS STRING * 1
    characterColumns            AS INTEGER
    bytesPerPage                AS INTEGER
    currentPageOffset           AS INTEGER
    cursorPositionPage0         AS INTEGER
    cursorPositionPage1         AS INTEGER
    cursorPositionPage2         AS INTEGER
    cursorPositionPage3         AS INTEGER
    cursorPositionPage4         AS INTEGER
    cursorPositionPage5         AS INTEGER
    cursorPositionPage6         AS INTEGER
    cursorPositionPage7         AS INTEGER
    cursorStartAndEnd           AS INTEGER
    currentDisplayPage          AS STRING * 1
    baseIoPortAddress           AS INTEGER
    currentMode                 AS STRING * 1
    currentPalette              AS STRING * 1
    characterRows               AS STRING * 1
    bytesPerCharacter           AS INTEGER
    activeDisplayCode           AS STRING * 1
    inactiveDisplayCode         AS STRING * 1
    currentNumberOfColors       AS INTEGER
    currentNumberOfPages        AS STRING * 1
    currentNumberOfScanLines    AS STRING * 1
    primaryCharacterSet         AS STRING * 1
    secondaryCharacterSet       AS STRING * 1
    miscStateInfo               AS STRING * 1
    reserved1                   AS STRING * 3
    amountOfVideoMemory         AS STRING * 1
    savePointerStateInfo        AS STRING * 1
    reserved2                   AS STRING * 13
END TYPE

' Define a data type for the character font information returned by
' the GetFontInfo procedure. The bytesPerCharacter and characterRows
' elements are used to return information for the *current* character
' font. The pointer elements are used to return the address of the
' character set specified in the procedure's pointerSpecifier parameter.
TYPE FontInfoType
    bytesPerCharacter   AS INTEGER
    characterRows       AS INTEGER
    pointerSegment      AS INTEGER
    pointerOffset       AS INTEGER
END TYPE

' Declare constants for the GetFontInfo procedure's pointerSpecifier
' parameter (ROM based character sets only).
CONST c8x14 = 2
CONST c8x8low128 = 3
CONST c8x8high128 = 4
CONST c9x14supplement = 5
CONST c8x16 = 6
CONST c9x16supplement = 7

' Declare the procedures.
DECLARE FUNCTION BaseAddress% ()
DECLARE FUNCTION CharacterWidth% ()
DECLARE FUNCTION CharScanLines% ()
DECLARE SUB ClearFontBlock (block%)
DECLARE FUNCTION DisplayPage% ()
DECLARE SUB GetColorValues (register%, red%, green%, blue%)
DECLARE SUB GetFontBlock (block%, bytesPerCharacter%, bitPatterns%())
DECLARE SUB GetFontInfo (fontInfo AS FontInfoType, pointerSpecifier%)
DECLARE SUB GetRomFont (pointerSpecifier%, bytesPerCharacter%, bitPatterns%())
DECLARE SUB GetVgaStateInfo (state AS FunctionalityStateType)
DECLARE SUB Load8x14CharacterSet (block%)
DECLARE SUB Load8x16CharacterSet (block%)
DECLARE SUB Load8x8CharacterSet (block%)
DECLARE SUB LoadUserCharacterSet (block%, charCode%, bytesPerCharacter%, bitPatterns%())
DECLARE FUNCTION PaletteValue% (register%)
DECLARE SUB RestoreVideoState (videoState%())
DECLARE SUB SaveVideoState (videoState%())
DECLARE SUB SetBlockSpecifier (primaryBlock%, secondaryBlock%, fDisableIntensity%)
DECLARE SUB ShowVgaStateInfo ()
DECLARE FUNCTION VideoMode% ()
DECLARE SUB WriteCharacter (charCode%)
DECLARE SUB WriteString (s$, fore%)

